/********************************************************************************
 * E N T I T Y 									*
 ********************************************************************************
 * File: EntityEditMode.h							*
 *                                                                         	*
 * Copyright (C) A.R.Hardie-Bick 2004-2005                                      *
 *										*
 * This program is free software; you can redistribute it and/or modify		*
 * it under the terms of the GNU General Public License as published by		*
 * the Free Software Foundation; either version 2 of the License, or (at	*
 * your option) any later version.						*
 *										*
 * This program is distributed in the hope that it will be useful, but		*
 * WITHOUT ANY WARRANTY; without even the implied warranty of			*
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU		*
 * General Public License for more details.					*
 *										*
 * You should have received a copy of the GNU General Public License		*
 * along with this program; if not, write to the Free Software			*
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.			*
 *******************************************************************************/

#ifndef __ENTITYEDITMODE_H__
#define __ENTITYEDITMODE_H__

#include <rtems.h>
#include <chameleon.h>
#include <panel.h>
#include <string.h>
#include <stdio.h>

#include "EntityLevelMeter.h"
#include "EntityGlobal.h"
#include "EntityRepeat.h"
#include "EntitySound.h"
#include "EntityPanel.h"
#include "EntityMode.h"

#define SLOW_REPEAT_TIME 330		// Time for navigation keys to autorepeat

class EntityEditMode : public EntityMode {
	
	EntityGlobal*	global;		// Used solely to access fast edit flag
	
	EntitySound*	sound;		// This is argument of enter()
	ParaStructure*	paras;		// Structure, with no data, of parameters
	
	Parameter*	parameter;	// Which parameter is currently being edited
	Page*		page;		// Which page...
	Group*		group;		// and which group
	int		parameterIndex;	// an index into all the above

	Parameter*	stackParameter;	// Specific parameter pointer 
	
	int		source;		// Takes the value -1..n, representing all source offsets
	bool		matrixEdit;
		
	bool		firstEdit;		// First ever edit may require special attention 
	bool		matrixUsageDisplay;	// Indicates matrix size has been displayed
	
	EntityLevelMeter*	levelMeter;		// Meter object for display over master volume parameter
	void			checkLevelMeter();	// Check to see whether to initiate/terminate level metering
	
	void		displayUpdate();
	void		displayValueUpdate();
	void		displayMatrixUsage();
	void		displayMidiEdit();
	void		checkNextDspPara(int);		// Helps avoid selecting non-dsp paras when editing matrix
	
	Status		incrementValue(int);
	void	 	parameterException();
	void		setPotsByContext();
	void		sourceInc(int);
	void		mute();
	void		unMute();
	
	void		buttonOffEvent();

	void		edit();
	void		shift();
	void		partUp();
	void		partDown();
	void		groupUp();
	void		groupDown();
	void		pageUp();
	void		pageDown();
	void		paramUp();
	void		paramDown();
	void		valueUp();
	void		valueDown();

	void		shiftPartUp();
	void		shiftPartDown();
	void		shiftGroupUp();
	void		shiftGroupDown();
	void		shiftPageUp();
	void		shiftPageDown();
	void		shiftParamUp();
	void		shiftParamDown();
	void		shiftValueUp();
	void		shiftValueDown();
	
	void		groupUpKeyUp();
	void		groupDownKeyUp();

	void		editShift();
	
	void		pot0Event(int p);
	void		pot1Event(int p);
	void		pot2Event(int p);

	void		autoEvent(int);
	
public:

	Mode		enter(EntitySound*);	

	EntityEditMode(EntityPanel*, EntityGlobal*, ParaStructure*);
};

	
	
















#endif /* __ENTITYEDITMODE_H__ */
