/********************************************************************************
 * E N T I T Y 									*
 ********************************************************************************
 * File: EntityMidi.cpp								*
 *                                                                         	*
 * Copyright (C) A.R.Hardie-Bick 2004-2005					*
 * 										*
 * EntityMidi::entitySysex(MidiEvPtr Ev)	 				*
 * based on Monosynth code example (C) Soundart 2001-2002 			*
 *******************************************************************************/
  
#include "EntityMidi.h"

//#define DEBUG_MIDI

static void receiveAlarm(short midiRef) {

	EntityMidi *midi = (EntityMidi*) MidiGetInfo(midiRef);

	midi->receiveEvents();
}

void EntityMidi::receiveEvents() { 

	MidiEvPtr ev;
	
	while((ev = MidiGetEv(midiRef))) {
	
		if(global->midiEcho()) MidiSendIm(midiRef, MidiCopyEv(ev));
		
		sendMessage(midq, &ev, 4);
	}
}

void EntityMidi::Run() { 
	
	MidiEvPtr ev; 
	
	while(1) { 
		
		getMessage(midq, &ev);
	       	
		#ifdef DEBUG_MIDI
		
		if(EvType(ev) != typeActiveSens) {
			TRACE("Midi: 0x%.2X ",(int)EvType(ev));
			for(int f = 0; f < MidiCountFields(ev); f++) 
				TRACE("0x%.2X ", (int)MidiGetField(ev, f));
			TRACE("\n");
		}
		
		#endif
		
		// Jump to decode vector

		(this->*event[EvType(ev)])(ev); 

		MidiFreeEv(ev);
	} 
}

EntityMidi::EntityMidi(EntityGlobal *global_) {

	global = global_;
	
	synq = getQueueId("SYNQ");
	ctrq = getQueueId("CTRQ");
	midq = getQueueId("MIDQ");

	lastNote = -1;
	lastBank = -1;

	highResNRPNPending = false;

	sysexFlashBufferReadPos = 0;
	sysexFlashBufferWritePos = 0;
	
	// Initialise LUT of vectors for decoding incoming midi events
	
	for(int x = 0; x < N_MIDI_EVENT_TYPES; x++) event[x] = &EntityMidi::Null;

	event[typeKeyOn] 	= &EntityMidi::KeyOn;
	event[typeKeyOff] 	= &EntityMidi::KeyOff;
	event[typeCtrlChange]	= &EntityMidi::CtrlChange;
	event[typeProgChange]	= &EntityMidi::ProgChange;
	event[typeChanPress]	= &EntityMidi::ChanPress;
	event[typeKeyPress]	= &EntityMidi::KeyPress;
	event[typePitchWheel]	= &EntityMidi::PitchWheel;
	event[typeSysEx]	= &EntityMidi::SystemEx;
	event[typeReset]	= &EntityMidi::Reset;

	event[typeClock]	= &EntityMidi::Clock;
	
	// Initialise LUT of vectors for decoding controller events
	
	for(int x = 0; x < 128; x++) ctrlEvent[x] = &EntityMidi::CtrlCoarse;
	for(int x = 32; x < 64; x++) ctrlEvent[x] = &EntityMidi::CtrlFine;

	ctrlEvent[101]		= &EntityMidi::CtrlCoarse;	// RPN (not used)
	ctrlEvent[100]		= &EntityMidi::CtrlFine;
	
	ctrlEvent[99]		= &EntityMidi::CtrlCoarse;	// NRPN
	ctrlEvent[98]		= &EntityMidi::CtrlFine;
	ctrlEvent[97]		= &EntityMidi::CtrlDecrement;
	ctrlEvent[96]		= &EntityMidi::CtrlIncrement;
	
	ctrlEvent[7]		= &EntityMidi::CtrlVolume;
	ctrlEvent[10]		= &EntityMidi::CtrlPan;
	ctrlEvent[11]		= &EntityMidi::CtrlExpression;
	ctrlEvent[32]		= &EntityMidi::CtrlReserved;
	ctrlEvent[38]		= &EntityMidi::CtrlNonRegEdit;
	ctrlEvent[64]		= &EntityMidi::CtrlSustain;

	ctrlEvent[120]		= &EntityMidi::CtrlAllSoundOff;
	ctrlEvent[121]		= &EntityMidi::CtrlAllCtrOff;
	ctrlEvent[123]		= &EntityMidi::CtrlAllNotesOff;
	
	ctrlEvent[89]		= &EntityMidi::CtrlLFOSyncPhase;
	ctrlEvent[90]		= &EntityMidi::CtrlLFOSyncSustain;
	
	for(int x = 0; x < 128; x++) ctrlValue[x] = 0;

	ctrlValue[0] = -1;		// sound bank no change
	
	// Start Midishare
	
	if((midiRef = MidiOpen("ENTITY")) < 0) TRACE("ERROR: cannot open MidiShare\n");
	
	MidiSetInfo(midiRef, (void *) this);
	MidiSetRcvAlarm(midiRef, receiveAlarm);
	
	MidiConnect(0, midiRef, TRUE);
	MidiConnect(midiRef, 0, TRUE);
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
// Midishare acts as a bridge, facilitating the asynchronous creation of new midi 
// events by other tasks, that can safely call the new event methods below.
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

void EntityMidi::newKeyOnEvent(char note, char velocity) {	
	
	// Put new event into receive and transmit midi buffers

	MidiEvPtr ev;

	if ((ev = MidiNewEv(typeKeyOn))) {
	       	
		Pitch(ev) = note; 
		Vel(ev)  = velocity; 
		
		Chan(ev) = global->midiRxChan(); 
		MidiSendIm(0, MidiCopyEv(ev));

		Chan(ev) = global->midiTxChan();
		MidiSendIm(midiRef, ev);
	}
}

void EntityMidi::newKeyOffEvent(char note) {

	// Put new event into receive and transmit midi buffers

	MidiEvPtr ev;

	if ((ev = MidiNewEv(typeKeyOff))) {
	       	
		Pitch(ev) = note; 
		Vel(ev)  = 90; 
		
		Chan(ev) = global->midiRxChan(); 
		MidiSendIm(0, MidiCopyEv(ev));

		Chan(ev) = global->midiTxChan();
		MidiSendIm(midiRef, ev);
	}
}

void EntityMidi::newAllNotesOffEvent() {

	// Put new event into receive and transmit midi buffers

	MidiEvPtr ev;

	if ((ev = MidiNewEv(typeCtrlChange))) {
	       	
		MidiSetField(ev, 0, 123);	// All-notes-off
		MidiSetField(ev, 1, 0);		// always zero
		
		Chan(ev) = global->midiRxChan(); 
		MidiSendIm(0, MidiCopyEv(ev));

		Chan(ev) = global->midiTxChan();
		MidiSendIm(midiRef, ev);
	}
}

void EntityMidi::newCtrlChangeEvent(char controller, char value) {

	MidiEvPtr ev;

	if ((ev = MidiNewEv(typeCtrlChange))) {
	       	
		MidiSetField(ev, 0, controller);
		MidiSetField(ev, 1, value);
		
		Chan(ev) = global->midiTxChan();
		MidiSendIm(midiRef, ev);
	}
}

void EntityMidi::newCtrl14bEvent(short controller, short value) {

	MidiEvPtr ev;

	if ((ev = MidiNewEv(typeCtrl14b))) {
	       	
		MidiSetField(ev, 0, controller);
		MidiSetField(ev, 1, value);
		
		Chan(ev) = global->midiTxChan();
		MidiSendIm(midiRef, ev);
	}
}

void EntityMidi::newNonRegParamEvent(char parameter, char source, short value) {

	// This function generates four controller events:
	
	// controller 99 = Parameter coarse (matrix source, 0..17)
	// controller 98 = Parameter fine
	// controller 6  = Value coarse
	// controller 38 = Value fine
	
	MidiEvPtr ev;
	
	if ((ev = MidiNewEv(typeNonRegParam))) {
	       	
		MidiSetField(ev, 0, (source << 7) | parameter);
		MidiSetField(ev, 1, value);
		
		Chan(ev) = global->midiTxChan();
		MidiSendIm(midiRef, ev);
	}
}

void EntityMidi::newProgChangeEvent(char soundBank, char userSoundNumber) {
	
	// Always send sound bank and sound number 
	
	MidiEvPtr ev;

	if ((ev = MidiNewEv(typeCtrlChange))) {
	       	
		MidiSetField(ev, 0, 0);			
		MidiSetField(ev, 1, soundBank);		
		Chan(ev) = global->midiTxChan();
		MidiSendIm(midiRef, ev);
	}

	if ((ev = MidiNewEv(typeProgChange))) {		
	       	
		MidiSetField(ev, 0, userSoundNumber);
		Chan(ev) = global->midiTxChan();
		MidiSendIm(midiRef, ev);
	}
}

void EntityMidi::newSysexStoredSoundEvent(void *buffer, int size, int soundBank, int userSoundNumber) { 

	newSysexSoundEvent(buffer, size, soundBank, userSoundNumber, true);
}

void EntityMidi::newSysexEditedSoundEvent(void *buffer, int size, int soundBank, int userSoundNumber) {

	newSysexSoundEvent(buffer, size, soundBank, userSoundNumber, false);
}

void EntityMidi::newSysexSoundEvent(void *buf, int size, int soundBank, int userSoundNumber, bool stored) {

	unsigned char *buffer = (unsigned char*) buf;
	
	MidiEvPtr ev = MidiNewEv(typeSysEx);
	
	// SoundArt MIDI ID
	
	MidiAddField(ev, (long) 0x00);
	MidiAddField(ev, (long) 0x20);
	MidiAddField(ev, (long) 0x45);
	
	// Chameleon Product ID
	
	MidiAddField(ev, (long) 0x01);

	// Add Device ID 
	
	MidiAddField(ev, (long) global->midiDeviceId());

	// Entity Application ID
	
	MidiAddField(ev, (long) 0x01);
	MidiAddField(ev, (long) ENTITY_MODEL_NUMBER);

	// Message ID: Sound Dump
	
	if(stored) 	MidiAddField(ev, (long) 0x00);	// Stored Sound Dump
	else 		MidiAddField(ev, (long) 0x01);	// Edited Sound Dump

	// Define sound bank & number
	
	MidiAddField(ev, (long) soundBank);
	MidiAddField(ev, (long) userSoundNumber);
	
	// Sound Data
	
	convert8to7bits(buffer, size);	// Pack data for Sysex	

	for (int i = 0; i < size; i++) MidiAddField(ev, ((long) buffer[i]) & 0x7F);

	// Checksum
	
	BYTE checksum = 0;
	
	for (int i = 0; i < MidiCountFields(ev); i++) checksum ^= MidiGetField(ev, i);

	checksum &= 0x7F;
	
	MidiAddField(ev, (long) checksum);

	MidiSendIm(midiRef, ev);
}

void EntityMidi::newSysexStoredGlobalEvent(void *buffer, int size) { newSysexGlobalEvent(buffer, size, true); }

void EntityMidi::newSysexEditedGlobalEvent(void *buffer, int size) { newSysexGlobalEvent(buffer, size, false); }

void EntityMidi::newSysexGlobalEvent(void *buf, int size, bool stored) {

	unsigned char *buffer = (unsigned char*) buf;
	
	MidiEvPtr ev = MidiNewEv(typeSysEx);
	
	// SoundArt MIDI ID
	
	MidiAddField(ev, (long) 0x00);
	MidiAddField(ev, (long) 0x20);
	MidiAddField(ev, (long) 0x45);
	
	// Chameleon Product ID
	
	MidiAddField(ev, (long) 0x01);

	// Add Device ID 
	
	MidiAddField(ev, (long) global->midiDeviceId());

	// Entity Application ID
	
	MidiAddField(ev, (long) 0x01);
	MidiAddField(ev, (long) ENTITY_MODEL_NUMBER);

	// Message ID: Global Dump
	
	if(stored) 	MidiAddField(ev, (long) 0x02);	// Stored Global Dump
	else 		MidiAddField(ev, (long) 0x03);	// Edited Global Dump

	// Data

	convert8to7bits(buffer, size);		// Pack bytes for Sysex	

	for (int i = 0; i < size; i++) MidiAddField(ev, ((long) buffer[i]) & 0x7F);

	// Checksum
	
	BYTE checksum = 0;
	
	for (int i = 0; i < MidiCountFields(ev); i++) checksum ^= MidiGetField(ev, i);
	
	checksum &= 0x7F;
	
	MidiAddField(ev, (long) checksum);

	MidiSendIm(midiRef, ev);
}

void EntityMidi::convert8to7bits(void *buffer, int &size) { 
	
	unsigned char *buf = (unsigned char*) buffer;

	int 	nBits = size * 8;
	int	n7Bits = nBits / 7;
	int	n7r = nBits % 7;
	
	if(n7r) n7Bits++;

	int	lsl = n7r - 1; 	// ...6 5 4 3 2 1 0
	int	lsr = 8 - n7r; 	// ...1 2 3 4 5 6 7
		
	if(lsl < 0) { lsl = 6; lsr = 1; }

	unsigned char srcByte;
	
	buf[n7Bits-1] = 0;
	
	for(int x = size-1, y = n7Bits-1; y > 0;) {

		srcByte = buf[x--];

		buf[y--] |= srcByte >> lsr++;
		if(y >= 0) buf[y] = 0x7F & (srcByte << lsl--);
		
		if(lsl < 0) { 
			
			lsl = 6; 
			lsr = 1; 
			
			buf[--y] = 0; 
		}
	}
		
	size = n7Bits;
}	

void EntityMidi::convert7to8bits(void *buffer, int &size) { 

	unsigned char *buf = (unsigned char*) buffer;

	int 	nBits = size * 7;
	int	n8Bits = nBits / 8;
	
	int	lsr = 0;	// 0 1 2 3 4 5 6...
	int	lsl = 7;	// 7 6 5 4 3 2 1...

	unsigned char dstByte;
	
	for(int x = 0, y = 0; x < size; ) {

		dstByte = buf[x++] >> lsr++;
		dstByte |= buf[x] << lsl--;
	
		buf[y++] = dstByte;
		
		if(lsr == 7) { 
			
			lsr = 0; 
			lsl = 7;
		       	
			x++;
		}
	}
		
	size = n8Bits;
}

bool EntityMidi::sysexFlashBufferFull() {
	
	int rp = sysexFlashBufferReadPos;
	int wp = sysexFlashBufferWritePos;
	
	if(rp <= wp) rp += SYSEX_FLASH_BUFFER_SIZE;

	if((rp - wp) <= ENTITY_SYSEX_MAX_MESSAGE_SIZE) {
		
		TRACE("WARNING: sysexFlashBufferFull, probably a bug\n");
		return true;
	}

	return false;
}

unsigned char EntityMidi::writeSysexFlashBuffer(unsigned char v) { 
	
	return sysexFlashBuffer[sysexFlashBufferWritePos = (SYSEX_FLASH_BUFFER_SIZE-1) & (sysexFlashBufferWritePos+1)] = v; 
}

unsigned char EntityMidi::readSysexFlashBuffer() {
	
	return sysexFlashBuffer[sysexFlashBufferReadPos = (SYSEX_FLASH_BUFFER_SIZE-1) & (sysexFlashBufferReadPos+1)]; 
}

int EntityMidi::sysexFlashBufferGetWritePos() {
	
	return sysexFlashBufferWritePos; 
}

bool EntityMidi::sysexFlashBufferSetReadPos(int rp) {
	
	if(rp > (SYSEX_FLASH_BUFFER_SIZE-1)) return false;
	sysexFlashBufferReadPos = rp; 
	return true;
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
// 				MIDI DECODE VECTORS
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 

void EntityMidi::KeyOn(MidiEvPtr ev) { 

	highResNRPNPending = false;	// See EntityMidi::CtrlNonRegEdit(..)
	
	if(!global->midiReceive(Chan(ev))) return;

	int v = Vel(ev);
	int n = Pitch(ev);
	
	if(v < 1) return KeyOff(ev);  	// zero velocity?

	mBuf[0] = (char) MIDI_NOTE_ON;
	mBuf[1] = (char) n;
	mBuf[2] = (char) v;
	
	sendMessage(synq, mBuf, 3);

	lastNote = n;   
}

void EntityMidi::KeyOff(MidiEvPtr ev) { 

	highResNRPNPending = false;
	
	if(!global->midiReceive(Chan(ev))) return;

	mBuf[0] = (char) MIDI_NOTE_OFF;
	mBuf[1] = (char) Pitch(ev);

	sendMessage(synq, mBuf, 2);
}

void EntityMidi::ProgChange(MidiEvPtr ev) { 

	highResNRPNPending = false;
	
	if(!global->midiReceive(Chan(ev))) return;
	if(!global->midiProgReceive()) return;

	sendMessage(ctrq, SOUND_MODE);
	
	// ctrlValue[0] = sound bank
	
	mBuf[0] = (char) MIDI_PROG_CHANGE;
	mBuf[1] = (char) ctrlValue[0];
	mBuf[2] = (char) MidiGetField(ev, 0);

	sendMessage(ctrq, mBuf, 3); // CTRQ!

	// A word about interpreting this message (by CTRQ client):
	// If the bank is -1, it means "use the current sound bank";
	// if it is non-zero, it means switch to the specified bank.
	// The last byte is the program offset within a bank.
	
	ctrlValue[0] = -1;

	// Once the last received bank change message has been used 
	// (here), it is reset, so another bank change message will
	// be required to make the next prog change select a sound
	// in a specified bank.
	
	// This is in accordance with general practice for sound 
	// bank switching.
}

void EntityMidi::ChanPress(MidiEvPtr ev) { 

	highResNRPNPending = false;
	
	if(!global->midiReceive(Chan(ev))) return;
	if(!global->midiCtrlReceive()) return;
	
	int v = MidiGetField(ev, 0);

	mBuf[0] = (char) MIDI_CONTROLLER;
	
	// Controller #32 reserved in global ctrLUT 
	// for aftertouch (chan/poly agnostic)

	if((mBuf[1] = global->ctrLookup(32))) {

		mBuf[2] = (char) v;
		mBuf[3] = 0;

		sendMessage(synq, mBuf, 4);
	}	
}

void EntityMidi::KeyPress(MidiEvPtr ev) {

	highResNRPNPending = false;
	
	if(!global->midiReceive(Chan(ev))) return;
	if(!global->midiCtrlReceive()) return;
	
	// Key pressure and channel pressure are not distinguished. 
	// The global mapping merely specifies aftertouch, and 
	// both get applied identically. Key pressure is only
	// applied if it's for the last note:

	if(MidiGetField(ev, 0) != lastNote) return;
	
	int v = MidiGetField(ev, 1);

	mBuf[0] = (char) MIDI_CONTROLLER;
	
	// Controller #32 reserved in global ctrLUT 
	// for aftertouch (chan/poly agnostic)

	if((mBuf[1] = global->ctrLookup(32))) {

		mBuf[2] = (char) v;
		mBuf[3] = 0;

		sendMessage(synq, mBuf, 4);
	}	
}

void EntityMidi::PitchWheel(MidiEvPtr ev) { 

	highResNRPNPending = false;
	
	if(!global->midiReceive(Chan(ev))) return;
	
	mBuf[0] = (char) MIDI_PITCH;
	mBuf[1] = (char) MidiGetField(ev, 1); // MS7b
	mBuf[2] = (char) MidiGetField(ev, 0); // LS7b

	sendMessage(synq, mBuf, 3);
}

void EntityMidi::Reset(MidiEvPtr ev) { 

	highResNRPNPending = false;
	
	if(!global->midiReceive(Chan(ev))) return;
	
	sendMessage(synq, MIDI_RESET);
}

void EntityMidi::Clock(MidiEvPtr ev) { 
	
	sendMessage(synq, MIDI_CLOCK); 
}

void EntityMidi::Null(MidiEvPtr ev) { }

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
// 				Controller Vectors
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 

void EntityMidi::CtrlChange(MidiEvPtr ev) {
	
	if(!global->midiReceive(Chan(ev))) return;

	unsigned int controller = MidiGetField(ev, 0);
	
	// Select controller method by number

	if(controller < 128) (this->*ctrlEvent[controller])(ev);
}

void EntityMidi::CtrlLFOSyncSustain(MidiEvPtr ev) { 

	if(MidiGetField(ev, 1) > 63) 	sendMessage(synq, MIDI_LFO_SYNC_SUSTAIN_ON);
	else 				sendMessage(synq, MIDI_LFO_SYNC_SUSTAIN_OFF);
}

void EntityMidi::CtrlLFOSyncPhase(MidiEvPtr ev) { 

	mBuf[0] = MIDI_LFO_SYNC_PHASE;
	mBuf[1] = MidiGetField(ev, 1);

	sendMessage(synq, mBuf, 2);
}

void EntityMidi::CtrlCoarse(MidiEvPtr ev) {

	highResNRPNPending = false;

	char c = MidiGetField(ev, 0);
	char v = MidiGetField(ev, 1);

	if(!c) { 
		
		// Ctrl 0 is program bank select (takes effect on prog receive)
		
		if(!global->midiProgReceive()) return;
		ctrlValue[0] = v;

		return; // Otherwise CTR1, 2 & 3 get zeroed
	}
	
	ctrlValue[c] = v;
	
	/* Implement intelligence for NRPN data interpretation switch */
	if(c == 6) { highResNRPNPending = true; return; }
	
	if(!global->midiCtrlReceive()) return;

	mBuf[0] = (char) MIDI_CONTROLLER;
	
	// Look up bit pattern defining modulation sources mapped 
	// to the incoming midi controller event:
	// 
	// (CTR1 = 1, CTR2 = 2, CTR3 = 4, MMOD = 8)

	if((mBuf[1] = global->ctrLookup(c))) {

		mBuf[2] = v;
	
		mBuf[3] = c < 32 ? ctrlValue[c+32] : 0;

		sendMessage(synq, mBuf, 4);
	}
}
	
void EntityMidi::CtrlFine(MidiEvPtr ev) {

	highResNRPNPending = false;	// Fine data entry is not received here

	char c = MidiGetField(ev, 0);
	char v = MidiGetField(ev, 1);

	ctrlValue[c] = v;

	if(!global->midiCtrlReceive()) return;

	mBuf[0] = (char) MIDI_CONTROLLER;
	
	// Look up bit pattern defining modulation sources mapped 
	// to the incoming midi controller event:
	// 
	// (CTR1 = 1, CTR2 = 2, CTR3 = 4, MMOD = 8)

	if((mBuf[1] = global->ctrLookup(c-32))) {

		mBuf[2] = ctrlValue[c-32];

		mBuf[3] = v;
	
		sendMessage(synq, mBuf, 4);
	}
}

void EntityMidi::CtrlReserved(MidiEvPtr ev) { 
       
	highResNRPNPending = false;

 // Controller #32 (Bank fine) is an internally 
 // reserved controller number (see EntityGlobal.cpp)

}

void EntityMidi::CtrlAllCtrOff(MidiEvPtr ev) {

	// Effect of All-Controllers-Off (121) message:
	// 
	// 	Volume 		   100
	// 	Expression 	   127
	// 	Sustain 	     0
	// 	Pitch wheel	centre
	// 	MMOD		     0
	// 	CTR1		centre
	// 	CTR2		centre
	// 	CTR3		centre

	mBuf[0] = MIDI_VOLUME; 
	mBuf[1] = ctrlValue[7] = 100; 
	mBuf[2] = ctrlValue[39] = 0; 
	
	sendMessage(synq, mBuf, 3);

	mBuf[0] = MIDI_PAN; 
	mBuf[1] = ctrlValue[10] = 64; 
	mBuf[2] = ctrlValue[42] = 0; 
	
	sendMessage(synq, mBuf, 3);

	mBuf[0] = MIDI_EXPRESSION; 
	mBuf[1] = ctrlValue[11] = 127; 
	mBuf[2] = ctrlValue[43] = 0; 
	
	sendMessage(synq, mBuf, 3);

	highResNRPNPending = false;

	mBuf[0] = (char) MIDI_SUSTAIN;
	mBuf[1] = (char) 0;
	
	sendMessage(synq, mBuf, 2);

	mBuf[0] = (char) MIDI_PITCH;
	mBuf[1] = (char) 64;
	mBuf[2] = (char) 0;

	sendMessage(synq, mBuf, 3);

	mBuf[0] = (char) MIDI_CONTROLLER;
	
	// Use bit pattern defining modulation source
	// (CTR1 = 1, CTR2 = 2, CTR3 = 4, MMOD = 8)

	// MMOD
	mBuf[1] = 8; mBuf[2] = 0; mBuf[3] = 0;
	sendMessage(synq, mBuf, 4);

	// CTR1
	mBuf[1] = 1; mBuf[2] = 64; mBuf[3] = 0;
	sendMessage(synq, mBuf, 4);

	// CTR2
	mBuf[1] = 2; mBuf[2] = 64; mBuf[3] = 0;
	sendMessage(synq, mBuf, 4);

	// CTR3
	mBuf[1] = 4; mBuf[2] = 64; mBuf[3] = 0;
	sendMessage(synq, mBuf, 4);

	highResNRPNPending = false;
}
	
void EntityMidi::CtrlAllNotesOff(MidiEvPtr ev) {

	highResNRPNPending = false;
	sendMessage(synq, MIDI_ALL_NOTES_OFF);
}

void EntityMidi::CtrlAllSoundOff(MidiEvPtr ev) {

	highResNRPNPending = false;
	sendMessage(synq, MIDI_ALL_SOUND_OFF);
}

void EntityMidi::CtrlVolume(MidiEvPtr ev) {

	highResNRPNPending = false;

	if(!global->midiVolReceive()) return;

	ctrlValue[7] = MidiGetField(ev, 1); 
	
	mBuf[0] = MIDI_VOLUME; 
	mBuf[1] = ctrlValue[7]; 
	mBuf[2] = ctrlValue[39]; 
	
	sendMessage(synq, mBuf, 3);
}

void EntityMidi::CtrlPan(MidiEvPtr ev) {

	highResNRPNPending = false;

	ctrlValue[10] = MidiGetField(ev, 1); 
	
	mBuf[0] = MIDI_PAN; 
	mBuf[1] = ctrlValue[10]; 
	mBuf[2] = ctrlValue[42]; 
	
	sendMessage(synq, mBuf, 3);
}

void EntityMidi::CtrlExpression(MidiEvPtr ev) {

	highResNRPNPending = false;

	if(!global->midiVolReceive()) return;

	ctrlValue[11] = MidiGetField(ev, 1); 
	
	mBuf[0] = MIDI_EXPRESSION; 
	mBuf[1] = ctrlValue[11]; 
	mBuf[2] = ctrlValue[43]; 
	
	sendMessage(synq, mBuf, 3);
}

void EntityMidi::CtrlSustain(MidiEvPtr ev) {

	highResNRPNPending = false;

	mBuf[0] = MIDI_SUSTAIN;
	mBuf[1] = MidiGetField(ev, 1);
       	
	sendMessage(synq, mBuf, 2);
}		

void EntityMidi::CtrlIncrement(MidiEvPtr ev) {

	highResNRPNPending = false;

	if(!global->midiEditReceive()) return;

	mBuf[0] = (char) MIDI_NRPN_INCREMENT;

	mBuf[1] = ctrlValue[99];	// Parameter coarse
	mBuf[2] = ctrlValue[98];	// Parameter fine
	
	sendMessage(synq, mBuf, 3);
}
	
void EntityMidi::CtrlDecrement(MidiEvPtr ev) {

	if(!global->midiEditReceive()) return;

	mBuf[0] = (char) MIDI_NRPN_DECREMENT;

	mBuf[1] = ctrlValue[99];	// Parameter coarse
	mBuf[2] = ctrlValue[98];	// Parameter fine
	
	sendMessage(synq, mBuf, 3);
}
	
void EntityMidi::CtrlNonRegEdit(MidiEvPtr ev) {

	// See nrpn.txt for list of editable parameters

	if(!global->midiEditReceive()) {
		
		highResNRPNPending = false;
		return;
	}
	
	// This is the vector for controller 38 (data entry fine)
	// It is assumed that values for controllers 99 and 98  
	// have already been received, and the parameter and 
	// value are generated now that the full data is available.
	
	ctrlValue[38] = MidiGetField(ev, 1);
	
	if(highResNRPNPending) 	{
		
	// If a coarse data entry value immediately preceded this
	// reception of the fine data entry value, then this triggers
	// a high-resolution interpretation of the NRPN, in other
	// words, no translation is performed from the NRPN 14 bit
	// data value to the user value (apart from subtracting 8192).

		mBuf[0] = (char) MIDI_NRPN_HI_RES_UPDATE;

		mBuf[1] = ctrlValue[99];	// Parameter coarse
		mBuf[2] = ctrlValue[98];	// Parameter fine
		mBuf[3] = ctrlValue[6];		// Value coarse
		mBuf[4] = ctrlValue[38];	// Value fine
	
		sendMessage(synq, mBuf, 5);
		
		highResNRPNPending = false;
	}

	else {
	
	// If a coarse data entry value did not immediately precede this
	// fine data entry value, then the NRPN is interpreted as a low
	// resolution control signal, whose scaling and offset is 
	// adjusted on a parameter-by-parameter basis in accordance
	// with what would be appropriate for control by a 128-step 
	// pot.
		
		mBuf[0] = (char) MIDI_NRPN_LO_RES_UPDATE;
		
		mBuf[1] = ctrlValue[99];	// Parameter coarse
		mBuf[2] = ctrlValue[98];	// Parameter fine
		mBuf[3] = ctrlValue[38];	// Value fine
	
		sendMessage(synq, mBuf, 4);
	}
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
//
// 	SYSTEM EXCLUSIVE 
// 				
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
// 
//	ENTITY System Exclusive Encoding
//	--------------------------------
//
// 	0xF0		Start of Sysex
//	0x00 0x20 0x45	SoundArt MIDI ID
//	0x01		Product ID (Chameleon = 0x01)
//	xxxx		Device ID (0x00..0x0F, 0x7F = broadcast)
//	0x01 0x??	Application ID (0x01 0x?? = Entity family)
//	xxxx		Message ID
//	....		Data (size variable)
//	xxxx		Checksum (XOR of all data bytes truncated to 7 bits, 0x7F is always valid)
//	0xF7		End of Sysex
//
//	Stored Sound Dump (when sent to synth, sound is saved automatically)
//	--------------------------------------------------------------------
//	
//	0x00		Message ID (0x00 = Stored Sound Dump)
//	xxxx		Sound Bank
//	yyyy		User sound number 
//	....		Dump data (compressed sound data, variable length)
//	
//	Edited Sound Dump (when sent to synth, sound is updated but not saved)
//	----------------------------------------------------------------------
//	
//	0x01		Message ID (0x01 = Edited Sound Dump)
//	xxxx		Sound Bank
//	yyyy		User sound number 
//	....		Dump data (compressed sound data, variable length)
//	
//	Stored Global Dump (new global settings saved automatically)
//	------------------------------------------------------------
//	
//	0x02		Message ID (0x02 = Stored Sound Dump)
//	....		Dump data (global parameters)
//	
//	Edited Global Dump (global settings updated but not saved)
//	----------------------------------------------------------
//	
//	0x03		Message ID (0x03 = Edited Sound Dump)
//	....		Dump data (global parameters)
//	
//	Stored Sound Request
//	--------------------
//	
//	0x04		Message ID (0x04 = Stored Sound Request)
//	xxxx		Sound bank 
//	yyyy		User sound number 
//	
//	Edited Sound Request
//	--------------------
//	
//	0x05		Message ID (0x05 = Edited Sound Request)
//	xxxx		Sound bank 
//	yyyy		User sound number 
//	
//	Stored Active Sound Request
//	---------------------------
//	
//	0x06		Message ID (0x06 = Stored Active Sound Request)
//	
//	Edited Active Sound Request
//	---------------------------
//	
//	0x07		Message ID (0x07 = Edited Active Sound Request)
//	
//	Stored Global Request
//	---------------------
//	
//	0x08		Message ID (0x08 = Stored Global Request)
//	
//	Edited Global Request
//	---------------------
//	
//	0x09		Message ID (0x09 = Edited Global Request)
//	
//	Sound Bank Request
//	------------------
//	
//	0x0A		Message ID (0x0A = Sound Bank Request)
//	xxxx		Sound bank
//
//	All Sounds Request
//	------------------
//	
//	0x0B		Message ID (0x0B = All Sounds Request)
//
//	All Sounds & Global Request
//	---------------------------
//	
//	0x0C		Message ID (0x0C = All Sounds and Global Request)
//
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 

// The System Exclusive code below is based on CMonoSynthEngine::OnMidiSysex()
// (and others) from Chameleon.sdk/src/examples/monosynth/MonoSynthEngine.cpp

void EntityMidi::SystemEx(MidiEvPtr ev) { 

	//for(int x = 0; x < 10; x++) TRACE("Sysex field %i = %i\n", x, (int)MidiGetField(ev, x));
		
	highResNRPNPending = false;
	if(entitySysex(ev)) return;
	if(tuningSysex(ev)) return;
}

bool EntityMidi::entitySysex(MidiEvPtr ev) {
	
	// Check message size is reasonable

	if (MidiCountFields(ev) < 9) return false;				// Mininum for Soundart
	if (MidiCountFields(ev) > ENTITY_SYSEX_MAX_MESSAGE_SIZE) return false;	// Maximum for Entity

	// Check for SoundArt MIDI ID

	if ((MidiGetField(ev, 0) != 0x00) ||	
	    (MidiGetField(ev, 1) != 0x20) ||
	    (MidiGetField(ev, 2) != 0x45)) return false;

	// Check for Chameleon Product ID

	if (MidiGetField(ev, 3) != 0x01) return false;

	// Check Device ID 

	if (!global->midiDeviceReceive(MidiGetField(ev, 4))) return false;

	// Check for Entity Application ID

	if ((MidiGetField(ev, 5) != 0x01) ||		
	    (MidiGetField(ev, 6) != ENTITY_MODEL_NUMBER)) return false;		// (see EntityVersion.h)

	// Check for valid checksum

	if (MidiGetField(ev, MidiCountFields(ev) - 1) != 0x7F) {
	
		int i;
		BYTE checksum = 0;

		for (i = 0; i < (MidiCountFields(ev) - 1); i++) checksum ^= MidiGetField(ev, i);

		checksum &= 0x7F;

		if (checksum != MidiGetField(ev, MidiCountFields(ev) - 1)) {
			
			TRACE("Bad sysex checksum\n");
			return false;
		}
	}

	// Decode Sysex Message

 	switch (MidiGetField(ev, 7)) {

		case 0x00:	soundDump(ev, LOAD_FLASH_SOUND_FROM_MIDI); break;
		case 0x01:	soundDump(ev, LOAD_EDITED_SOUND_FROM_MIDI); break;

		case 0x02:	globalDump(ev, LOAD_FLASH_GLOBAL_FROM_MIDI); break;
		case 0x03:	globalDump(ev, LOAD_EDITED_GLOBAL_FROM_MIDI); break;
				
		case 0x04:	if(MidiCountFields(ev) == 11) soundRequest(ev, REQUEST_STORED_SOUND_TO_MIDI); break;
		case 0x05:	if(MidiCountFields(ev) == 11) soundRequest(ev, REQUEST_EDITED_SOUND_TO_MIDI); break;

		case 0x06:	if(MidiCountFields(ev) == 9)  activeSoundRequest(REQUEST_STORED_ACTIVE_SOUND_TO_MIDI); break;
		case 0x07:	if(MidiCountFields(ev) == 9)  activeSoundRequest(REQUEST_EDITED_ACTIVE_SOUND_TO_MIDI); break;

		case 0x08:	if(MidiCountFields(ev) == 9)  globalRequest(REQUEST_STORED_GLOBAL_TO_MIDI); break;
		case 0x09:	if(MidiCountFields(ev) == 9)  globalRequest(REQUEST_EDITED_GLOBAL_TO_MIDI); break;

		case 0x0A:	if(MidiCountFields(ev) == 10) soundBankRequest(ev); break;
		case 0x0B:	if(MidiCountFields(ev) == 9)  allSoundsRequest(); break;
		case 0x0C:	if(MidiCountFields(ev) == 9)  allSoundsAndGlobalRequest(); break;

		default:	return false;
	}
	
	return true;
}

void EntityMidi::soundDump(MidiEvPtr ev, char updateType) {
	
	if(sysexFlashBufferFull()) return;		// Discard message if receiving process cannot keep up

	int bufSize = 0;
	
	int soundBank = MidiGetField(ev, 8);
	int userSoundNumber = MidiGetField(ev, 9);

	for(int x = 10; x < MidiCountFields(ev)-1; x++) sysexBufferIn[bufSize++] = MidiGetField(ev, x);

	TRACE("Received bank %i sound %i size %i\n", soundBank, userSoundNumber, bufSize);
		
	convert7to8bits(sysexBufferIn, bufSize);	// Convert back from 7 bit Sysex data format

	int writePos = sysexFlashBufferGetWritePos();	// This info supplied to receiver in case it missed a previous message

	for(int x = 0; x < bufSize; x++) writeSysexFlashBuffer(sysexBufferIn[x]);

	sendMessage(ctrq, SYSEX_MODE);			// Causes CTRP to switch to sysex mode
	
	mBuf[0] = updateType;				// LOAD_FLASH_SOUND_FROM_MIDI or LOAD_EDITED_SOUND_FROM_MIDI 

	mBuf[1] = (writePos >> 16) & 0xff;		// Start location in Sysex Flash Buffer from which to read
	mBuf[2] = (writePos >> 8) & 0xff;		// (this information is supplied to the receiving process 
	mBuf[3] = (writePos) & 0xff;			// in case a previous message was missed)
	
	mBuf[4] = (bufSize >> 8) & 0xff;		// Size of data
	mBuf[5] = (bufSize) & 0xff;
	
	mBuf[6] = soundBank;		
	mBuf[7] = userSoundNumber;
	
	sendMessage(ctrq, mBuf, 8);
}

void EntityMidi::globalDump(MidiEvPtr ev, char updateType) { 

	if(sysexFlashBufferFull()) return;		// Discard message if receiving process cannot keep up

	int bufSize = 0;
	
	for(int x = 8; x < MidiCountFields(ev)-1; x++) sysexBufferIn[bufSize++] = MidiGetField(ev, x);

	convert7to8bits(sysexBufferIn, bufSize);	// Convert back from 7 bit Sysex data format

	int writePos = sysexFlashBufferGetWritePos();	// This info supplied to receiver in case it missed a previous message

	for(int x = 0; x < bufSize; x++) writeSysexFlashBuffer(sysexBufferIn[x]);

	sendMessage(ctrq, SYSEX_MODE);			// Causes CTRP to switch to sysex mode
	
	mBuf[0] = updateType;				// LOAD_FLASH_GLOBAL_FROM_MIDI or LOAD_EDITED_GLOBAL_FROM_MIDI 

	mBuf[1] = (writePos >> 16) & 0xff;		// Start location in Sysex Flash Buffer from which to read
	mBuf[2] = (writePos >> 8) & 0xff;		// (this information is supplied to the receiving process 
	mBuf[3] = (writePos) & 0xff;			// in case a previous message was missed)
	
	mBuf[4] = (bufSize >> 8) & 0xff;		// Size of data
	mBuf[5] = (bufSize) & 0xff;
	
	sendMessage(ctrq, mBuf, 6);
}

void EntityMidi::soundRequest(MidiEvPtr ev, char updateType) { 

 	TRACE("Sound Request\n");

	sendMessage(ctrq, SYSEX_MODE);		

	mBuf[0] = updateType;			// REQUEST_STORED_SOUND_TO_MIDI or REQUEST_EDITED_SOUND_TO_MIDI
	mBuf[1] = MidiGetField(ev, 8);		// Sound Bank
	mBuf[2] = MidiGetField(ev, 9);		// User Sound Number
	
	sendMessage(ctrq, mBuf, 3);
}

void EntityMidi::activeSoundRequest(char updateType) { 

 	TRACE("Active sound Request\n");

	sendMessage(ctrq, SYSEX_MODE);		
	sendMessage(ctrq, updateType);		// REQUEST_STORED_ACTIVE_SOUND_TO_MIDI or REQUEST_EDITED_ACTIVE_SOUND_TO_MIDI
}

void EntityMidi::globalRequest(char updateType) { 

 	TRACE("Global Request\n");

	sendMessage(ctrq, SYSEX_MODE);		
	sendMessage(ctrq, updateType);		// REQUEST_STORED_GLOBAL_TO_MIDI or REQUEST_EDITED_GLOBAL_TO_MIDI
}

void EntityMidi::soundBankRequest(MidiEvPtr ev) { 
	
 	TRACE("Sound Bank Request\n");

	sendMessage(ctrq, SYSEX_MODE);		

	mBuf[0] = REQUEST_BANK_TO_MIDI;
	mBuf[1] = MidiGetField(ev, 8);		// Sound Bank (0..n)
	
	sendMessage(ctrq, mBuf, 2);
}

void EntityMidi::allSoundsRequest() { 
	
 	TRACE("All Sounds Request\n");

	sendMessage(ctrq, SYSEX_MODE);		
	sendMessage(ctrq, REQUEST_ALL_SOUNDS_TO_MIDI);
}

void EntityMidi::allSoundsAndGlobalRequest() { 
	
 	TRACE("All Sounds and Global Request\n");

	sendMessage(ctrq, SYSEX_MODE);		
	sendMessage(ctrq, REQUEST_ALL_TO_MIDI);
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
// 
//	Bulk Tuning Dump System Exclusive Encoding
//	------------------------------------------
//
// 	0xF0		Start of Sysex
//	0x7E		Universal non-realtime header
//	nnnn		Device ID, as midi channel (0x00..0x0F, 0x7F = broadcast)
//	0x08		Midi Tuning ID
//	0x01		Bulk dump ID
//	tttt		Tuning program number
//	|nn|		Tuning program name (16 ASCII characters)	
//	|xx yy zz|	Pitch data for 128 notes, three 7bit words for each note
//	cccc		Checksum (XOR all bytes in message from 7E to end of pitch data, bit 7 = 0)
//	0xF7		End of Sysex
//
//	Total 406 bytes (excluding F0 and F7 start and end bytes) 
//
//	Pitch data format
//	-----------------
//	
//	xx = semitone
//	yy = 1/128 semitones
//	zz = 1/16384 semitones
//
// 	Special case: 7F 7F 7F indicates no change to current tuning of note
// 
// 	Internal synth pitch format
// 	---------------------------
//
// 	0 = C5, semitone 60, units cents
//
// 	Translation from tuning data to internal synth pitch format
// 	-----------------------------------------------------------
//
// 	Tuning units are essentially semitones, synth units are essentially cents.
//
// 	Conversion from semitones to cents is x100, to cents is x400.
// 	Scaling of tuning to semitones is /16384.
// 	Offset is 60 x 400 cents
//
// 	P = (T * 400) >> 14
// 	
// 	P = P - 60 * 400
//
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 

bool EntityMidi::tuningSysex(MidiEvPtr ev) {	
	
	if(sysexFlashBufferFull()) return false;

	if (MidiCountFields(ev) != 406) return false;		// Check message size
	if (MidiGetField(ev, 0) != 0x7E) return false;		// Sysex non-realtime

	// Check for tuning bulk dump message

	if ((MidiGetField(ev, 2) != 0x08) ||			// Midi Tuning ID
	    (MidiGetField(ev, 3) != 0x01)) return false;	// Bulk dump ID

	// Check for Device ID that matches MIDI Receive channel, including broadcast

	if (!global->midiDeviceReceive(MidiGetField(ev, 4))) return false;

	// Check for valid checksum

	if (MidiGetField(ev, MidiCountFields(ev) - 1) != 0x7F) {
	
		int i;
		BYTE checksum = 0;

		for (i = 0; i < (MidiCountFields(ev) - 1); i++) checksum ^= MidiGetField(ev, i);

		checksum &= 0x7F;

		if (checksum != MidiGetField(ev, MidiCountFields(ev) - 1)) return false;
	}

	// Transmit tuning data to application 

	int writePos = sysexFlashBufferGetWritePos();

	for(int x = 4; x < 405; x++) writeSysexFlashBuffer(MidiGetField(ev, x));

	sendMessage(ctrq, SYSEX_MODE);		// Causes CTRP to switch to sysex mode	

	mBuf[0] = LOAD_TUNING_FROM_MIDI;			

	mBuf[1] = (writePos >> 16) & 0xff;	// Start location in Sysex Flash Buffer from which to read
	mBuf[2] = (writePos >> 8) & 0xff;	// (this information is supplied to the receiving process 
	mBuf[3] = (writePos) & 0xff;		// in case a previous message was missed)
	
	mBuf[4] = (384 >> 8) & 0xff;		// Size of data
	mBuf[5] = (384) & 0xff;
	
	sendMessage(ctrq, mBuf, 6);

	return true;
}

