/********************************************************************************
 * E N T I T Y 									*
 ********************************************************************************
 * File: EntityMidi.h								*
 *                                                                         	*
 * Copyright (C) A.R.Hardie-Bick 2004-2005					*
 *										*
 * This program is free software; you can redistribute it and/or modify		*
 * it under the terms of the GNU General Public License as published by		*
 * the Free Software Foundation; either version 2 of the License, or (at	*
 * your option) any later version.						*
 *										*
 * This program is distributed in the hope that it will be useful, but		*
 * WITHOUT ANY WARRANTY; without even the implied warranty of			*
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU		*
 * General Public License for more details.					*
 *										*
 * You should have received a copy of the GNU General Public License		*
 * along with this program; if not, write to the Free Software			*
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.			*
 *******************************************************************************/

#ifndef __ENTITYMIDI_H__
#define __ENTITYMIDI_H__

#include <midishare.h>

#include "EntityMessages.h"
#include "EntityProcess.h"
#include "EntityGlobal.h"
#include "EntityVersion.h"
#include "EntityDefines.h"

class EntityMidi : public EntityProcess {

protected:

	typedef 	void (EntityMidi::*Vector)(MidiEvPtr);

	// Midi event vector LUT

	Vector		event[N_MIDI_EVENT_TYPES];

	// Midi event decode vectors
	
	void		KeyOn(MidiEvPtr);
	void		KeyOff(MidiEvPtr);
	void		CtrlChange(MidiEvPtr);  
	void		ProgChange(MidiEvPtr);
	void		ChanPress(MidiEvPtr); 	 
	void		KeyPress(MidiEvPtr); 	 
	void		PitchWheel(MidiEvPtr);  
	void		SystemEx(MidiEvPtr);
	void		Reset(MidiEvPtr); 	 
	void		Null(MidiEvPtr);
	void		Clock(MidiEvPtr);
	
	// Midi controller vector LUT
	
	Vector		ctrlEvent[128];

	void		CtrlCoarse(MidiEvPtr);
	void		CtrlFine(MidiEvPtr);
	void		CtrlVolume(MidiEvPtr);
	void		CtrlPan(MidiEvPtr);
	void		CtrlExpression(MidiEvPtr);
	void		CtrlReserved(MidiEvPtr);
	void		CtrlSustain(MidiEvPtr);
	void		CtrlAllSoundOff(MidiEvPtr);
	void		CtrlAllCtrOff(MidiEvPtr);
	void		CtrlAllNotesOff(MidiEvPtr);
	void		CtrlNonRegEdit(MidiEvPtr);
	void		CtrlIncrement(MidiEvPtr);
	void		CtrlDecrement(MidiEvPtr);
	void		CtrlLFOSyncSustain(MidiEvPtr);
	void		CtrlLFOSyncPhase(MidiEvPtr);
	
	char		ctrlValue[128];
	
	// Sysex methods 
	
	bool		entitySysex(MidiEvPtr);
	bool		tuningSysex(MidiEvPtr);
	void		soundDump(MidiEvPtr, char updateType); 
	void		globalDump(MidiEvPtr, char updateType); 
	void		soundRequest(MidiEvPtr, char updateType);
	void		activeSoundRequest(char updateType);
	void		globalRequest(char updateType);
	void		soundBankRequest(MidiEvPtr); 
	void		allSoundsAndGlobalRequest();
	void		allSoundsRequest();

	// Everything else

	void			convert7to8bits(void *buffer, int &size);
	void			convert8to7bits(void *buffer, int &size);
	
	EntityGlobal		*global;

	short			midiRef;
	char			mBuf[16];

	rtems_id		synq;
	rtems_id		ctrq;
	rtems_id		midq;

	char			lastNote;
	char			lastBank;

	bool			highResNRPNPending;	// If MSB sent, use high res (literal) interpretation of Data Entry
	
	unsigned char		sysexBufferIn[ENTITY_SYSEX_MAX_MESSAGE_SIZE];	// Individual message input buffer
	unsigned char 		sysexFlashBuffer[SYSEX_FLASH_BUFFER_SIZE];	// Multiple message circular buffer
	int			sysexFlashBufferReadPos;
	int			sysexFlashBufferWritePos;
	
	bool			sysexFlashBufferFull();
	int			sysexFlashBufferGetWritePos();
	unsigned char		writeSysexFlashBuffer(unsigned char x);

	void			newSysexSoundEvent(void *buffer, int size, int soundBank, int userSoundNumber, bool stored);
	void			newSysexGlobalEvent(void *buffer, int size, bool stored);

public:

	void		receiveEvents();
	
	void		newCtrlChangeEvent(char controller, char value);
	void		newCtrl14bEvent(short controller, short value);
	void		newNonRegParamEvent(char parameter, char source, short value);
	void		newProgChangeEvent(char bank, char userSoundNumber);
	void		newKeyOnEvent(char note, char velocity);
	void		newKeyOffEvent(char note);
	void		newAllNotesOffEvent();
	
	void		newSysexStoredSoundEvent(void *buffer, int size, int soundBank, int userSoundNumber);
	void		newSysexEditedSoundEvent(void *buffer, int size, int soundBank, int userSoundNumber);

	void		newSysexStoredGlobalEvent(void *buffer, int size);
	void		newSysexEditedGlobalEvent(void *buffer, int size);

	bool 		sysexFlashBufferSetReadPos(int rp);
	unsigned char	readSysexFlashBuffer();
	
	EntityMidi(EntityGlobal*);

	void		Run();
};

#endif /* __ENTITYMIDI_H__ */
