/********************************************************************************
 * E N T I T Y 									*
 ********************************************************************************
 * File: EntityNameMode.cpp							*
 *                                                                         	*
 * Copyright (C) A.R.Hardie-Bick 2004-2005					*
 *										*
 * This program is free software; you can redistribute it and/or modify		*
 * it under the terms of the GNU General Public License as published by		*
 * the Free Software Foundation; either version 2 of the License, or (at	*
 * your option) any later version.						*
 *										*
 * This program is distributed in the hope that it will be useful, but		*
 * WITHOUT ANY WARRANTY; without even the implied warranty of			*
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU		*
 * General Public License for more details.					*
 *										*
 * You should have received a copy of the GNU General Public License		*
 * along with this program; if not, write to the Free Software			*
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.			*
 *******************************************************************************/

#include "EntityNameMode.h"

EntityNameMode::EntityNameMode(EntityPanel *panel_, EntityFlash *flash_) : EntityMode(panel_) {

	flash = flash_;

	/* Initialise symbol LUTs */

	int i = 0, t = 0;
		
	symbolStop[t++] = i;	// A..Z
	for(int x = 'A'; x <= 'Z'; x++) symbolLUT[i++] = x;

	symbolStop[t++] = i;	// a..z
	for(int x = 'a'; x <= 'z'; x++) symbolLUT[i++] = x;

	symbolStop[t++] = i;	// 0..9
	for(int x = '0'; x <= '9'; x++) symbolLUT[i++] = x;

	symbolStop[t++] = i;	// all the symbols
	for(int x = '!'; x <= '/'; x++) symbolLUT[i++] = x;
	for(int x = ':'; x <= '@'; x++) symbolLUT[i++] = x;
	for(int x = '['; x <= '`'; x++) symbolLUT[i++] = x;
	for(int x = '{'; x <= '~'; x++) symbolLUT[i++] = x;

	symbolStop[t++] = i;	// space has its own range
	symbolLUT[i++] = 32;	

	// Unfortunately the Chameleon LCD has a non-standard 
	// character set, so accented characters aren't provided.
	
	symbolStop[t++] = i;

	nSymbols = i;
	nSymbolTypes = t-1;
}

Mode EntityNameMode::enter(EntitySound *sound_, int soundNumber_, Mode oldMode_) {

	sound = sound_;
	oldMode = oldMode_;
	soundNumber = soundNumber_;

	if(soundNumber >= flash->getnWritableSounds()) return SAVE; 
	
	strcpy(soundName, sound->getName());

	panel->clearLEDs(PANEL01_LED_CTRL1, PANEL01_LED_CTRL2, PANEL01_LED_CTRL3);
	panel->setEncoderAcceleration(100);
	
	/* Find out if name is critical for bullet */

	int cp = sound->getnChangedParas();
	
	if(cp == 1 && sound->nameChanged()) nameIsCritical = true;
	else if(!cp) nameIsCritical = true;
	else nameIsCritical = false;

	/* Make sure name characters are valid */

	for(int x = 0; x < SOUND_NAME_LENGTH; x++)
		soundName[x] = soundName[x] > 126 ? 32 : soundName[x];

	for(int x = 0; x < VISIBLE_NAME_LENGTH; x++)
		soundName[x] = soundName[x] < 32 ? 32 : soundName[x];

	soundName[VISIBLE_NAME_LENGTH] = 0;	
	
	/* Display sound number & name */

	displayUpdate();

	/* Set cursor to blink. Forcing blink to false
	 * delays blanking by one cycle - used when 
	 * character is changing */

	panel->blink(CURSOR_BLINK, 200, CURSOR_BLINK_TIME);
	blink = false; 

	cursorOffset = 0; // (position offset within name)
	getSymbol();	  // Use a symbol set different from ASCII
	
	processEvents();

	sound->setName(soundName);	// always set sound name
	
	return rMode;
}

void EntityNameMode::shift() { 
	
	switch(oldMode) {
							
		case EDIT : 	rMode = SOUND;
		case SOUND : 	rMode = EDIT;
		default :	rMode = RETURN;
	}
}

void EntityNameMode::edit() { 
	
	sound->setNumber(soundNumber);
	sound->setName(soundName);

	panel->clearRepeat();
	panel->clearBlink();
					
	panel->clear();
	panel->print(1, 0, "  SAVED SOUND  ");

	flash->save(sound);
	
	rtems_task_wake_after(750);	
	panel->clear();
	
	rMode = RETURN;
}

void EntityNameMode::editHold() { }

void EntityNameMode::paramUp() { 

	panel->repeat(AUTO_PARAM_UP, 825, 120);
	moveCursor(1);
}

void EntityNameMode::paramDown() { 

	panel->repeat(AUTO_PARAM_DOWN, 825, 120);
	moveCursor(-1);
}
	
void EntityNameMode::groupUp() {

	soundName[cursorOffset] = '1';
	
	getSymbol();
	checkBullet();
	displayValueUpdate(); 
	blink = false;
}

void EntityNameMode::groupDown() {

	soundName[cursorOffset] = ' ';

	getSymbol();
	checkBullet();
	displayValueUpdate(); 
	blink = false;
}

void EntityNameMode::pageUp() {

	soundName[cursorOffset] = 'A';

	getSymbol();
	checkBullet();
	displayValueUpdate(); 
	blink = false;
}

void EntityNameMode::pageDown() {

	soundName[cursorOffset] = 'a';

	getSymbol();
	checkBullet();
	displayValueUpdate(); 
	blink = false;
}

void EntityNameMode::partUp() { 

	panel->repeat(AUTO_INSERT_CHAR, 825, 120);
	insertChar(); 
}

void EntityNameMode::partDown() { 

	panel->repeat(AUTO_DELETE_CHAR, 825, 120);
	deleteChar(); 
}

void EntityNameMode::valueUp() {

	panel->repeat(AUTO_VALUE_UP, 825, 120);
	incrementValue(+1);
}	

void EntityNameMode::valueDown() {

	panel->repeat(AUTO_VALUE_DOWN, 825, 120);
	incrementValue(-1);
}	

void EntityNameMode::autoEvent(int event) {

	switch(event) {

		case CURSOR_BLINK : 	cursorBlink(blink); break;
		case AUTO_VALUE_UP :	incrementValue(+1); break;
		case AUTO_VALUE_DOWN :	incrementValue(-1); break;
		case AUTO_PARAM_UP :	moveCursor(1); break;
		case AUTO_PARAM_DOWN :	moveCursor(-1); break;
		case AUTO_DELETE_CHAR :	deleteChar(); break;
		case AUTO_INSERT_CHAR :	insertChar(); break;
		case LED_BLINK :	blinkSaveLEDs(); break;
	}
}

void EntityNameMode::displayUpdate() {

	char 	line[PANEL_LCD_MAX_LINE_LEN + 7];

	char*		s;	// string to copy
	int 		p;	// position
		
	panel->clear();

	s = sound->getNumberStr(soundNumber);
	p = 0; while(*s) line[p++] = *s++;

	line[p++] = sound->getBullet();

	namePosition = p;

	s = soundName;

	for(int x = 0; x < VISIBLE_NAME_LENGTH; x++)
		line[p++] = *s++;
	
	line[p++] = 0;

	panel->print(0, 0, line);
}

void EntityNameMode::displayValueUpdate() {

	char line[2]; 
	
	line[0] = soundName[cursorOffset];
	line[1] = 0;

	/* slightly delay start of next cursor blink */
	panel->blink(CURSOR_BLINK, 300, CURSOR_BLINK_TIME); blink = false;
	
	panel->print(0, namePosition + cursorOffset, line);
}

bool EntityNameMode::moveCursor(int i) {

	int np = cursorOffset + i;
	
	if(np < 0) np = 0;
	else if(np >= VISIBLE_NAME_LENGTH) np = VISIBLE_NAME_LENGTH - 1;

	if(cursorOffset == np) return false;
	
	displayValueUpdate();	// Draw character at old pos
	cursorOffset = np;	// new pos
	getSymbol();		// 

	blink = true; 		// cursor initially visible
	cursorBlink(blink); 	// draw cursor

	return true;
}

void EntityNameMode::getSymbol() {

	char c = soundName[cursorOffset];

	for(int x = 0; x < nSymbolTypes; x++) {
		for(int y = symbolStop[x]; y < symbolStop[x+1]; y++) {
			if(c == symbolLUT[y]) {
				symbolType = x;
				symbol = y;
				return;
			}
		}
	}
}

Status EntityNameMode::incrementValue(int i) {

	int ns = symbol + i;
	static bool stopped = true;

	if(i > 0) {

		stopped = false;
		
		if(ns >= symbolStop[symbolType+1]) {
	
			symbolType++;
			
			if(symbolType >= nSymbolTypes) {
				symbolType = 0;
				ns = 0;
			}
			
			else ns = symbolStop[symbolType];		
			stopped = true;
		}
	}

	else {

		if(!stopped) {
		
			if(ns <= symbolStop[symbolType]) {
					
				ns = symbolStop[symbolType];
				stopped = true;
			}
		}

		else {

			stopped = false;
			
			symbolType--;
			
			if(symbolType < 0) {

				symbolType = nSymbolTypes - 1;
				ns += nSymbols;
			}
			
			if(ns <= symbolStop[symbolType]) {
	
				ns = symbolStop[symbolType];
				stopped = true;
			}
		}
	}

	if(ns == symbol) return PARAMETER_UNCHANGED;

	soundName[cursorOffset] = symbolLUT[symbol = ns];
	
	checkBullet();
	displayValueUpdate(); 

	blink = false;

	if(stopped) return PARAMETER_THRESHOLD;

	return PARAMETER_CHANGED;
}

void EntityNameMode::checkBullet() {

	if(nameIsCritical) {
		
		sound->setName(soundName);

		char line[2]; 
	
		line[0] = sound->getBullet();
		line[1] = 0;

		panel->print(0, namePosition - 1, line);
	}
}

void EntityNameMode::cursorBlink(bool &blink) {

	char line[2];

	line[0] = blink ? '_' : soundName[cursorOffset];
	line[1] = 0; blink = !blink;
	
	panel->print(0, namePosition + cursorOffset, line);
}

void EntityNameMode::blinkSaveLEDs() {
	
	if(ledBlink) {
				
		ledBlink = false;
		panel->setLEDs(PANEL01_LED_EDIT);
		panel->clearLEDs(PANEL01_LED_SHIFT);
	}
						
	else {
		ledBlink = true;
		panel->setLEDs(PANEL01_LED_SHIFT);
		panel->clearLEDs(PANEL01_LED_EDIT);
	}
}
						
void EntityNameMode::deleteChar() {

	char *s = &soundName[cursorOffset+1];
	char *d = &soundName[cursorOffset];

	for(int x = cursorOffset; x < VISIBLE_NAME_LENGTH - 1; x++) {

		*d++ = *s++;
	}

	*d++ = ' ';
	*d++ = 0;

	getSymbol();	
	checkBullet();		
	panel->print(0, namePosition, soundName);
}

void EntityNameMode::insertChar() {

	char *s = &soundName[VISIBLE_NAME_LENGTH-1];
	char *d = &soundName[VISIBLE_NAME_LENGTH];

	*s = 0;
	
	for(int x = VISIBLE_NAME_LENGTH; x > cursorOffset; x--) {

		*d-- = *s--;
	}

	*d = ' ';

	getSymbol();
	checkBullet();		
	panel->print(0, namePosition, soundName);
}

