/********************************************************************************
 * E N T I T Y 									*
 ********************************************************************************
 * File: EntityNameMode.h							*
 *                                                                         	*
 * Copyright (C) A.R.Hardie-Bick 2004-2005					*
 *										*
 * This program is free software; you can redistribute it and/or modify		*
 * it under the terms of the GNU General Public License as published by		*
 * the Free Software Foundation; either version 2 of the License, or (at	*
 * your option) any later version.						*
 *										*
 * This program is distributed in the hope that it will be useful, but		*
 * WITHOUT ANY WARRANTY; without even the implied warranty of			*
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU		*
 * General Public License for more details.					*
 *										*
 * You should have received a copy of the GNU General Public License		*
 * along with this program; if not, write to the Free Software			*
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.			*
 *******************************************************************************/

#ifndef __ENTITYNAMEMODE_H__
#define __ENTITYNAMEMODE_H__

#include <rtems.h>
#include <chameleon.h>
#include <string.h>

#include "EntitySound.h"
#include "EntityFlash.h"
#include "EntityMode.h"

#define NAME_SAVE_HOLD_DELAY 200	// Milliseconds before save countdown
#define NAME_SAVE_HOLD_TIME 1000	// Milliseconds to hold down save button
#define VISIBLE_NAME_LENGTH 12		// Number of name characters shown
#define CURSOR_BLINK_TIME 250		// Time for half a blink cycle

class EntityNameMode : public EntityMode {
			
	EntitySound*	sound;
	EntityFlash*	flash;

	Mode		oldMode;
	int		soundNumber;
	char		soundName[SOUND_NAME_LENGTH];

	char		symbolLUT[256];
	char		symbolStop[10];
	int		nSymbols;
	char		symbol;
	int		symbolType;
	int		nSymbolTypes;
	int		cursorOffset;		// cursor offset within name
	int		namePosition;		// position of 1st char of name on display
	bool		blink;			// cursor state
	bool		nameIsCritical;		// true if name is the only changing para in sound

	void		displayUpdate();
	void		displayValueUpdate();
	Status		incrementValue(int);
	
	void		blinkSaveLEDs();

	bool		moveCursor(int);
	void		incrementChar(int);
	void		cursorBlink(bool&);
	void		deleteChar();
	void		insertChar();
	void		checkBullet();
	void		getSymbol();

	void		autoEvent(int);

	void		edit();
	void		editHold();
	void		shift();
	void		partUp();
	void		partDown();
	void		groupUp();
	void		groupDown();
	void		pageUp();
	void		pageDown();
	void		paramUp();
	void		paramDown();
	void		valueUp();
	void		valueDown();

public:

	Mode		enter(EntitySound*, int, Mode);

	EntityNameMode(EntityPanel*, EntityFlash*);
};

#endif /* __ENTITYNAMEMODE_H__ */

























