/********************************************************************************
 * E N T I T Y 									*
 ********************************************************************************
 * File: EntityPanel.cpp							*
 *                                                                         	*
 * Copyright (C) A.R.Hardie-Bick 2004-2005					*
 *										*
 * This program is free software; you can redistribute it and/or modify		*
 * it under the terms of the GNU General Public License as published by		*
 * the Free Software Foundation; either version 2 of the License, or (at	*
 * your option) any later version.						*
 *										*
 * This program is distributed in the hope that it will be useful, but		*
 * WITHOUT ANY WARRANTY; without even the implied warranty of			*
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU		*
 * General Public License for more details.					*
 *										*
 * You should have received a copy of the GNU General Public License		*
 * along with this program; if not, write to the Free Software			*
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.			*
 *******************************************************************************/
  
#include "EntityPanel.h"

//#define DEBUG_PANEL

const unsigned int keyArray [] = { 
	
		PANEL01_KEY_EDIT, 	PANEL01_KEY_SHIFT, 	PANEL01_KEY_PART_UP, 	
		PANEL01_KEY_PART_DOWN,	PANEL01_KEY_GROUP_UP, 	PANEL01_KEY_GROUP_DOWN,	
		PANEL01_KEY_PAGE_UP, 	PANEL01_KEY_PAGE_DOWN,	PANEL01_KEY_PARAM_UP, 	
		PANEL01_KEY_PARAM_DOWN,	PANEL01_KEY_VALUE_UP,	PANEL01_KEY_VALUE_DOWN 
	};

const int N_PANEL_KEYS = sizeof(keyArray) / sizeof(keyArray[0]);

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	
void EntityPanel::Run() {
	
	char	m[7];

	rtems_clock_get(RTEMS_CLOCK_GET_TICKS_SINCE_BOOT, &ticks);
	oldTicks = ticks;
	
	// - - - - - - - Encoder Debounce - - - - - - - 
	
	dbOldTicks = ticks;
	dbcps = 0; dbEncoder = 0; dbSign = 0;

	// - - - - - - - - - - - - - - - - - - - - - - -
	
	/* Repeater tasks for writing to queues */

	repeater1->create("CTRQ", "REP1", 100);
	repeater2->create("CTRQ", "REP2", 100);
	repeater3->create("CTRQ", "REP3", 100);
	repeater4->create("SYNQ", "REP4", 100);

	while(1) {	
	
		panel_in_new_event(panelHandle, TRUE);

		if(panel_in_encoder(panelHandle, &eidx, &encoder)) {

			rtems_clock_get(RTEMS_CLOCK_GET_TICKS_SINCE_BOOT, &ticks);

			// - - - - - - - - - - - - - - - Encoder debounce - - - - - - - - - - - - - - - - - - - 
	
			/* Calculate encoder velocity */

			dbTimeInterval = ticks - dbOldTicks;
			dbOldTicks = ticks;

			dbcps = ((int)dbEncoder << 10) / (int)dbTimeInterval; // ~= enc*1000 / t-in-ms

			/* This code aims to fix encoder contact bounce of either prevailing polarity. This was an 
			 * anomalous problem found only on the Chameleon on which Entity was originally developed.
			 * It's left in as it doesn't seem to noticably slow things down, and may be useful. */
			
			if(encoder > 4 | encoder < -4) encoder = dbEncoder; // If anomalous magnitude, use previous data 
			else {						   // else check for sign reversal

				/* If three same sign raw encoder values in succession, disable opposite polarity fix */
				if(encoder > 0) dbSign++; else dbSign--; dbSign = dbSign > 2 ? 2 : dbSign < -2 ? -2 : dbSign;
			
				/* If sudden sign change to positive, use previous data */
				if(dbcps < -20) if(encoder > 0 & dbSign < 2) encoder = dbEncoder;

				/* If sudden sign change to negative, use previous data */
				else if(dbcps > 20) if(encoder < 0 & dbSign > -2) encoder = dbEncoder;
			}

			/* This should work for string of errors */
			dbEncoder = encoder;

			// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -  
	
			/* Encoder used as modulation source */

			if(encoderMod) {
			
				m[0] = PANEL_ENCODER_MOD;
				m[1] = (char) eidx;
				m[2] = (char) encoder;

				status = rtems_message_queue_send(synq, m, 3);

				#ifdef DEBUG_PANEL
				if(status != RTEMS_SUCCESSFUL) TRACE("PANP -> SYNQ problem (encoder)\n");
				#endif
			}

			/* Encoder used as data source */
		
			else {
			
				encoderAccumulator += encoder;
				
				if(int encoderDiv4 = encoderAccumulator>>2) {
			
					timeInterval = ticks - oldTicks;
					oldTicks = ticks;

					if(timeInterval) {
			
						int absEnc = encoderDiv4 < 0 ? ~encoderDiv4 + 1 : encoderDiv4;
					
						// acceleration is a percentage, ie, an int in range 0..100
						int cps = (absEnc * acceleration * 10) / timeInterval;	// Counts-per-second

						/* Filter the (rather coarse) velocity */
				
						static int fcps = 0;	// Filtered counts-per-second
				
						/* Calculate filter coefficient (pragmatic) */
				
						int offset = 25; // Increasing this reduces filtering of fast rotation (25)
						
						int fci = (timeInterval + offset) > 256 ? 256 : timeInterval + offset;
						int fc = 256 - fci;
				
						fcps = (fcps * fc + cps * fci) >> 8;

						/* Sigmoid velocity scaling */

						int sAddr = fcps > sigmoidSize ? sigmoidSize : fcps;
						encoderDiv4 += sigmoid[sAddr] * encoderDiv4;
				
						/* Transmit encoder data */
				
						encoderAccumulator &= 3;

						m[0] = PANEL_ENCODER_DATA;
						m[1] = (char) eidx;
						m[2] = (char) encoderDiv4;
				
						status = rtems_message_queue_send(ctrq, m, 3);

						#ifdef DEBUG_PANEL
						if(status != RTEMS_SUCCESSFUL) TRACE("PANP -> CTRQ problem (encoder)\n");
						#endif
					}
				}					
			}
		}
	
		if(panel_in_keypad(panelHandle, &keyBits)) {

			/* The next line seems to fix a very strange bug, causing false on/off event pairs
			 * for keys with low order bits common to the true key event. This may be related
			 * to button-state transfers from the PIC being perfomed non-atomically (the PIC 
			 * is eight bits), and possibly this has emerged because I have messed with the 
			 * TFLA priority, or set up a priority scenario that causes problems either for 
			 * PIC data transfer, or for key bit pattern assembly (by byte shifting) in the 
			 * Coldfire. (NB This fault was duplicated on a second Chameleon.) */
			
			/* Fix false key-on event by receiving corresponding immediate false key-off */ 
			if(panel_in_new_event(panelHandle, false)) panel_in_keypad(panelHandle, &keyBits);

			/* Business as usual */
			
			changedKeyBits = keyBits ^ oldKeyBits; 
			
			oldKeyBits = keyBits;

			for(int k = 0; k < N_PANEL_KEYS; k++) {		// Scan all keys

				unsigned int key = keyArray[k];

				if(changedKeyBits & key) {		// If key has changed state
					
					m[0] = PANEL_KEYPAD;
					
					if(keyBits & key) {		// If key is now on...
						
						m[1] = (char) KEY_ON;
				
					} else {
						
						m[1] = (char) KEY_OFF;
					}
				
					m[2] = (char) (key >> 24);	// Define which key changed
					m[3] = (char) (key >> 16);	// (send only ms 16 bits)

					m[4] = (char) (keyBits >> 24);	// Define key states
					m[5] = (char) (keyBits >> 16);

					status = rtems_message_queue_send(ctrq, m, 6);

					#ifdef DEBUG_PANEL
					if(status != RTEMS_SUCCESSFUL) TRACE("PANP -> CTRQ problem (keypad)\n");
					#endif
				}

			}	// (continue scanning in case of multiple simultaneous key events)
		}
		
		if(panel_in_potentiometer(panelHandle, &pidx, &potValue)) {

			potVal[pidx] = potValue;

			m[0] = PANEL_POTENTIOMETER;
			m[1] = (char) pidx;
			m[2] = (char) potValue;

			if(!potMod[pidx]) {
				
				status = rtems_message_queue_send(ctrq, m, 3);
			}

			else {
				if(pidx == PANEL01_POT_CTRL3) m[1] = pot3Assignment;
				status = rtems_message_queue_send(synq, m, 3);
			}

			#ifdef DEBUG_PANEL
			if(status != RTEMS_SUCCESSFUL) TRACE("PANP messaging problem (potentiometer)\n");
			#endif
		}
	}
}

bool EntityPanel::clear() { 
	
	return panel_out_lcd_clear(panelHandle); 
}

bool EntityPanel::clearLine(int line) { 
	
	return panel_out_lcd_print(panelHandle, line, 0, "                "); 
}

bool EntityPanel::print(int y, int x, char *str) { 
	
	return panel_out_lcd_print(panelHandle, y, x, str); 
}

bool EntityPanel::clearLEDs(int a, int b, int c, int d, int e) {

	ledStates &= ~a & ~b & ~c & ~d & ~e;
	return panel_out_led(panelHandle, ledStates);
}

bool EntityPanel::setLEDs(int a, int b, int c, int d, int e) {

	ledStates |= a | b | c | d | e;
	return panel_out_led(panelHandle, ledStates);
}

void EntityPanel::setPotsForModulation(int a, int b, int c, int d) {

	if(a > -1) setPotForModulation(a);
	if(b > -1) setPotForModulation(b);
	if(c > -1) setPotForModulation(c);
	if(d > -1) setPotForModulation(d);
}

void EntityPanel::setPotsForModulation() {

	setPotForModulation(PANEL01_POT_VOLUME);
	setPotForModulation(PANEL01_POT_CTRL1);
	setPotForModulation(PANEL01_POT_CTRL2);
	setPotForModulation(PANEL01_POT_CTRL3);
}

void EntityPanel::setPotsForData(int a, int b, int c, int d) {

	if(a > -1) setPotForData(a);
	if(b > -1) setPotForData(b);
	if(c > -1) setPotForData(c);
	if(d > -1) setPotForData(d);
}

void EntityPanel::setPotsForData() {

	setPotForData(PANEL01_POT_VOLUME);
	setPotForData(PANEL01_POT_CTRL1);
	setPotForData(PANEL01_POT_CTRL2);
	setPotForData(PANEL01_POT_CTRL3);
}

void EntityPanel::setPotForModulation(int p) {

	switch(p) {

		case PANEL01_POT_CTRL1 : clearLEDs(PANEL01_LED_CTRL1); break;
		case PANEL01_POT_CTRL2 : clearLEDs(PANEL01_LED_CTRL2); break; 
		case PANEL01_POT_CTRL3 : clearLEDs(PANEL01_LED_CTRL3); break;
	}

	potMod[p] = true;	
}

void EntityPanel::setPotForData(int p) {

	switch(p) {

		case PANEL01_POT_CTRL1 : setLEDs(PANEL01_LED_CTRL1); break;
		case PANEL01_POT_CTRL2 : setLEDs(PANEL01_LED_CTRL2); break; 
		case PANEL01_POT_CTRL3 : setLEDs(PANEL01_LED_CTRL3); break;
	}

	potMod[p] = false;	
}

void EntityPanel::transmitPot3EventsAs(char panelPot) {

	pot3Assignment = panelPot;
}

void EntityPanel::requestPotUpdates() {

	if(!ctrq) { TRACE("ERROR: ctrq not initialised\n"); return; }
	
	static char m[3];

	for(int x = 0; x < 4; x++) {

		if(!potMod[x]) {

			m[0] = PANEL_POTENTIOMETER;
			m[1] = (char) x;
			m[2] = (char) potVal[x];

			rtems_message_queue_send(ctrq, m, 3);
		}
	}
}

void EntityPanel::requestPot1Update() {

	if(potMod[1]) return;
	
	if(!ctrq) { TRACE("ERROR: ctrq not initialised\n"); return; }
	
	static char m[3];

	m[0] = PANEL_POTENTIOMETER;
	m[1] = (char) 1;
	m[2] = (char) potVal[1];

	rtems_message_queue_send(ctrq, m, 3);
}

bool EntityPanel::repeat(char message, int delay, int repeat) {

	return repeater1->activate(message, delay, repeat);
}

bool EntityPanel::clearRepeat() { 
	
	return repeater1->clear(); 
}

bool EntityPanel::delay(char message, int delay, int repeat) {

	return repeater2->activate(message, delay, repeat);
}

bool EntityPanel::clearDelay() { 
	
	return repeater2->clear(); 
}

bool EntityPanel::blink(char message, int delay, int repeat) {

	return repeater3->activate(message, delay, repeat);
}

bool EntityPanel::clearBlink() { 
	
	return repeater3->clear(); 
}

bool EntityPanel::request(char message, int delay, int repeat) {

	return repeater4->activate(message, delay, repeat);
}

bool EntityPanel::clearRequest() { 
	
	return repeater4->clear(); 
}

void EntityPanel::setEncoderAcceleration(int percent) {

	if(percent > 100) percent = 100;
	else if(percent < 0) percent = 0;

	acceleration = percent;
}

void EntityPanel::resetEncoder() { encoder = 0; }

void EntityPanel::setEncoderForData() { encoderMod = false; }
void EntityPanel::setEncoderForModulation() { encoderMod = true; }

void EntityPanel::clearAllRepeaters() {

	clearRequest();
	clearRepeat(); 
	clearBlink();
	clearDelay();
}

EntityPanel::EntityPanel() {

	/* Queue on which main event messages will be placed */
	ctrq = getQueueId("CTRQ");
	
	/* Queue for Real Time Control */
	synq = getQueueId("SYNQ");
	
	if(!(panelHandle = panel_init())) { 
		
		TRACE("ERROR: Could not obtain handle in EntityPanel\n"); 
		return; 
	}

	/* Prepare sigmoid for encoder velocity scaling */

	const int 	threshold = 40;		// cps level at which scaling saturates (40)
	const int 	factor = 20;		// maximum encoder scaling factor (20)
	
	sigmoidSize = 128;			// max cps (typical max = ~50 - 70)

	sigmoid = new int [sigmoidSize];

	for(int cps = 0; cps < sigmoidSize; cps++) {	// counts-per-second

		float I = (float) cps / (float) threshold;
		if(I > 1.0) I = 1.0;
		
		float Sigma = 0.5 + sin(I * M_PI - M_PI / 2.0 ) / 2.0;	// 'S' -  x:0..1, y:0..1

		sigmoid[cps] = (int)(Sigma * (float) factor);
	}

	static rtems_unsigned8 matrix_route_lcd_symbol [8] =

		{
			0x00,	/* 00000 */
			0x04,	/* 00100 */
			0x02,	/* 00010 */
			0x1f,	/* 11111 */
			0x02,	/* 00010 */
			0x04,	/* 00100 */
			0x00,	/* 00000 */
			0x00	/* 00000 */
		};
		
	static rtems_unsigned8 bullet_lcd_symbol [8] =

		{
			0x00,	/* 00000 */
			0x00,	/* 00000 */
			0x00,	/* 00000 */
			0x00,	/* 00000 */
			0x00,	/* 00000 */
			0x00,	/* 00000 */
			0x04,	/* 00100 */
			0x00	/* 00000 */
		};
		
	static rtems_unsigned8 level_l_lcd_symbol [8] =

		{
			0x00,	/* 00000 */
			0x00,	/* 00000 */
			0x1f,	/* 11111 */
			0x00,	/* 00000 */
			0x00,	/* 00000 */
			0x00,	/* 00000 */
			0x00,	/* 00000 */
			0x00	/* 00000 */
		};
		
	static rtems_unsigned8 level_r_lcd_symbol [8] =

		{
			0x00,	/* 00000 */
			0x00,	/* 00000 */
			0x00,	/* 00000 */
			0x00,	/* 00000 */
			0x1f,	/* 11111 */
			0x00,	/* 00000 */
			0x00,	/* 00000 */
			0x00	/* 00000 */
		};
		
	static rtems_unsigned8 level_lm_lcd_symbol [8] =

		{
			0x04,	/* 00100 */
			0x04,	/* 00100 */
			0x1f,	/* 11111 */
			0x04,	/* 00100 */
			0x04,	/* 00100 */
			0x00,	/* 00000 */
			0x00,	/* 00000 */
			0x00	/* 00000 */
		};
		
	static rtems_unsigned8 level_rm_lcd_symbol [8] =

		{
			0x00,	/* 00000 */
			0x00,	/* 00000 */
			0x04,	/* 00100 */
			0x04,	/* 00100 */
			0x1f,	/* 11111 */
			0x04,	/* 00100 */
			0x04,	/* 00100 */
			0x00	/* 00000 */
		};
		
	static rtems_unsigned8 stack_minus_lcd_symbol [8] =

		{
			0x00,	/* 00000 */
			0x00,	/* 00000 */
			0x00,	/* 00000 */
			0x07,	/* 00111 */
			0x00,	/* 00000 */
			0x00,	/* 00000 */
			0x00,	/* 00000 */
			0x00	/* 00000 */
		};
		
	panel_out_lcd_redefine(panelHandle, 1, matrix_route_lcd_symbol);
	panel_out_lcd_redefine(panelHandle, 2, bullet_lcd_symbol);
	panel_out_lcd_redefine(panelHandle, 3, level_l_lcd_symbol);
	panel_out_lcd_redefine(panelHandle, 4, level_r_lcd_symbol);
	panel_out_lcd_redefine(panelHandle, 5, level_lm_lcd_symbol);
	panel_out_lcd_redefine(panelHandle, 6, level_rm_lcd_symbol);
	panel_out_lcd_redefine(panelHandle, 7, stack_minus_lcd_symbol);

	/* Other initialisations */

	repeater1 = new EntityRepeat;		// ie for repeating increments with button
	repeater2 = new EntityRepeat;		// ie for cursor flash, or LED blink
	repeater3 = new EntityRepeat;		// ie for delaying display update after warning
	repeater4 = new EntityRepeat;		// ie for requesting meter readings from synth

	ledStates = 0;
	panel_out_led(panelHandle, ledStates);
	
	keyBits = 0;
	oldKeyBits = 0;

	acceleration = 100;	// default 100%
	encoderAccumulator = 0;	
	
	for(int x = 0; x < 4; x++) potVal[x] = 64;
	setPotsForModulation();

	setEncoderForData();
	
	transmitPot3EventsAs(PANEL01_POT_CTRL3);
}

EntityPanel::~EntityPanel() {

	delete [] sigmoid;

	delete repeater1;
	delete repeater2;
	delete repeater3;
	delete repeater4;
}

