/********************************************************************************
 * E N T I T Y 									*
 ********************************************************************************
 * File: EntityPresets.cpp							*
 *                                                                         	*
 * Copyright (C) A.R.Hardie-Bick 2005 		          			*
 *										*
 * This program is free software; you can redistribute it and/or modify		*
 * it under the terms of the GNU General Public License as published by		*
 * the Free Software Foundation; either version 2 of the License, or (at	*
 * your option) any later version.						*
 *										*
 * This program is distributed in the hope that it will be useful, but		*
 * WITHOUT ANY WARRANTY; without even the implied warranty of			*
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU		*
 * General Public License for more details.					*
 *										*
 * You should have received a copy of the GNU General Public License		*
 * along with this program; if not, write to the Free Software			*
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.			*
 ********************************************************************************/

#include "EntityPresets.h"
#include "EntityPresetData.h"

EntityPresets::EntityPresets() { 

	nPresets = 0;
	int addr = 0;
	int offset = 1;	
	
	while(offset) {
		
		offset = 256 * (unsigned char) entityPresetData[addr++];
		offset += (unsigned char) entityPresetData[addr++];

		if(offset) {

			index[nPresets++] = addr;
			addr += offset;
			
			if(nPresets >= MAX_N_PRESETS) TRACE("ERROR: Too many presets\n");
		}
	}
}

char* EntityPresets::snoopName(int presetNumber) {

	if(presetNumber >= nPresets) return 0;
	return (char*) &entityPresetData[index[presetNumber]];
}

bool EntityPresets::load(EntitySound *snd, int soundNumber, int presetNumber) {

	snd->deCompress((void*)&entityPresetData[index[presetNumber]]);
	snd->setNumber(soundNumber);

	return true;
}

int EntityPresets::getnPresets() {

	return nPresets;
}

bool EntityPresets::loadDefault(EntitySound *snd, int soundNumber) {

	if(!(entityDefaultData[0] || entityDefaultData[1])) return false;

	snd->deCompress((void*)&entityDefaultData[2]);
	snd->setNumber(soundNumber);

	return true;
}

int EntityPresets::load(void *buffer, int presetNumber) {

	unsigned char *dst = (unsigned char*) buffer;
	
	if(presetNumber >= nPresets) return loadDefault(buffer);

	unsigned char *src = (unsigned char*) &entityPresetData[index[presetNumber]-2];
	
	int length = 256 * *src++;
	length += *src++;

	for(int x = 0; x < length; x++) *dst++ = *src++;

	return length;
}

int EntityPresets::loadDefault(void *buffer) {

	unsigned char *dst = (unsigned char*) buffer;
	unsigned char *src = (unsigned char*) entityDefaultData;
	
	int length = 256 * *src++;
	length += *src++;

	for(int x = 0; x < length; x++) *dst++ = *src++;

	return length;
}

