/********************************************************************************
 * E N T I T Y 									*
 ********************************************************************************
 * File: EntityRepeat.cpp							*
 *                                                                         	*
 * Copyright (C) A.R.Hardie-Bick 2004-2005					*
 *										*
 * This program is free software; you can redistribute it and/or modify		*
 * it under the terms of the GNU General Public License as published by		*
 * the Free Software Foundation; either version 2 of the License, or (at	*
 * your option) any later version.						*
 *										*
 * This program is distributed in the hope that it will be useful, but		*
 * WITHOUT ANY WARRANTY; without even the implied warranty of			*
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU		*
 * General Public License for more details.					*
 *										*
 * You should have received a copy of the GNU General Public License		*
 * along with this program; if not, write to the Free Software			*
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.			*
 *******************************************************************************/
  
#include "EntityRepeat.h"

void EntityRepeat::Run() {

	activated = true;

	rtems_task_wake_after(delayTime);

	while(1) {
			
		if(message != -1) {
			
			status = rtems_message_queue_send(destination_queue_id, &message, 1);
			
			if(status != RTEMS_SUCCESSFUL) {
				
				TRACE("EntityRepeat object cannot write to a queue:\n");

				switch(status) {

					case RTEMS_TOO_MANY: 		TRACE("Queue's limit has been reached\n"); break;
					case RTEMS_UNSATISFIED:		TRACE("Out of message buffers\n"); break;
					case RTEMS_INVALID_SIZE: 	TRACE("Invalid message size\n"); break;
					case RTEMS_INVALID_ID: 		TRACE("Invalid queue ID\n"); break;

					default:			TRACE("Unknown queue error\n"); break;
				}

				TRACE("In EntityRepeat process %s writing to %s:\n", pName, qName);  
			}
		}

		if(!repeatTime) {
			
			activated = false;
			rtems_task_suspend(RTEMS_SELF);
		}
		
		else rtems_task_wake_after(repeatTime);
	}
}

void EntityRepeat::wait(int t) {

	activate(-1, t, 0);
}

bool EntityRepeat::clear() {

	return rtems_task_suspend(task_id);
}

bool EntityRepeat::activate(char m, int d, int r) {

	message = m;
	delayTime = d;
	repeatTime = r;

	// Activate regardless of whether previous cycle completed
	status = rtems_task_restart(task_id, (rtems_task_argument) this);

	if (status != RTEMS_SUCCESSFUL) {
		TRACE("ERROR: cannot restart EntityRepeat task.\n");
		return false;
	}
	
	return true;
}

bool EntityRepeat::running() {

	return activated;
}

bool EntityRepeat::create(	char* 			destQueue, 
				char* 			tname, 
				rtems_task_priority 	priority ) {
				
	// For debugging the process and queue names are useful

	for(int x = 0; x < 4; x++) pName[x] = tname[x]; pName[4] = 0;
	for(int x = 0; x < 4; x++) qName[x] = destQueue[x]; qName[4] = 0;
	
	message = -1;
	delayTime = 0;
	repeatTime = 0;
	activated = false;

	destination_queue_id = getQueueId(destQueue);
	
	return this->EntityProcess::create(tname, priority);	
}

