/********************************************************************************
 * E N T I T Y 									*
 ********************************************************************************
 * File: EntitySaveGlobalMode.cpp						*
 *                                                                         	*
 * Copyright (C) A.R.Hardie-Bick 2004-2005					*
 *										*
 * This program is free software; you can redistribute it and/or modify		*
 * it under the terms of the GNU General Public License as published by		*
 * the Free Software Foundation; either version 2 of the License, or (at	*
 * your option) any later version.						*
 *										*
 * This program is distributed in the hope that it will be useful, but		*
 * WITHOUT ANY WARRANTY; without even the implied warranty of			*
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU		*
 * General Public License for more details.					*
 *										*
 * You should have received a copy of the GNU General Public License		*
 * along with this program; if not, write to the Free Software			*
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.			*
 *******************************************************************************/

#include "EntitySaveGlobalMode.h"

EntitySaveGlobalMode::EntitySaveGlobalMode(EntityPanel *panel_, EntityFlash *flash_, EntityGlobal *global_) : EntityMode(panel_) {

	global = global_;
	flash = flash_;
	panel = panel_;
}

Mode EntitySaveGlobalMode::enter() {

	panel->clearAllRepeaters();
	
	panel->clearLEDs(PANEL01_LED_EDIT, PANEL01_LED_CTRL1, PANEL01_LED_CTRL2, PANEL01_LED_CTRL3);
	panel->setLEDs(PANEL01_LED_SHIFT);
	
	panel->blink(LED_BLINK, 0, SAVE_GLOBAL_LED_BLINK_TIME);
	
	ledBlink = true;
	
	displayUpdate();
	
	processEvents(); 

	panel->clearLEDs(PANEL01_LED_SHIFT);

	// If saved, go back to previous mode (before 
	// global) otherwise go back to global

	if(rMode == RETURN) return GLOBAL;
	return RETURN;
}

void EntitySaveGlobalMode::edit() { 

	panel->clear();
	
	panel->print(1, 0, " SAVING CONFIG  ");
	
	flash->save(global);
	
	rtems_task_wake_after(750);
	panel->clear();
	
	rMode = SOUND;
}	

void EntitySaveGlobalMode::editHold() { }				

void EntitySaveGlobalMode::shift() { 

	rMode = RETURN;
}

void EntitySaveGlobalMode::autoEvent(int event) { 

	switch(event) {

		case LED_BLINK :

			blinkSaveLEDs();
			
			break;		
	}
}

void EntitySaveGlobalMode::blinkSaveLEDs() {

	if(ledBlink) {
		
		ledBlink = false;
		panel->setLEDs(PANEL01_LED_EDIT);
		panel->clearLEDs(PANEL01_LED_SHIFT);
	}
				
	else {
		ledBlink = true;
		panel->setLEDs(PANEL01_LED_SHIFT);
		panel->clearLEDs(PANEL01_LED_EDIT);
	}
}

void EntitySaveGlobalMode::displayValueUpdate() { }

void EntitySaveGlobalMode::displayUpdate() {

	panel->clear();
	panel->print(0, 0, "  SAVE CONFIG   ");
}

