/********************************************************************************
 * E N T I T Y 									*
 ********************************************************************************
 * File: EntitySaveMode.cpp							*
 *                                                                         	*
 * Copyright (C) A.R.Hardie-Bick 2004-2005					*
 *										*
 * This program is free software; you can redistribute it and/or modify		*
 * it under the terms of the GNU General Public License as published by		*
 * the Free Software Foundation; either version 2 of the License, or (at	*
 * your option) any later version.						*
 *										*
 * This program is distributed in the hope that it will be useful, but		*
 * WITHOUT ANY WARRANTY; without even the implied warranty of			*
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU		*
 * General Public License for more details.					*
 *										*
 * You should have received a copy of the GNU General Public License		*
 * along with this program; if not, write to the Free Software			*
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.			*
 *******************************************************************************/

#include "EntitySaveMode.h"

EntitySaveMode::EntitySaveMode(EntityPanel *panel_, EntityFlash *flash_) : EntityMode(panel_) {

	flash = flash_;
	nameMode = new EntityNameMode(panel_, flash_);
	
	ledBlinker = new EntityRepeat;
	ledBlinker->create("CTRQ", "LDBP", 100);
}

EntitySaveMode::~EntitySaveMode() {

	delete nameMode;
}

Mode EntitySaveMode::enter(EntitySound *sound_, Mode oldMode_) {

	sound = sound_;
	oldMode = oldMode_;

	soundNumber = sound->getNumber();

	panel->clearLEDs(PANEL01_LED_EDIT, PANEL01_LED_CTRL1, PANEL01_LED_CTRL2, PANEL01_LED_CTRL3);
	panel->setLEDs(PANEL01_LED_SHIFT);
	
	ledBlinker->activate(LED_BLINK, 0, SAVE_LED_BLINK_TIME);
	ledBlink = true;
	
	panel->setEncoderAcceleration(0);

	rMode = SAVE;

	while(1) {
	
		switch(rMode) {

			case SAVE :	
				
				processEvents(); 
				break;
				
			case NAME :	
				
				rMode = nameMode->enter(sound, soundNumber, oldMode); 
				break;
				
			default	  :	

				panel->clearLEDs(PANEL01_LED_SHIFT);
				ledBlinker->clear();				
				return rMode;
		}
	}

	return rMode;
}

void EntitySaveMode::edit() { rMode = NAME; }				

void EntitySaveMode::editHold() { }				

void EntitySaveMode::shift() { 
	
	switch(oldMode) {
							
		case EDIT : 	rMode = SOUND;
		case SOUND : 	rMode = EDIT;
		default :	rMode = RETURN;
	}
}

void EntitySaveMode::paramUp() { incrementEvent(SOUND_BANK_SIZE); }
void EntitySaveMode::paramDown() { decrementEvent(SOUND_BANK_SIZE); }

void EntitySaveMode::valueUp() { 
	
	panel->repeat(AUTO_VALUE_UP, 825, 50); 
	incrementEvent();
}

void EntitySaveMode::valueDown() { 

	panel->repeat(AUTO_VALUE_DOWN, 825, 50);
	decrementEvent();
}

void EntitySaveMode::autoEvent(int event) { 

	switch(event) {

		case AUTO_VALUE_UP : 	
			
			incrementEvent(); 

			break;

		case AUTO_VALUE_DOWN :	
			
			decrementEvent(); 

			break;

		case LED_BLINK :

			blinkSaveLEDs();
			
			break;		
	}
}

void EntitySaveMode::blinkSaveLEDs() {

	if(ledBlink) {
		
		ledBlink = false;
		panel->setLEDs(PANEL01_LED_EDIT);
		panel->clearLEDs(PANEL01_LED_SHIFT);
	}
				
	else {
		ledBlink = true;
		panel->setLEDs(PANEL01_LED_SHIFT);
		panel->clearLEDs(PANEL01_LED_EDIT);
	}
}

void EntitySaveMode::displayUpdate() {

	char 	line[PANEL_LCD_MAX_LINE_LEN + 7];

	char*		s;	// string to copy
	int 		p;	// position
		
	panel->clear();

	s = "SAVE ";
	p = 0; while(*s) line[p++] = *s++;

	s = sound->getNumberStr();
	while(*s) line[p++] = *s++;

	s = " TO ";
	while(*s) line[p++] = *s++;

	dp = p;
	
	if(soundNumber >= flash->getnWritableSounds()) s = "---"; 
	else s = sound->getNumberStr(soundNumber);
	while((line[p++] = *s++));

	panel->print(0, 0, line);
}

void EntitySaveMode::displayValueUpdate() {

	char 	line[PANEL_LCD_MAX_LINE_LEN + 7];

	char*		s;	// string to copy
	int 		p;	// position
		
	if(soundNumber >= flash->getnWritableSounds()) s = "---"; 
	else s = sound->getNumberStr(soundNumber);
	p = 0; while(*s) line[p++] = *s++;
		
	s = "  "; while((line[p++] = *s++));

	panel->print(0, dp, line);
}
	
Status EntitySaveMode::incrementValue(int x) {

	int z = soundNumber + x;

	/* Reset destination to A1, instead of factory location */
	if(x < 0 && soundNumber >= flash->getnWritableSounds()) z = 0;

	z = flash->validSaveSoundNumber(z);
	
	if(z == soundNumber) return PARAMETER_UNCHANGED;

	soundNumber = z;
	return PARAMETER_CHANGED;
}

