/********************************************************************************
 * E N T I T Y 									*
 ********************************************************************************
 * File: EntitySound.cpp							*
 *                                                                         	*
 * Copyright (C) A.R.Hardie-Bick 2004-2005                                      *
 *										*
 * This program is free software; you can redistribute it and/or modify		*
 * it under the terms of the GNU General Public License as published by		*
 * the Free Software Foundation; either version 2 of the License, or (at	*
 * your option) any later version.						*
 *										*
 * This program is distributed in the hope that it will be useful, but		*
 * WITHOUT ANY WARRANTY; without even the implied warranty of			*
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU		*
 * General Public License for more details.					*
 *										*
 * You should have received a copy of the GNU General Public License		*
 * along with this program; if not, write to the Free Software			*
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.			*
 ********************************************************************************/

#include "EntitySound.h"

//-----------------------------------
//   Instantiate object counters
//-----------------------------------

int Page::nPages = 0;
int Parameter::nParameters = 0;
int Parameter::nParametersTotal = 0;

//----------------------------------------------------------------------
//
// 			   Matrix Sources
//
//----------------------------------------------------------------------

// Make sure this matches the matrix enum in the header file

char *matrixSource[] = {

	"CTR1", "CTR2",	"CTR3", "MMOD",
	"KFLW", "KVEL", "VDIF", "STDF",
	"LFO1", "LFO2",	"LFO3", "LFO4", 
	"LFO5", "LFO6", "ENV1", "ENV2",  };

const int N_MATRIX_SOURCES = sizeof(matrixSource) / sizeof(matrixSource[0]);

//----------------------------------------------------------------------
//
// 				ParaName
//
//----------------------------------------------------------------------

ParaName::ParaName(char* n, char* sn) { 

	name = n;
	shortName = sn;
}

char* ParaName::getName() { return name; }
char* ParaName::getName(int x) { return fixName(name, x); }
char* ParaName::getShortName() { return shortName; }
char* ParaName::getShortName(int x) { return fixName(shortName, x); }

char* ParaName::fixName(char* n, int x) {

	switch(x) {
	
		case 1:	return lcase(n);
		case 2: return caps(n);
		case 3: return initcap(n);
		
		default: return n;
	}
}

char* ParaName::lcase(char* n) {
	int x = 0;
	while((fixedName[x++] = (char)tolower(*n++)));	
	return fixedName;
}

char* ParaName::caps(char* n) {
	int x = 0;
	while((fixedName[x++] = (char)toupper(*n++)));	
	return fixedName;
}

char* ParaName::initcap(char* n) {
	int x = 0;
	fixedName[x++] = (char)toupper(*n++);
	while((fixedName[x++] = *n++));	
	return fixedName;
}
	
//-----------------------------------------------------------------------------------------------
//
// 					Parameters
//
//-----------------------------------------------------------------------------------------------

char *Parameter::valStr(int i, int &strLen) {

	/* This is a right justified sign plus 4-digit
	 * number-to-string converter. There are always
	 * 5 characters in the output string. There is
	 * no checking of the input value. */

	static char number[6];
	int 	s, x, y;
	char 	*n;
	bool 	d;
	
	d = TRUE;	// Still on first digit?
	n = number;
	
	if(!i) {

		for(x = 0; x < 4; x++) 
			*n++ = 32;
		*n++ = '0';
		*n++ = 0;

		strLen = 5;
		return number;
	}		

	if(i < 0) {

		s = '-';
		i = ~i + 1;
	} 
	
	else s = ' ';

	for(x = 1000; (x); x /= 10) {
	
		y = i / x;

		if(y) {
	
			i = i % x;
	
			if(d) {
				*n++ = s;
				d = FALSE;
			} 	
			
			*n++ = y + '0';	
			
		} else 	d ? *n++ = ' ' : *n++ = '0';
	}

	if(d) {
		*n++ = ' ';
		*n++ = '0';
	}

	*n++ = 0;
	
	strLen = 5;

	return number;
}

char *Parameter::getValStr(int value, int source, int &strLen) {

	if(source == -1) return baseValStr(value, strLen);
	return valStr(value, strLen);
}

char *Parameter::baseValStr(int value, int &strLen) {

	return valStr(value, strLen);
}

int Parameter::validValue(int x, int source) { 
	
	if(source < 0) {
	
		if(x < min) {
			x = min;
			return x;
		}
		if(x > max) {
			x = max;
			return x;
		}
		return x;
		
	} else {
		
		if(x < MATRIX_MIN) {
			x = MATRIX_MIN;
			return x;
		}
		if(x > MATRIX_MAX) {
			x = MATRIX_MAX;
			return x;
		}
		return x;
	}
	return x; 
}

Parameter::Parameter(char* n, char* sn, int rmin, int rmax) : ParaName(n, sn) { 

	paraIndex = nParametersTotal * (N_MATRIX_SOURCES + 1);
	paraNumber = nParametersTotal;
	
	min = rmin;
	max = rmax;

	if(min == 0) 	nrpnSetPotMinMax(0, 1000);
	else		nrpnSetPotMinMax(-1000, 1000);

	nParameters++;
	nParametersTotal++;
}

void Parameter::nrpnSetPotMinMax(int pmin, int pmax) {

	float	pRange, scale, offset;
	
	/* Base value nrpn pot scaling */

	pRange = pmax - pmin;

	scale = 1024.0 * pRange / 127.0; 
	offset = 1024.0 * pmin + 512.0;

	nrpnPotScale = (int) scale;
	nrpnPotOffset = (int) offset;

	/* Depth value nrpn pot scaling */

	pRange = 1000.0 - (-1000.0);
	
	scale = 1024.0 * pRange / 127.0;
	offset = 1024.0 * -1000.0 + 512.0;

	nrpnPotScaleDepth = (int) scale;
	nrpnPotOffsetDepth = (int) offset;
}

int Parameter::translate(int v, int source) { 
	
	if(source > -1) return mTranslate(v);

	// 1000 = fsd here, 2^23 = DSP fsd,
	// 2 ^ 20 = gives dsp fsd (matrix gain = 8);
	// translation factor = 2^20 / 1000.

	const float K = 1048.576;		// 2^20 / 1000
	float V = v;
	
	return (int)(V * K);
}

int Parameter::mTranslate(int v) { 
	
	// A user value or gain of 1000 should 
	// give a matrix gain of unity. A 
	// direct matrix depth of 2^20 should 
	// correspond to a user depth of 1000.
	
	// (Using FP to maintain pitch accuracy)

	const float K = 1048.576;		// 2^20 / 1000
	float V = v;
	
	return (int)(V * K);
}

int Parameter::translateNRPNPotValue(int potVal, int source) { 
	
	if(source == -1) return (potVal * nrpnPotScale + nrpnPotOffset) >> 10;
	else return (potVal * nrpnPotScaleDepth + nrpnPotOffsetDepth) >> 10;
}

PhaseParameter::PhaseParameter(char *n, char *sn, int min, int max):Parameter(n, sn, min, max) { }

int PhaseParameter::translate(int v, int source) { 

	// Divide standard parameter translation by eight
	// to prevent matrix output from saturating
	// when matrix phase additions wraparound
	
	if(source > -1) return mTranslate(v);

	const float K = 1048576.0 / 8000.0;		// 2^20 / 8000
	float V = v;
	
	return (int)(V * K);
}

int PhaseParameter::mTranslate(int v) { 
	
	const float K = 1048576.0 / 8000.0;		// 2^20 / 8000
	float V = v;
	
	return (int)(V * K);
}

Test1Parameter::Test1Parameter(char *n, char *sn, int min, int max):Parameter(n, sn, min, max) { }

int Test1Parameter::translate(int v, int source) { 
	
	if(source > -1) return mTranslate(v);

	// 8000 = fsd here, 2^23 = DSP fsd,
	// 2 ^ 20 = gives dsp fsd (matrix gain = 8);
	// translation factor = 2^20 / 1000.

	const float K = 1048.576 / 8.0;	// 2^20 / 8000
	float V = v;
	
	return (int)(V * K);
}

Test2Parameter::Test2Parameter(char *n, char *sn, int min, int max):Parameter(n, sn, min, max) { }

int Test2Parameter::translate(int v, int source) { 
	
	if(source > -1) return mTranslate(v);

	return (v - 16) * 65536;
}

/* All positive parameters that translate to -1..+1 in the DSP */

PosParameter::PosParameter(char *n, char *sn, int min, int max):Parameter(n, sn, min, max) { 

	nrpnSetPotMinMax(0, 1000);
}

int PosParameter::translate(int v, int source) { 
	
	if(source > -1) return mTranslate(v);

	// For PosParameter class:
	// 
	// 	   0 -> -2 ^ 20
	// 	1000 -> +2 ^ 20

	const float K = 2097.152;		// 2 x 2^20 / 1000
	float V = v - 500;			// rerange to -500..500
	return (int)(V * K);
}

vLFORateParameter::vLFORateParameter(char *n, char *sn, int min, int max):PosParameter(n, sn, min, max) { 

	nrpnSetPotMinMax(0, 800);
}

sLFORateParameter::sLFORateParameter(char *n, char *sn, int min, int max):PosParameter(n, sn, min, max) { 

	nrpnSetPotMinMax(0, 800);
}

int sLFORateParameter::translate(int v, int source) { 
	
	if(source > -1) return mTranslate(v);

	// For ordinary parameter values:
	// 
	// 	   0 -> -2 ^ 20
	// 	1000 -> +2 ^ 20

	if(v >= 0) {
	
		const float K = 2097.152;		// 2 x 2^20 / 1000
		float V = v - 500;			// rerange to -500..500
	
		return (int)(V * K);
	}
	
	// For midi sync values:
	 
	// -24 .. -1 -> -2^21-23 .. -2^21 
	// -24 .. -1 -> -2097175 .. -2097152 
	//
	// (This special range of values is 
	// identified by the LFO subroutine
	// in the DSP code)

	return v - 2097151;
}

char *sLFORateParameter::baseValStr(int value, int &strLen) {

	if(value >= 0) {

		static char v[16];	// *not* threadsafe! 
		
		char *c = valStr(value, strLen);

		int s = 9-strLen;
		v[s] = 0; while(s) v[--s] = ' ';	// pad with left spaces
		strLen = 9;
		return strcat(v, c);
	}

	/* Values below zero are midi sync values */

	static char *midiSyncStrs[] = {

		"Sync 32/1",	// indexed by -1
		"Sync 24/1",
		"Sync 16/1",
		"Sync 12/1",
		" Sync 9/1",
		" Sync 8/1",	
		" Sync 7/1",	
		" Sync 6/1",	
		" Sync 5/1",	
                " Sync 4/1",	
		" Sync 3/1",	
		" Sync 2/1",	
		" Sync 1/1",	
                " Sync 1/2",	
		" Sync 1/3",	
		" Sync 1/4",	
		" Sync 1/6",	
                " Sync 1/8",
		"Sync 1/12",	
		"Sync 1/16",	
		"Sync 1/24",	
                "Sync 1/32",	
		"Sync 1/48",	
		"Sync 1/96"	// indexed by -24
	};

	strLen = 9;
	
	return midiSyncStrs[(-value) -1];
}
	
SwitchParameter::SwitchParameter(char *n, char *sn, int min, int max):Parameter(n, sn, min, max) { 

	/* See java/switching/Switching.java  */
	int j = (max - min) + 1; 
	ip = 8388608 / (8 * j); 
	
	nrpnSetPotMinMax(min, max);
}

int SwitchParameter::translate(int v, int source) { 

	if(source > -1) return mTranslate(v);

	/* See java/switching/Switching.java for documentation */
	return ip * ((v << 1) - max - min); 	/* Equation #1 */
}

bool SwitchParameter::isContinuous(int source) { 
	
	return source == -1 ? false : true; 
}

OffsetParameter::OffsetParameter(char *n, char *sn, int min, int max):Parameter(n, sn, min, max) { 

	nrpnSetPotMinMax(-50, 50);
}

int OffsetParameter::translate(int v, int source) { 
	
	if(source > -1) return mTranslate(v);

	// 8000 = fsd here, 2^23 = DSP fsd,
	// 2 ^ 20 = gives dsp fsd (matrix gain = 8);
	// translation factor = 2^20 / 8000.

	const float K = 131.072;	// 2^20 / 8000
	float V = v;
	
	return (int)(V * K);
}

char *OffsetParameter::baseValStr(int v, int &strLen) {

	static char s[16];
	if(v < 0) { v *= -1; sprintf(s, "-%2.2i.%2.2i Hz", v / 100, v % 100); }
	else sprintf(s, " %2.2i.%2.2i Hz", v / 100, v % 100);
	strLen = 9;
	return s;
}

PitchParameter::PitchParameter(char *n, char *sn):Parameter(n, sn, -8000, 8000) { 

	nrpnSetPotMinMax(-50, 50);
}

int PitchParameter::translate(int v, int source) { 
	
	if(source > -1) return mTranslate(v);
	return v << 7;
}

char *PitchParameter::baseValStr(int v, int &strLen) {

	static char s[16];
	if(v < 0) { v *= -1; sprintf(s, "-%2.2i.%2.2i", v / 100, v % 100); }
	else sprintf(s, " %2.2i.%2.2i", v / 100, v % 100);
	strLen = 6;
	return s;
}

CutoffParameter::CutoffParameter(char *n, char *sn, int min, int max):Parameter(n, sn, min, max) { 

	if(min == 0) 	nrpnSetPotMinMax(250, 1300);	// cutoff
	else	 	nrpnSetPotMinMax(-1000, 1000);	// cutoff'
}

/* With the PitchParameter class used for oscillators, a user range of 0..1000 maps onto the full
 * 64K P2F LUT, which is addressed by matrix output values from -1..1, ie, it's a bipolar LUT.
 * With the CutoffParameter class, we are dealing with two subtly different user parameters:
 * cutoff and cutoff'. When filter channels are unlinked, both these have a user range 0..1000 
 * mapping onto the 64K LUT. However, when linked, it is necessary for cutoff' to be able to
 * go negative, and so the base parameters supplied to the matrix must be ordinarily only
 * positive, so the full range of the 64K P2F LUT is addressed by matrix outputs from 0..1.
 * This is achieved below by halving the matrix depth parameters, so that bipolar signals
 * (such as KFLW) having range -1..1 have their range reduced to one, corresponding to the
 * full range of addresses for the 64K P2I LUT. */

int CutoffParameter::translate(int v, int source) { 
	
	if(source > -1) return mTranslate(v);

	// Filter cutoff units are tenths of semitones.
	// The path to the LUT addressed in cents is
	//
	//  *8   (matrix gain)
	//  /128 (right shift from matrix to LUT) 
	// 
	// so gain is:
	//
	//  g = 40 * 128 / 8 = 640
	
	return v * 640;
}

int CutoffParameter::mTranslate(int v) { 
	
	// Half usual gain for a matrix depth

	const float K = 1048.576 / 2;
	float V = v;
	
	return (int)(V * K);
}

char *CutoffParameter::baseValStr(int v, int &strLen) {

	static char s[16];

	if(v < 0) { v *= -1; sprintf(s, "-%3.3i.%1.1i", v / 10, v % 10); }
	else sprintf(s, " %3.3i.%1.1i", v / 10, v % 10);

	strLen = 6;
	return s;
}

LevelParameter::LevelParameter(char *n, char *sn, int min, int max):Parameter(n, sn, min, max) { }

int LevelParameter::translate(int v, int source) {

	if(source > -1) return mTranslate(v);

	// A user level base value of 1000 should give 
	// a level gain of unity. Levels are eventually 
	// shifted left 3 bits in the matrix, so a user 
	// level of 1000 corresponds to a DSP base 
	// level value of 2^20.
	
	const int t = 1048;	// 2^20 / 1000 (do not round up)

	return v * t;
}

int LevelParameter::mTranslate(int v) {

	// Level matrix depths are only 
	// valid for -1000 to +1000, since
	// a greater than unity level change
	// has no useful meaning.
	
	v = v > 1000 ? 1000 : v < -1000 ? -1000 : v;

	// User matrix gain (for level matrix) of 1000
	// corresponds to a DSP matrix depth of unity
	
	const int t = 8388;	// 2^23 / 1000 (do not round up)

	return v * t;
}

VolumeParameter::VolumeParameter(char *n, char *sn, int min, int max):LevelParameter(n, sn, min, max) { }

StringParameter::StringParameter(char *n, char *sn, int a):SwitchParameter(n, sn, 0, a-1) {

	paraString = new char* [a+1];
	paraLength = new int [a+1];

	addIndex = 0;
}

void StringParameter::addString(char *s) {

	int i = 0; while(s[i]) i++;
	paraLength[addIndex] = i;
	paraString[addIndex] = s;

	addIndex++;
}

char *StringParameter::getValStr(int value, int source, int &strLen) {

	if(source >= 0) return valStr(value, strLen);

	strLen = paraLength[value];
	return paraString[value];
}

StringParameter::~StringParameter() {

	for(int x = 0; x <= max; x++) {
		delete paraString[x];
	}

	delete [] paraString;
	delete [] paraLength;
}

WavetypeParameter::WavetypeParameter(char *n, char *sn):StringParameter(n, sn, 2) {

	addString("Sawtooth");
	addString("  Square");
}

OnOffParameter::OnOffParameter(char *n, char *sn):StringParameter(n, sn, 2) {

	addString("Off");
	addString(" On");
}

VoiceAssignParameter::VoiceAssignParameter(char *n, char *sn):OnOffParameter(n, sn) { }

SourceParameter::SourceParameter(char *n, char *sn):StringParameter(n, sn, 9) {

	addString("  O1");
	addString("  O2");
	addString("  O3");
	addString("  NG");
	addString("  RM");
	addString("FLTL");
	addString("FLTR");
	addString("EXTL");
	addString("EXTR");
}

LFOWaveformParameter::LFOWaveformParameter(char *n, char *sn):StringParameter(n, sn, 12) {

	addString("Triangle");
	addString("    Sine");
	addString("Sawtooth");
	addString(" Saw-exp");
	addString("  Square");
	addString("   Pulse");
	addString("  Random");
	addString("Gaussian");
	addString("    Pink");
	addString(" Rand-sm");
	addString("Gauss-sm");
	addString(" Pink-sm");
}

vLFOSyncParameter::vLFOSyncParameter(char *n, char *sn):StringParameter(n, sn, 27) {

	addString("      Off");
	addString("     Note");
	
	addString("LFO1 Anti");
	addString("LFO1 Quad");
	addString("LFO2 Anti");
	addString("LFO2 Quad");
	addString("LFO3 Anti");
	addString("LFO3 Quad");
	addString("LFO4 Anti");
	addString("LFO4 Quad");
	addString("LFO5 Anti");
	addString("LFO5 Quad");
	addString("LFO6 Anti");
	addString("LFO6 Quad");

	addString(" LFO1 S&H");
	addString(" LFO1 NSH");
	addString(" LFO2 S&H");
	addString(" LFO2 NSH");
	addString(" LFO3 S&H");
	addString(" LFO3 NSH");
	addString(" LFO4 S&H");
	addString(" LFO4 NSH");
	addString(" LFO5 S&H");
	addString(" LFO5 NSH");
	addString(" LFO6 S&H");
	addString(" LFO6 NSH");
	
	addString(" Note S&H");
}

sLFOSyncParameter::sLFOSyncParameter(char *n, char *sn):StringParameter(n, sn, 12) {

	addString("      Off");
	
	addString("LFO5 Anti");
	addString("LFO5 Quad");
	addString("LFO6 Anti");
	addString("LFO6 Quad");
	
	addString(" LFO1 S&H");
	addString(" LFO2 S&H");
	addString(" LFO3 S&H");
	addString(" LFO4 S&H");
	addString(" LFO5 S&H");
	addString(" LFO6 S&H");
	
	addString(" Note S&H");
}

//-----------------------------------------------------------------------------------------------
//
// 					Pages
//
//-----------------------------------------------------------------------------------------------

Page::Page(char* n, char* sn):ParaName(n, sn) { 

	Page::nPages++;
	currentParameterIndex = 0;
}

bool Page::parameterInc(int x) { 
	
	x += currentParameterIndex;
	
	x = x < 0 ? 0 : x >= nParas ? nParas - 1 : x; 

	if(currentParameterIndex == x) return false;
	
	currentParameterIndex = x;

	return true;
}

Parameter* Page::getParameterByIndex(int x) {

	if(x < 0) x = 0;
	if(x >= nPages) x = nPages - 1;
	
	return parameter[x];
}

Parameter* Page::getCurrentParameter() {

	if(currentParameterIndex < 0) currentParameterIndex = 0;
	if(currentParameterIndex >= nPages) currentParameterIndex = nPages - 1;
	
	return parameter[currentParameterIndex];
}

int Page::setCurrentParameterByIndex(int x) { 
	
	return currentParameterIndex = x < 0 ? 0 : x >= nParas ? nParas - 1 : x; 
}

bool Page::setCurrentParameterByPot(int p) { 
	
	int newParameterIndex = potLUT[p & 0xff];

	if(newParameterIndex == currentParameterIndex) return false;
	
	currentParameterIndex = newParameterIndex; 
	return true;
}

int Page::getCurrentParameterIndex() { 
	
	return currentParameterIndex; 
}

OscillatorPage::OscillatorPage(char* n, char* sn):Page(n, sn) {

	Parameter::countZero();
	Parameter **p = parameter;

	*p++ = new Parameter("brightness", "bright", 0);
	*p++ = new PhaseParameter("phase", "phase");		// "Shape" parameter in DSP code
	*p++ = new WavetypeParameter("wavtype", "wave");
	
	*p++ = new SwitchParameter("octave", "octave", -10, 10);
	*p++ = new PitchParameter("pitch", "pitch");
	*p++ = new OffsetParameter("offset", "offset");

	*p++ = new SourceParameter("FM source", "FM src");
	*p++ = new Parameter("FM", "FM", 0);

	*p++ = new OnOffParameter("phase sync","phSyn");	

	*p++ = new Parameter("pan", "pan");	
	*p++ = new LevelParameter("level", "level");

	nParas = Parameter::getCount();
	initPotLUT(nParas, potLUT);
}

FilterPage::FilterPage(char* n, char* sn):Page(n, sn) { 
	
	Parameter::countZero();
	Parameter **p = parameter;
		
	*p++ = new CutoffParameter("cutoff", "cut", 0);
	*p++ = new Parameter("resonance", "res", 0);
	*p++ = new LevelParameter("overdrive", "ovdr");
		
	nParas = Parameter::getCount();
	initPotLUT(nParas, potLUT);
}

NoiseGenPage::NoiseGenPage(char* n, char* sn):Page(n, sn) { 

	Parameter::countZero();
	Parameter **p = parameter;

	*p++ = new LevelParameter("level", "level");
		
	nParas = Parameter::getCount();
	initPotLUT(nParas, potLUT);
}

ExtInputPage::ExtInputPage(char* n, char* sn):Page(n, sn) { 

	Parameter::countZero();
	Parameter **p = parameter;

	*p++ = new LevelParameter("level", "level");
		
	nParas = Parameter::getCount();
	initPotLUT(nParas, potLUT);
}

RingModPage::RingModPage(char* n, char* sn):Page(n, sn) { 

	Parameter::countZero();
	Parameter **p = parameter;

	*p++ = new SourceParameter("source A", "src A");
	*p++ = new SourceParameter("source B", "src B");
	*p++ = new Parameter("pan", "pan");
	*p++ = new LevelParameter("level", "level");
		
	nParas = Parameter::getCount();
	initPotLUT(nParas, potLUT);
}

vLFOPage::vLFOPage(char* n, char* sn):Page(n, sn) { 

	Parameter::countZero();
	Parameter **p = parameter;

	*p++ = new vLFORateParameter("rate", "rate");
	*p++ = new LFOWaveformParameter("wavform", "wave");
	*p++ = new vLFOSyncParameter("sync", "sync");
	*p++ = new PosParameter("delay", "del");
	*p++ = new LevelParameter("level", "levl");

	nParas = Parameter::getCount();
	initPotLUT(nParas, potLUT);
}

sLFOPage::sLFOPage(char* n, char* sn):Page(n, sn) { 

	Parameter::countZero();
	Parameter **p = parameter;

	*p++ = new sLFORateParameter("rate", "rate");
	*p++ = new LFOWaveformParameter("wavform", "wave");
	*p++ = new sLFOSyncParameter("sync", "sync");
	*p++ = new PosParameter("delay", "del");
	*p++ = new LevelParameter("level", "levl");

	nParas = Parameter::getCount();
	initPotLUT(nParas, potLUT);
}

ENVPage::ENVPage(char* n, char* sn):Page(n, sn) { 

	Parameter::countZero();
	Parameter **p = parameter;

	*p++ = new PosParameter("attack", "att");
	*p++ = new PosParameter("decay", "dec");
	*p++ = new PosParameter("sustain", "sus");
	*p++ = new PosParameter("release", "rel");

	nParas = Parameter::getCount();
	initPotLUT(nParas, potLUT);
}

OutputPage::OutputPage(char* n, char* sn):Page(n, sn) { 

	Parameter::countZero();
	Parameter **p = parameter;

	*p++ = new SwitchParameter("octave",  "Octave", -5, 5);
	
	*p++ = new VoiceAssignParameter("stack", "Stack");
	*p++ = new VoiceAssignParameter("polyphony", "Polyphony");
	*p++ = new PosParameter("portamento",        "Portamnto");
	*p++ = new PosParameter("KVEL decay",  "KVEL decy");
	
	*p++ = new Parameter("balance", "Balance");
	*p++ = new Parameter("spread", "Spread");
	*p++ = new Parameter("pan", "Pan");
		
	*p++ = new VolumeParameter("volume", "Volume");

	nParas = Parameter::getCount();
	initPotLUT(nParas, potLUT);
}

//-----------------------------------------------------------------------------------------------
//
// 					Groups
//
//-----------------------------------------------------------------------------------------------

Group::Group() {
	
	currentPageIndex = 0;
	nPages = 0;
}

Page* Group::addPage(Page *p, int groupIndex) {	
	
	page[nPages++] = p;
	p->setIndeces(groupIndex, nPages - 1);  

	return p;
}

Page* Group::getPageByIndex(int x) {

	return x < 0 ? page[0] : x >= nPages ? page[nPages-1] : page[x]; 
}

Page* Group::getCurrentPage() { 

	if(currentPageIndex < 0) currentPageIndex = 0;
	if(currentPageIndex >= nPages) currentPageIndex = nPages - 1;
	
	return page[currentPageIndex]; 
}

bool Group::pageInc(int x) { 
	
	x += currentPageIndex;
	
	x = x < 0 ? 0 : x >= nPages ? nPages - 1 : x; 

	if(currentPageIndex == x) return false;
	
	currentPageIndex = x;

	return true;
}

int Group::getCurrentPageIndex() { 
	
	return currentPageIndex; 
}

int Group::setCurrentPageByIndex(int x) { 
	
	return currentPageIndex = x < 0 ? 0 : x >= nPages ? nPages - 1 : x; 
}

//-----------------------------------------------------------------------------------------------
//
// 				Parameter Structure
//
//-----------------------------------------------------------------------------------------------

ParaStructure::ParaStructure() {

	Parameter::allCountZero();
	
	Page::countZero();
	Page** p = page;   
	
	int n = 0;
	Group** g = group;

	for(int x = 0; x < MAX_GROUPS; x++) g[x] = new Group;

	g[n]->addPage(*p++ = new FilterPage	("FILTER",     "FLT"), n);

	n++;
	
	g[n]->addPage(*p++ = new OscillatorPage	("OSCILLATOR",  "O1"), n);	
	g[n]->addPage(*p++ = new OscillatorPage	("OSCILLATOR",  "O2"), n);		
	g[n]->addPage(*p++ = new OscillatorPage	("OSCILLATOR",  "O3"), n);
	
	n++;
		
	g[n]->addPage(*p++ = new NoiseGenPage	("NOISE GEN",   "NG"), n);	
	
	n++;
		
	g[n]->addPage(*p++ = new ExtInputPage	("EXTERNAL",   "EXT"), n);	
	
	n++;
		
	g[n]->addPage(*p++ = new RingModPage	("RING MOD",    "RM"), n);	
	
	n++;
		
	g[n]->addPage(*p++ = new vLFOPage	("LFO",	      "LFO1"), n);
	g[n]->addPage(*p++ = new vLFOPage	("LFO",	      "LFO2"), n);
	g[n]->addPage(*p++ = new vLFOPage	("LFO",	      "LFO3"), n);
	g[n]->addPage(*p++ = new vLFOPage	("LFO",	      "LFO4"), n);
	g[n]->addPage(*p++ = new sLFOPage	("LFO",	      "LFO5"), n);
	g[n]->addPage(*p++ = new sLFOPage	("LFO",	      "LFO6"), n);
	
	n++;
		
	g[n]->addPage(*p++ = new ENVPage	("ENV",	      "ENV1"), n);
	g[n]->addPage(*p++ = new ENVPage	("ENV",	      "ENV2"), n);
	
	n++;
		
	g[n]->addPage(*p++ = new OutputPage	("",		  ""), n);
	
	n++;
		
	nGroups = n;	
	nPages = Page::getCount();
	nParas = Parameter::getCountTotal();
	nameLength = SOUND_NAME_LENGTH;
	tuningLength = SOUND_TUNING_LENGTH;
	tuningNameLength = SOUND_TUNING_NAME_LENGTH;
	nMatrixSources = N_MATRIX_SOURCES;
	nValues = nParas + nParas * nMatrixSources;

	initPotLUT(nPages, pagePotLUT);	
	initPotLUT(nMatrixSources, matrixPotLUT);
	
	currentGroupIndex = 0;
	currentSource = 0;
	
	// Construct parameter list in order used by interface

	userPara = new Parameter* [nParas];
	makeUserParaList();

	// Construct parameter list in order used by the dsp

	dspPara = new Parameter* [nParas];
	makeDspParaList();

	// Hash table takes a while to initialise -  perhaps use in development only

	hashTable = 0;
	makeHash();
	
	TRACE("nGroups = %i\n", nGroups);
	TRACE("nPages = %i\n", nPages);
	TRACE("nParas = %i\n", nParas);
	TRACE("nDspParas = %i\n", nDspParas);
	TRACE("N_MATRIX_SOURCES = %i\n", N_MATRIX_SOURCES);
}

bool ParaStructure::groupInc(int x) { 
	
	int newGroupIndex = currentGroupIndex + x;
	
	if(newGroupIndex < 0) newGroupIndex = 0;
	if(newGroupIndex >= nGroups) newGroupIndex = nGroups - 1;
       
	if(currentGroupIndex == newGroupIndex) return false;

	currentGroupIndex = newGroupIndex;

	return true;
}

int ParaStructure::setCurrentGroupByIndex(int x) {

	return currentGroupIndex = x < 0 ? 0 : x >= nGroups ? nGroups - 1 : x; 
}

Group *ParaStructure::getCurrentGroup() { 
	
	return group[currentGroupIndex]; 
}

bool ParaStructure::sourceInc(int x) { 

	x += currentSource;
	
	x = x < 0 ? 0 : x >= nMatrixSources ? nMatrixSources - 1 : x; 

	if(currentSource == x) return false;
	
	currentSource = x;

	return true;
}

int ParaStructure::getCurrentSource() {

	return currentSource;
}

bool ParaStructure::setCurrentPageByPot(int p) {
	
	Page *newPage = page[pagePotLUT[p & 0xff]];
	Page *currentPage = group[currentGroupIndex]->getCurrentPage();

	if(newPage == currentPage) return false;
	
	currentGroupIndex = newPage->getGroupIndex();
	group[currentGroupIndex]->setCurrentPageByIndex(newPage->getPageIndex());
	
	return true;
}

int ParaStructure::setCurrentSource(int s) {

	return currentSource = s < 0 ? 0 : s >= nMatrixSources ? nMatrixSources - 1 : s; 
}

bool ParaStructure::setCurrentSourceByPot(int p) {

	int newSource = matrixPotLUT[p & 0xff]; 
	
	if(newSource != currentSource) {
		
		currentSource = newSource;
		return true;
	}	
	
	return false; 
}

void ParaStructure::makeUserParaList() {

	int i = 0;
	
	for(int g = 0; g < nPages; g++) {
		Page *pg = page[g];
		for(int p = 0; p < pg->getnParas(); p++) {
			userPara[i++] = pg->getParameterByIndex(p);
		}
	}
}
	
void ParaStructure::makeDspParaList() {

	/* Optional code generation for file dsp/sound_equ.asm */
	//#define GENERATE_DSP_SOUND_STRUCTURE

	/* The idea here is that you generate a list of parameters in a special
	 * order for the dsp. The only reason for doing this is to enable the
	 * dsp to calculate the matrix without having to calculate parameter
	 * jumps on each matrix row. There are two different types of matrix
	 * calculation, direct and product-of-products. From the run-time 
	 * point-of-view, the quickest system is to have a list of parameters 
	 * with all the direct paras in one block, then a block for those that 
	 * are calculated by product-of-products. The only other info is how 
	 * many of these are in each block. The dspParaIndex in each para is 
	 * required for individual dsp parameter updates (sound updates just 
	 * use the ordered list). */
	
	/* During development, it is necessary to create a list of equates for
	 * the dsp assembler. If the parameter order is changed, or the style of 
	 * any parameter's matrix calculation is changed, then this list of equates
	 * needs to be regenerated and copied into the appropriate assembler file
	 * (although the assembler file can be changed manually for small changes) */

	#ifdef GENERATE_DSP_SOUND_STRUCTURE
	TRACE("\n; Ordered list of matrix destinations to go in file: dsp/sound_equ.asm\n\n");

	#endif
	
	int i = 0;
	
	int nDirect = 0;
	int nPoPs = 0;
	
	/* 1st pass - get parameters whose matrix computation is direct */
	
	#ifdef GENERATE_DSP_SOUND_STRUCTURE
	TRACE("\n; 1. Parameters whose matrix computation is direct\n\n");
	#endif

	for(int g = 0; g < nPages; g++) {
		Page *pg = page[g];
		for(int p = 0; p < pg->getnParas(); p++) {
			Parameter *para = pg->getParameterByIndex(p);
			if(para->isDspParameter()) {
				if(!para->usesPoPsMatrix()) {
					para->setDspParaIndex(i);
					generateDspCodeLine(pg, para);
					dspPara[i++] = para;
					nDirect++;
				}
			}
		}
	
		#ifdef GENERATE_DSP_SOUND_STRUCTURE
		TRACE("\n");
		#endif
	}
			
	/* 2nd pass - get parameters whose matrix computation is product-of-products */
	
	#ifdef GENERATE_DSP_SOUND_STRUCTURE
	TRACE("\n; 2. Parameters whose matrix computation is product-of-products\n\n");
	#endif

	for(int g = 0; g < nPages; g++) {
		Page *pg = page[g];
		for(int p = 0; p < pg->getnParas(); p++) {
			Parameter *para = pg->getParameterByIndex(p);
			if(para->isDspParameter()) {
				if(para->usesPoPsMatrix()) {
					para->setDspParaIndex(i);
					generateDspCodeLine(pg, para);
					dspPara[i++] = para;
					nPoPs++;
				}
			}
		}
	
		#ifdef GENERATE_DSP_SOUND_STRUCTURE
		TRACE("\n");
		#endif
	}

	nDspParas = i;

	/* Place non-dsp paras at the end of the list where they can easily be excluded */
	
	for(int g = 0; g < nPages; g++) {
		Page *pg = page[g];
		for(int p = 0; p < pg->getnParas(); p++) {
			Parameter *para = pg->getParameterByIndex(p);
			if(!para->isDspParameter()) {
				para->setDspParaIndex(i);
				dspPara[i++] = para;
			}
		}
	}

	if((nPoPs * N_MATRIX_SOURCES) > 256) {

		TRACE("CONFIGURATION ERROR: Too many product-of-product type paras.\n");
		TRACE("nPoPs * N_MATRIX_SOURCES > 256. The number of words\n");
		TRACE("reserved in DSP memory for X:soundPnData is 256.\n\n");
	}
	
	#ifdef GENERATE_DSP_SOUND_STRUCTURE

	TRACE("\n; Structure of parameter list by matrix style:\n\n");
	
	TRACE("nDirect			equ	%i\n", nDirect);
	TRACE("nPoPs			equ	%i\n", nPoPs);

	TRACE("\n; Size of data:\n\n");
	
	TRACE("nMatrixSources		equ	%i\n", N_MATRIX_SOURCES);	
	TRACE("nParas			equ	%i\n", i);	
	TRACE("soundDataSize		equ	%i\n", nDspParas*(N_MATRIX_SOURCES+1));	
	
	TRACE("\n; Matrix sources:\n\n");
	
	for(int s = 0; s < N_MATRIX_SOURCES; s++) {

		TRACE("%s\t		equ	%i\n", matrixSource[s], s);
	}

	TRACE("\n; End of auto-generated code\n\n");
	
	#endif
}

void ParaStructure::generateDspCodeLine(Page *pg, Parameter *para) {

	#ifdef GENERATE_DSP_SOUND_STRUCTURE
	
	char keyText[32];

	char *a = pg->getShortName();
	char *b = para->getShortName();
	char *c = keyText;

	// short page name ie "NFT"
			
	if(!*a) a = "SND";
	while(*a) { if(*a != 32) *c++ = toupper(*a); a++; }

	// short para name ie "freq"

	while(*b) {

		switch(*b) {

			case 32 : 	break;			// skip over any spaces
			case 39 :	*c++ = 'b'; break;	// replace trailing apostrophe
			default	:	*c++ = tolower(*b); break;
		}
		
		b++;
	}

	*c = 0;

	int d = 16 - strlen(keyText);
	while(d--) *c++ = 32;
	*c++ = 0;

	TRACE("%s	ds	1\n", keyText);

	#endif
}

Parameter* ParaStructure::getDspPara(int i) { return dspPara[i]; };

Page* ParaStructure::getPageByIndex(int i) { return page[i]; }

char *ParaStructure::getSourceName(int index) {

	index = index > N_MATRIX_SOURCES ? N_MATRIX_SOURCES : index;
	index = index < 0 ? 0 : index;

	return matrixSource[index];
}

Parameter *ParaStructure::getParaByIndex(int i) {
	
	if((i < 0) || (i >= nParas)) return 0;
	return userPara[i];
}

Parameter *ParaStructure::getParaByName(char* keyText) {

	char lcKeyText[32];

	char *c = lcKeyText;
	while(*keyText) *c++ = tolower(*keyText++);
	*c++ = 0;

	if(!hashTable) { TRACE("Creating hashTable\n"); makeHash(); }
	Parameter *para = (Parameter*)hashTable->lookup(lcKeyText);

	if(!para) { TRACE("ERROR: Could not set named parameter: %s\n", lcKeyText); return false; }

	return para;
}

bool ParaStructure::setCurrentGroupPageParaByIndex(int index) {

	index /= (N_MATRIX_SOURCES + 1);
	if(index >= nParas) return false;

	for(int gi = 0; gi < nGroups; gi++) {
		
		Group *grp = group[gi];
		
		for(int ai = 0; ai < grp->getnPages(); ai++) {

			Page *pge = grp->getPageByIndex(ai);

			for(int pi = 0; pi < pge->getnParas(); pi++) {

				Parameter *para = pge->getParameterByIndex(pi);

				if(para->getParaNumber() == index) { // found it

					pge->setCurrentParameterByIndex(pi);
					grp->setCurrentPageByIndex(ai);
					currentGroupIndex = gi;

					return true;
				}
			}
		}
	}

	return false;
}
	
void ParaStructure::makeHash() {

	// - - - - - - - - - - - - - - - - - - - - - - - - - - 
	// Create hash table of parameters for access by name

	// This takes rather a long time (~2 seconds), so 
	// should only be be used during development. It
	// creates a significant initialisation delay.
	// In use, the hash table is extremely fast.
	
	hashTable = new Hash();

	for(int g = 0; g < nPages; g++) {

		Page *pg = page[g];

		for(int p = 0; p < pg->getnParas(); p++) {

			Parameter *para = pg->getParameterByIndex(p);

			char keyText[32];

			char *a = pg->getShortName();
			char *b = para->getShortName();
			char *c = keyText;

			while(*a) *c++ = tolower(*a++);	// short page name ie "NFT"
			if(c != keyText) *c++ = ' ';	// Space only if page name has chars

			while(*b) *c++ = tolower(*b++);	// short para name ie "freq"
			*c++ = 0;
 
			if(!hashTable->add(keyText, para)) {
				TRACE("ERROR: Duplicate hash key: %s\n", keyText);
			}
		}
	}

	//int a, b; hashTable->getStats(a, b); // for debugging
	//TRACE("Direct hash entries = %i, of total = %i\n", a, b);	
}

//-----------------------------------------------------------------------------------------------
//
// 				     EntitySound
//
//-----------------------------------------------------------------------------------------------

EntitySound::EntitySound(ParaStructure* p) {

	paras = p;

	value = new int[paras->getnValues()];
	oldval = new int[paras->getnValues()];
	mute = new bool[paras->getnValues()];

	name = new char[paras->getnameLength()];
	oldName = new char[paras->getnameLength()];

	tuning = new int[paras->gettuningLength()]; // No. of notes
	tuningName = new char[paras->gettuningNameLength()+2];	
	
	pot = new char[4];

	clearSound();		// reset values appropriately
	makeAvailable();	// this sound available in sound object pool
}

EntitySound::~EntitySound() {
	
	delete [] pot;
	delete [] tuning;
	delete [] tuningName;
	delete [] oldName;
	delete [] name;
	delete [] mute;
	delete [] oldval;
	delete [] value;
}

// An explanation of the indexing process: Each Parameter object contains
// the paraIndex value, that gives an offset into the ints of the sound.
// If there are n matrix sources, each parameter is associated with n+1
// data values, including the base value. The default arguments defined 
// in EntitySound.h give source=-1, which is then added to at least one
// to provide the appropriate offset into the field of ints:

int EntitySound::getTranslatedValue(Parameter* parameter, int source) {

	int i = parameter->getParaIndex() + source + 1;

	if(mute[i]) return parameter->translate(0, source);
	return parameter->translate(value[i], source);
}
	
int EntitySound::getValue(Parameter* parameter, int source) {

	int i = parameter->getParaIndex() + source + 1;
	if(mute[i]) return 0;
	
	int v = value[i];
	return parameter->validValue(v, source);	
}

Status EntitySound::incValue(Parameter* parameter, int incr, int source) {
	
	/* incValue() is the function used when changing parameter values during
	 * editing. It includes all the checks for changes, zero-crossing, change
	 * counting, etc. For loading a new sound, use loadValue(), as this updates 
	 * the oldValue[] array, and unmutes the parameter. */

	/* Zero point crossing is now part of general stop value processing, 
	 * which facilitates encoder pauses at -1000, 0, +1000, for most 
	 * continuous parameters. */
	 	
	/* With muting, the following behaviour is defined: If parameter was muted
	 * and user turns encoder, muting is turned off and encoder increments upon
	 * the muted value of zero (same with inc/dec buttons). */
	
	int i = parameter->getParaIndex() + source + 1;

	int oldValue, newValue;
	
	if(mute[i]) 	{ oldValue = 0; newValue = incr; }
	else		{ oldValue = value[i]; newValue = oldValue + incr; }
		
	newValue = parameter->validValue(newValue, source);

	int delta = newValue - oldValue;

	int stopValue = 1000;	// Default stops at -1000, 0 and +1000

	/* Disable non-zero stop values for base parameters that don't want them */
	if(source == -1 && !parameter->isStoppedAt1000()) stopValue = 100000;	
		
	if(delta > 0) {
	
		if(oldValue < -stopValue) {		// check for -1000

			if(newValue >= -stopValue) {

				newValue = -stopValue;
				value[i] = newValue;
		
				return PARAMETER_THRESHOLD;
			}
		}

		else if(oldValue < 0) {			// check for zero

			if(newValue >= 0) {
			
				newValue = 0;
				updatenChangedParas(oldval[i], oldValue, newValue);
				if(source > -1) nMatrixLinks--;
				value[i] = newValue;
		
				return PARAMETER_THRESHOLD;
			}
		}

		else if(oldValue < stopValue) {		// check for +1000

			if(newValue >= stopValue) {

				newValue = stopValue;
				value[i] = newValue;
		
				return PARAMETER_THRESHOLD;
			}
		}

		if(newValue && !oldValue) {		// Check for added matrix link

			if(source > -1) {
				
				if(++nMatrixLinks > MAX_MATRIX_LINKS) {
			
					--nMatrixLinks;
					return PARAMETER_EXCEPTION;	// Indicates full matrix
				}
			
			}
			
			mute[i] = 0;
		}

		updatenChangedParas(oldval[i], oldValue, newValue);

		value[i] = newValue;

		return PARAMETER_CHANGED;
	}

	else 
		
	if(delta < 0) {
	
		if(oldValue > stopValue) {		// check for +1000

			if(newValue <= stopValue) {

				newValue = stopValue;
				value[i] = newValue;
		
				return PARAMETER_THRESHOLD;
			}
		}

		else if(oldValue > 0) {			// check for zero

			if(newValue <= 0) {
			
				newValue = 0;
				updatenChangedParas(oldval[i], oldValue, newValue);
				if(source > -1) nMatrixLinks--;
				value[i] = newValue;
		
				return PARAMETER_THRESHOLD;
			}
		}

		else if(oldValue > -stopValue) {	// check for -1000

			if(newValue <= -stopValue) {

				newValue = -stopValue;
				value[i] = newValue;
		
				return PARAMETER_THRESHOLD;
			}
		}

		if(newValue && !oldValue) {		// Check for added matrix link

			if(source > -1) {
				
				if(++nMatrixLinks > MAX_MATRIX_LINKS) {
			
					--nMatrixLinks;
					return PARAMETER_EXCEPTION;	// Indicates full matrix
				}
			
			}
			
			mute[i] = 0;
		}

		updatenChangedParas(oldval[i], oldValue, newValue);

		value[i] = newValue;

		return PARAMETER_CHANGED;
	}

	else
	
	return PARAMETER_UNCHANGED;
}

Status EntitySound::changeValue(Parameter* parameter, int newValue, int source) {
	
	/* changeValue() is the function used when modifying parameter values via midi */

	int i = parameter->getParaIndex() + source + 1;

	int oldValue = value[i];

	if(mute[i]) oldValue = 0; 

	newValue = parameter->validValue(newValue, source);

	if(newValue && !oldValue) {		// Check for added matrix link

		if(source > -1) {
				
			if(++nMatrixLinks > MAX_MATRIX_LINKS) {
			
				--nMatrixLinks;
				return PARAMETER_EXCEPTION;	// Indicates full matrix
			}
		
		}
			
		mute[i] = 0;
	}

	updatenChangedParas(oldval[i], oldValue, newValue);

	value[i] = newValue;

	return PARAMETER_CHANGED;
}

void EntitySound::updatenChangedParas(int sValue, int pValue, int nValue) {

	// sValue = stored value
	// pValue = previous value
	// nValue = new value
	
	if(sValue == pValue && nValue != pValue) nChangedParas++;
	else if(nValue != pValue && nValue == sValue) nChangedParas--;
}

void EntitySound::loadValue(Parameter *parameter, int intendedValue, int source) {

	/* loadValue() is used when loading sounds, as it updates the oldValue[]  
	 * array as well as the value[] array. Also it avoids all the checking
	 * that's required when a parameter is edited, so it may be a bit faster 
	 * than setValue(). After all paras have been loaded, set nChangedValues 
	 * to zero (you have to do this manually) */
	
	int newValue = parameter->validValue(intendedValue, source);
	int i = parameter->getParaIndex() + source + 1;	

	if((source > -1) && newValue) nMatrixLinks++;
	
	value[i] = newValue;
	oldval[i] = newValue;
	mute[i] = false;
}

bool EntitySound::loadValue(char* paraname, int intendedValue, int source) {

	Parameter* para = paras->getParaByName(paraname);

	if(!para) { TRACE("ERROR: Could not set named parameter: %s\n", paraname); return false; }

	loadValue(para, intendedValue, source); 
	return true;	
}

Parameter *EntitySound::getParameterByIndex(int index) {

	return paras->getParaByIndex(index);
}

char *EntitySound::getValStr(Parameter* parameter, int source, int &strLen) {

	int i = parameter->getParaIndex() + source + 1;
	int v = value[i];
	if(mute[i]) v = 0;
	return parameter->getValStr(v, source, strLen);
}

bool EntitySound::getMute(Parameter* parameter, int source) {

	return mute[parameter->getParaIndex() + source + 1];
}

Status EntitySound::setMute(Parameter* parameter, bool m, int source) {

	if((source < 0) && (!parameter->isMutable())) return PARAMETER_UNCHANGED;

	int v;

	int i = parameter->getParaIndex() + source + 1;

	if(mute[i] == m) return PARAMETER_UNCHANGED;

	if((source > -1) && (value[i])) {
	
		if(!m) {
		
			if(++nMatrixLinks > MAX_MATRIX_LINKS) {
			
				--nMatrixLinks;
				return PARAMETER_EXCEPTION;
			}	
			
		} else nMatrixLinks--;

	}

	mute[i] = m;

	v = m ? 0 : value[i];
	
	updatenChangedParas(oldval[i], value[i], v);
	
	return PARAMETER_CHANGED;
}	

Status EntitySound::toggleMute(Parameter* parameter, int source) {

	if((source < 0) && (!parameter->isMutable())) return PARAMETER_UNCHANGED;

	int v;
	
	int i = parameter->getParaIndex() + source + 1;

	bool m = !mute[i];

	if((source > -1) && (value[i])) {
		
		if(!m) {
		
			if(++nMatrixLinks > MAX_MATRIX_LINKS) {
			
				--nMatrixLinks;
				return PARAMETER_EXCEPTION;
			}	
			
		} else nMatrixLinks--;

	}

	mute[i] = m;

	m ? v = 0 : v = value[i];
	
	updatenChangedParas(oldval[i], value[i], v);

	return PARAMETER_CHANGED;
}	

int EntitySound::gettuning(int note) {

	return note < 0 ? tuning[0] : note > 127 ? tuning[127] : tuning[note];
}

void EntitySound::loadDefaultValues(int n) {

	clearSound(); 	// (includes clearing name)
	number = n;	// Sound number;

	/* NB - use matrix abbreviated form of name, case irrelevant */

	loadValue("Polyphony", 1);

	loadValue("FLT cut", 800);
	loadValue("FLT cut", 500, CTR1);
	loadValue("FLT cut", 100, KFLW);
	loadValue("FLT res", 673);
	loadValue("FLT res", 500, CTR2);
	loadValue("FLT ovdr", 400);
	loadValue("FLT ovdr", 1000, CTR3);
	
	loadValue("O1 bright", 369);
	loadValue("O1 pitch", 1000, KFLW);
	loadValue("O1 FM src", 2);
	loadValue("O1 level", 327);
	
	loadValue("O2 bright", 369);
	loadValue("O2 octave", 1);
	loadValue("O2 pitch", 1000, KFLW);
	loadValue("O2 FM src", 0);
	
	loadValue("O3 bright", 369);
	loadValue("O3 octave", 0);
	loadValue("O3 pitch", 1000, KFLW);
	loadValue("O3 FM src", 1);
	
	loadValue("RM src A", 0);
	loadValue("RM src B", 1);

	loadValue("LFO1 rate", 100);
	loadValue("LFO2 rate", 200);
	loadValue("LFO3 rate", 300);
	loadValue("LFO4 rate", 400);
	loadValue("LFO5 rate", 500);
	loadValue("LFO6 rate", 600);

	loadValue("LFO1 levl", 1000);
	loadValue("LFO2 levl", 1000);
	loadValue("LFO3 levl", 1000);
	loadValue("LFO4 levl", 1000);
	loadValue("LFO5 levl", 1000);
	loadValue("LFO6 levl", 1000);
	
	loadValue("ENV1 att", 50);
	loadValue("ENV1 sus", 1000);

	loadValue("ENV2 att", 50);
	loadValue("ENV2 sus", 1000);

	loadValue("KVEL decy", 700);
	loadValue("Spread", 1000);
	loadValue("Volume", 300);
	loadValue("Volume", 1000, ENV1);

	/* Testing octave switching */
	//loadValue("FLT ovdr", 0, CTR3);
	//loadValue("O1 wave", 300, CTR3);
}

void EntitySound::clearSound() {

	for(int g = 0; g < paras->getnPages(); g++) {
		Page *page = paras->getPageByIndex(g);
		for(int p = 0; p < page->getnParas(); p++) {
			Parameter *para = page->getParameterByIndex(p);
			for(int s = -1; s < paras->getnMatrixSources(); s++) {
				loadValue(para, 0, s);
			}
		}
	}
	
	int x = 0; while(x < (paras->getnameLength() - 2)) name[x++] = ' '; name[x++] = 0; strcpy(oldName, name);

	/* Equal tempered semitone tuning, middle C (60) = 0, 400 steps per semitone ( cent intervals) */
	for(int note = 0; note < paras->gettuningLength(); note++) tuning[note] = (note - 60) * 400;
	
	strcpy(tuningName, "12-tone Equal");	// Tuning name not accessible via UI, but stored anyway
	tuningNumber = 0;			// From midi tuning standard (unused)
	
	number = 0;
	nMatrixLinks = 0;	
	nChangedParas = 0;
	
	pot[0] = pot[1] = pot[2] = 0;
}

void EntitySound::loadRandomValues() {

// This is mainly for testing compression and filing,
// but could be used as a basis for generating sounds

	static r250 r; // static ensures different values in each call

	nMatrixLinks = 0;

	/* First set all non-matrix parameters randomly */

	for(int g = 0; g < paras->getnPages(); g++) {

		Page *page = paras->getPageByIndex(g);
		
		for(int p = 0; p < page->getnParas(); p++) {

			Parameter *para = page->getParameterByIndex(p);

			int range = para->getmax() - para->getmin();
			int mask = 0; while(mask < range) mask = (mask << 1) | 1;

			/* analytic assignment of random value */
			int v; while((v = r.rng() & mask) > range);
			
			loadValue(para, v + para->getmin());

			/* In preparation, clear the matrix depths and set mutes */

			for(int s = 0; s < paras->getnMatrixSources(); s++) {
				loadValue(para, 0, s);
				setMute(para, true, s); // to track if depth has been set
			}
		}
	}

	/* Now select 500 matrix depths at random, setting them with random values */

	int mrange = MATRIX_MAX - MATRIX_MIN;
	int mmask = 0; while(mmask < mrange) mmask = (mmask << 1) | 1;
	int srange = paras->getnMatrixSources() - 1;
	int smask = 0; while(smask < srange) smask = (smask << 1) | 1;

	int grange = paras->getnPages() - 1;
	int gmask = 0; while(gmask < grange) gmask = (gmask << 1) | 1;
	
	for(int x = 0; x < MAX_MATRIX_LINKS; x++) {
		
		bool m = 0;

		do {	int g; while((g = r.rng() & gmask) > grange);

			Page *page = paras->getPageByIndex(g);
		
			int prange = page->getnParas() - 1;
			int pmask = 0; while(pmask < prange) pmask = (pmask << 1) | 1;

			int p; while((p = r.rng() & pmask) > prange);
			
			Parameter *para = page->getParameterByIndex(p);

			int s; while(( s = r.rng() & smask) > srange);

			int v; while((v = r.rng() & mmask) > mrange);
	
			v += MATRIX_MIN;

			if(v && (m = getMute(para, s))) {	// Must be a non-zero value
								// for a muted matrix depth				
				loadValue(para, v, s);
			}
				// Repeat until a muted parameter is set. Avoids setting 
		} while(!m);	// same depth twice, so 500 *different* depths are set.

	}

	nChangedParas = 0;
}

char EntitySound::getBullet() const {

	return !nChangedParas ? 32 : 2;	
}

void EntitySound::setName(char *newName) {

	bool a = nameChanged();

	int i = 0; char *s = newName; char *d = name;		
	while((*d++ = *s++)) i++;
	while(i++ < paras->getnameLength()) *d++ = 0;	

	bool b = nameChanged();

	if(a && !b) { nChangedParas--; return; }
	if(b && !a) { nChangedParas++; return; }	
}

char *EntitySound::getNumberStr(int i) {

	static char 	numStr[7];
	static int 	x, y;
	static char 	*n;
	static bool 	d;
	
	numStr[0] = (char)(i / SOUND_BANK_SIZE) + 'A';

	/* Preset sound banks start at 'X' */
	if(numStr[0] >= 'F') numStr[0] += 'X' - 'F';
			
	i = (i % SOUND_BANK_SIZE) + 1;

	n = numStr + 1;
	d = true;
	
	for(x = 100; (x); x /= 10) {
	
		y = i / x;

		if(y) {
	
			i = i % x;

			*n++ = y + '0';	

			d = false;
			
		} else 	if(!d) *n++ = '0';
	}

	*n++ = 0;
	
	return numStr;
}	

// Matrix scan functions - These take a current position in the matrix, defined by a 
// source and a destination, and scan 1) to the next destination with a non-zero (and 
// unmuted) value depth for the same source or 2) to the next Source with a non-zero
// value depth for the same destination.

bool EntitySound::findNextMatrixDestination(int mSrc, Parameter *para) {

	// Now that there are a couple of parameters whose matrix values are not supplied
	// direct to the user, these paras must be skipped when they are found in a
	// destination scan.

	int incr = N_MATRIX_SOURCES + 1;
	int index = para->getParaIndex() + mSrc + 1 + incr;
	int indexLimit = paras->getnValues();
	
	while(index < indexLimit) {

		if(!mute[index] && value[index]) {

			Parameter *p = getParameterByIndex(index / (N_MATRIX_SOURCES + 1));
			
			if(p->isDspParameter()) { // Could be one of the two non-dsp paras
				
				return paras->setCurrentGroupPageParaByIndex(index);
			}
				
		}

		index += incr;
	}

	index = mSrc + 1;
	indexLimit = para->getParaIndex() + mSrc + 1;

	while(index < indexLimit) {

		if(!mute[index] && value[index]) {

			Parameter *p = getParameterByIndex(index / (N_MATRIX_SOURCES + 1));
			
			if(p->isDspParameter()) { // Could be one of the two non-dsp paras
				
				return paras->setCurrentGroupPageParaByIndex(index);
			}
				
		}

		index += incr;
	}

	return true;
}

int EntitySound::findNextMatrixSource(int mSrc, Parameter *para) {

	// Since the button used to access this operation is a "down"
	// button, it makes intuitive sense to scan backwards through
	// the matrix sources.

	// No need to avoid the non-dsp paras here, because it should not
	// have been possible to enter matrix scan with these paras selected
	// (and they cannot be accessed by destination scan, since that has been
	// coded to exclude them).

	int source = mSrc - 1; 				// -1 to start at next src
	if(mSrc < 0) mSrc = N_MATRIX_SOURCES - 1;

	int index = para->getParaIndex() + source + 1; 	// +1 to access matrix paras

	while(source >= 0) {

		if(!mute[index] && value[index]) return paras->setCurrentSource(source);

		source--;
		index--;
	}

	source = N_MATRIX_SOURCES - 1;
	index = para->getParaIndex() + source + 1;
	
	while(source > mSrc) {

		if(!mute[index] && value[index]) return paras->setCurrentSource(source);

		source--;
		index--;
	}

	return mSrc;			
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
// A word of explanation about compression and decompression. Compression is performed on
// the assumption that only a subset of the matrix depths are non-zero (the maximum number
// is defined as N_MAX_MATRIX_LINKS). Since there is a fixed flash file size, the
// compression must be very predictable, and any advantage can only be obtained by looking
// at the worst case scenario. So it is assumed that all values are 14 bit quantities
// (which may be helpful also when sending sounds over MIDI System Exclusive). Most of these
// values are set to zero, because we are imposing a limit on the number of active matrix
// links, to enable storage to be made more efficient. Huffman analysis showed that a
// single bit should be used for storing these zeros, and other bit lengths for other
// data. To keep it very predictable, and to avoid unnecessary processing for storage and
// retrieval, bit lengths for the other data values are all set to 15.
//
// The sound name, in uncompressed form, is put at the start of the compressed data stream. 
// This makes file saving more elegant, and the name can be snooped easily.
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

int EntitySound::compress(void *buf) {

	char *buffer = (char*) buf;

	for(int x = 0; x < paras->getnameLength(); x++) *buffer++ = name[x];

	//static BitStream b; b.init(buffer);
	// Changed to non-static, since this function may need to be thread safe
	BitStream b; b.init(buffer);

	for(int x = 0; x < paras->getnValues(); x++) {

		int v = mute[x] ? 0 : value[x];

		/* Use 15 bits, MS bit = 0, or use 1 bit = 1 for zero */
		
		v ? b.put(v & ~0x4000, 15) : b.put(1,1); /* (MS bit is first) */
	}

	for(int x = 0; x < 3; x++) b.put(pot[x], 7); 	/* Control pot values */

	buffer += b.getnBytes();
	
	for(int x = 0; x < paras->gettuningLength(); x++) { 	/* Tuning data */
		
		*buffer++ = (char)((tuning[x] >> 8) & 0xff);
		*buffer++ = (char)(tuning[x] & 0xff);
	}

	for(int x = 0; x < paras->gettuningNameLength(); x++) {	 /* Tuning name */

		*buffer++ = (char)tuningName[x];
	}

	*buffer++ = (char)tuningNumber;				/* Tuning number */
	
	for(int x = 0; x < paras->getnValues(); x++) oldval[x] = mute[x] ? 0 : value[x];
	strcpy(oldName, name);
	nChangedParas = 0;

	int dataSize = 	b.getnBytes() + 
			paras->getnameLength() + 
			paras->gettuningLength() * 2 +
			paras->gettuningNameLength() +
			1; // (for tuning number)

	// Save Sounds as Presets - - - - - - - - - - - - 

	#ifdef DEBUG	
	#ifdef SAVE_TO_PRESET // (define in EntitySound.h)

	TRACE("\n\n\t/* Sound: %s %s */\n\n\t", getNumberStr(), getName());
	TRACE("0x%.2X, 0x%.2X,\t/* Number of bytes (MSB, LSB) */\n\n\t", dataSize/256, dataSize%256);
	
	int i = 0; unsigned char *pbuf = (unsigned char*) buf;

	for(int x = 0; x < dataSize / 8; x++) {

		for(int y = 0; y < 8; y++) {

			TRACE("0x%.2X, ",  pbuf[i++]);
		}
	
		TRACE("\n\t");
		
		rtems_task_wake_after(25);	// TRACE needs this time to catch up
	}

	while(i < dataSize) TRACE("0x%.2X, ", pbuf[i++]); 

	TRACE("\n\n"); 
	
	#endif // SAVE_TO_PRESET
	
	// Save Sounds as Human-readable - - - - - - - - 

	#ifdef SAVE_AS_HUMAN_READABLE // (define in EntitySound.h)
	
	TRACE("\n\n\t/* Sound: %s %s */\n\n", getNumberStr(), getName());

	TRACE("Matrix sources\n\n");
	
	for(int s = 0; s < N_MATRIX_SOURCES; s++) {
		if(s == 8) TRACE("\n");
		TRACE("%s ", matrixSource[s]);
	}

	TRACE("\n");
		
	for(int g = 0; g < paras->getnPages(); g++) {

		Page *page = paras->getPageByIndex(g);

		TRACE("\n%s\n\n", page->getShortName());

		for(int p = 0; p < page->getnParas(); p++) {

			Parameter *para = page->getParameterByIndex(p);

			int strlen;
			
			TRACE("%10s %10s\t\t", para->getName(), getValStr(para, -1, strlen));

			for(int s = 0; s < paras->getnMatrixSources(); s++) {
				
				TRACE("%i ", getValue(para, s));

			}
			
			rtems_task_wake_after(35);	// TRACE needs this time to catch up
			TRACE("\n");
		}
	}

	#endif // SAVE_AS_HUMAN_READABLE

	#endif // DEBUG
	
	return dataSize;
}

void EntitySound::deCompress(void *buf) {

	char *buffer = (char*) buf;

	for(int x = 0; x < paras->getnameLength(); x++) oldName[x] = name[x] = *buffer++;
		
	static BitStream b; b.init(buffer);

	nMatrixLinks = 0;
	
	for(int g = 0; g < paras->getnPages(); g++) {

		Page *page = paras->getPageByIndex(g);

		for(int p = 0; p < page->getnParas(); p++) {

			Parameter *para = page->getParameterByIndex(p);

			for(int s = -1; s < paras->getnMatrixSources(); s++) {
				
				int v = b.get(1) ? 0 : b.get(14);

				/* Sign extension from bit 13 to bit 31 */
				v = v & 0x2000 ? v | ~0x1fff : v & 0x1fff;

				loadValue(para, v, s);
			}
		}
	}

	for(int x = 0; x < 3; x++) pot[x] = b.get(7);

	buffer += b.getnBytes();
	
	for(int x = 0; x < paras->gettuningLength(); x++) {
		
		tuning[x] = *buffer++ << 8;
		tuning[x] |= (unsigned char)(*buffer++) & 0xff;
	}

	for(int x = 0; x < paras->gettuningNameLength(); x++) {	

		tuningName[x] = (char)*buffer++;
	}

	tuningNumber = (char)*buffer++;
	
	nChangedParas = 0;
}

unsigned int EntitySound::checksum() const {

	CRC32 c; c.init();

	for(int x = 0; x < paras->getnValues(); x++) {

		int v = mute[x] ? 0 : value[x];
		
		c.add((unsigned char)v >> 8);
		c.add((unsigned char)v);
	}

	return c.checksum();
}

void EntitySound::setTuning(void *buf) {

	unsigned char *buffer = (unsigned char*) buf;

	tuningNumber = *buffer++;

	for(int n = 0; n < 16; n++) tuningName[n] = (char) *buffer++;
	
	for(int n = 0; n < 128; n++) {

		int T, P;	

		// Unpack note tuning
		
		T  = *buffer++ << 14;
		T += *buffer++ << 7;
		T += *buffer++;

		// 7F 7F 7F = no change

		if(T == 0x1FFFFF) continue;

		// Convert to cents

		P = (T * 400) >> 14;
		P = P - 60 * 400;

		tuning[n] = P;
	}
}

