/********************************************************************************
 * E N T I T Y 									*
 ********************************************************************************
 * File: EntitySysexMode.cpp							*
 *                                                                         	*
 * Copyright (C) A.R.Hardie-Bick 2004-2005					*
 *										*
 * This program is free software; you can redistribute it and/or modify		*
 * it under the terms of the GNU General Public License as published by		*
 * the Free Software Foundation; either version 2 of the License, or (at	*
 * your option) any later version.						*
 *										*
 * This program is distributed in the hope that it will be useful, but		*
 * WITHOUT ANY WARRANTY; without even the implied warranty of			*
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU		*
 * General Public License for more details.					*
 *										*
 * You should have received a copy of the GNU General Public License		*
 * along with this program; if not, write to the Free Software			*
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.			*
 *******************************************************************************/

#include "EntitySysexMode.h"

EntitySysexMode::EntitySysexMode(EntityPanel *panel_, EntityMidi *midi_, EntityFlash *flash_, EntityGlobal *global_, EntitySoundManager *soundManager_) : EntityMode(panel_) { 

	midi = midi_; 
	flash = flash_;
	global = global_;
	soundManager = soundManager_;
}

Mode EntitySysexMode::enter() {

	panel->clearLEDs(PANEL01_LED_EDIT, PANEL01_LED_CTRL1, PANEL01_LED_CTRL2, PANEL01_LED_CTRL3);
	panel->setEncoderAcceleration(0);	

	// Ignore all events that are not related to sysex transfer, exit after timeout 

	do {	rMode = SELF;
	
		autoExit();
		
		getMessage(ctrq);

		switch(mBuffer[0]) {

			case LOAD_FLASH_SOUND_FROM_MIDI:		loadFlashSoundFromMidi(); break;
			case LOAD_EDITED_SOUND_FROM_MIDI:		loadEditedSoundFromMidi(); rMode = RETURN; break;
			case LOAD_FLASH_GLOBAL_FROM_MIDI:		loadFlashGlobalFromMidi(); break;
			case LOAD_EDITED_GLOBAL_FROM_MIDI:		loadEditedGlobalFromMidi(); rMode = RETURN; break;
			case LOAD_TUNING_FROM_MIDI:			loadTuningFromMidi(); rMode = RETURN; break;
									
			case REQUEST_STORED_SOUND_TO_MIDI: 		saveSoundToMidi(true); break;
			case REQUEST_EDITED_SOUND_TO_MIDI:		saveSoundToMidi(false); break;
			case REQUEST_STORED_ACTIVE_SOUND_TO_MIDI:	saveActiveSoundToMidi(true); rMode = RETURN; break;
			case REQUEST_EDITED_ACTIVE_SOUND_TO_MIDI:	saveActiveSoundToMidi(false); rMode = RETURN; break;
			case REQUEST_STORED_GLOBAL_TO_MIDI: 		saveGlobalToMidi(true); break;
			case REQUEST_EDITED_GLOBAL_TO_MIDI:		saveGlobalToMidi(false); break;

			case REQUEST_BANK_TO_MIDI:			saveBankToMidi(); break;
			case REQUEST_ALL_SOUNDS_TO_MIDI:		saveAllSoundsToMidi(); break;
			case REQUEST_ALL_TO_MIDI:			saveAllToMidi(); break;

			case AUTO_EXIT :				rMode = RETURN; break;
		}

	} while(rMode == SELF);
			
	panel->clearDelay();

	return rMode;
}

void EntitySysexMode::loadFlashSoundFromMidi() {
	
	int soundNumber  = mBuffer[6] * soundManager->getBankSize() + mBuffer[7];
	
	int bufSize = loadMidiBufferIn();

	flash->save(soundNumber, midiBufferIn, bufSize);

	/* If sysex saved to flash file having current sound number, then update current sound */
	if(soundNumber == soundManager->getNumber()) soundManager->loadSound(soundNumber);
	
	panel->print(0, 0, "MIDI Sysex      ");
	char line[] = 	"Loaded sound    ";
	char* s = soundManager->getNumberStr(soundNumber);
	int p = 13; while(*s) line[p++] = *s++;	
	panel->print(1, 0, line);	
}

void EntitySysexMode::loadEditedSoundFromMidi() {

	int soundNumber = mBuffer[6] * soundManager->getBankSize() + mBuffer[7];
	
	loadMidiBufferIn();

	soundManager->loadSoundFromMidi(soundNumber, midiBufferIn);
}

void EntitySysexMode::loadFlashGlobalFromMidi() { 

	int bufSize = loadMidiBufferIn();

	flash->save(midiBufferIn, bufSize);
	global->unpack(midiBufferIn);

	panel->print(0, 0, "MIDI Sysex      ");
	panel->print(1, 0, "Loaded Config   ");	
}

void EntitySysexMode::loadEditedGlobalFromMidi() { 

	loadMidiBufferIn();
	global->unpack(midiBufferIn);
}

void EntitySysexMode::loadTuningFromMidi() {

	loadMidiBufferIn();
	soundManager->loadTuningFromMidi(midiBufferIn);
}

int EntitySysexMode::saveSoundToMidi(bool stored) {

	int soundBank = mBuffer[1];
	int userSoundNumber = mBuffer[2];
	
	int soundNumber = soundBank * soundManager->getBankSize() + userSoundNumber;

	return saveSoundToMidi(soundNumber, stored);
}

int EntitySysexMode::saveSoundToMidi(int soundNumber, bool stored) {

	int soundBank = soundNumber / soundManager->getBankSize();
	int userSoundNumber = soundNumber % soundManager->getBankSize();
	
	int bufSize = flash->load(soundNumber, midiBufferOut); if(!bufSize) return 0;

	if(stored) midi->newSysexStoredSoundEvent(midiBufferOut, bufSize, soundBank, userSoundNumber);
	else       midi->newSysexEditedSoundEvent(midiBufferOut, bufSize, soundBank, userSoundNumber);

	panel->print(0, 0, "MIDI Sysex      ");
	char line[] =   "Saved sound     ";
	char* s = soundManager->getNumberStr(soundNumber);
	int p = 12; while(*s) line[p++] = *s++;	
	panel->print(1, 0, line);

	return bufSize;	
}

int EntitySysexMode::saveActiveSoundToMidi(bool stored) {

	int soundNumber;
	
	int bufSize = soundManager->loadMidiFromSound(soundNumber, midiBufferOut);

	int soundBank = soundNumber / soundManager->getBankSize();
	int userSoundNumber = soundNumber % soundManager->getBankSize();
	
	if(stored) midi->newSysexStoredSoundEvent(midiBufferOut, bufSize, soundBank, userSoundNumber);
	else       midi->newSysexEditedSoundEvent(midiBufferOut, bufSize, soundBank, userSoundNumber);

	return bufSize;
}
	
int EntitySysexMode::saveGlobalToMidi(bool stored) {
	
	int bufSize = global->pack(midiBufferOut);

	if(stored) {
		
		midi->newSysexStoredGlobalEvent(midiBufferOut, bufSize);

		panel->print(0, 0, "MIDI Sysex      ");
		panel->print(1, 0, "Saved Config    ");

		rtems_task_wake_after(500); // So user can see message
	}

	else	{
		
		midi->newSysexEditedGlobalEvent(midiBufferOut, bufSize);
		rMode = RETURN;
	}

	return bufSize;
}

void EntitySysexMode::saveBankToMidi() { 

	int bank = mBuffer[1];

	int bankSize = soundManager->getBankSize();
	int firstSound = bank * bankSize;
	int lastSound = ((bank+1) * bankSize) - 1;

	for(int s = firstSound; s <= lastSound; s++) {
		
		waitForMidi(saveSoundToMidi(s, true));
	}
}

void EntitySysexMode::saveCurrentBankToMidi() { 

	int soundNumber = soundManager->getNumber();
	int bankSize = soundManager->getBankSize();
	int bank = soundNumber / bankSize;

	int firstSound = bank * bankSize;
	int lastSound = ((bank+1) * bankSize) - 1;

	for(int s = firstSound; s <= lastSound; s++) {
		
		waitForMidi(saveSoundToMidi(s, true));
	}
}

void EntitySysexMode::saveAllSoundsToMidi() { 

	for(int s = 0; s <= soundManager->getMaxWritableNumber(); s++) {
		
		waitForMidi(saveSoundToMidi(s, true));
	}
}

void EntitySysexMode::saveAllToMidi() { 

	saveAllSoundsToMidi();
	waitForMidi(saveGlobalToMidi(true));
}

void EntitySysexMode::waitForMidi(int nBytes) {

	// This midi delay limits the rate at which midi
	// sysex events are generated, for receipients
	// with limited buffer capacity.

	rtems_task_wake_after(nBytes / 2); 

	// Clear panel events that build up during 
	// transmission or reception
	
	clearQueue(ctrq);
}

int EntitySysexMode::loadMidiBufferIn() {

	int readPos      = ((unsigned char)mBuffer[1] << 16)
	      	         + ((unsigned char)mBuffer[2] << 8)
	      	         +  (unsigned char)mBuffer[3];

	int bufSize      = ((unsigned char)mBuffer[4] << 8)
	      	         +  (unsigned char)mBuffer[5];

	midi->sysexFlashBufferSetReadPos(readPos);

	for(int x = 0; x < bufSize; x++) midiBufferIn[x] = midi->readSysexFlashBuffer();

	return bufSize;
}

void EntitySysexMode::autoExit() {

	panel->delay(AUTO_EXIT, 1000, 0);
}

void EntitySysexMode::displayUpdate() { }

