;********************************************************************************
;* E N T I T Y 									*
;********************************************************************************
;* File: dsp/interrupts.asm							*
;*                                                                         	*
;* Copyright (C) A.R.Hardie-Bick 2004-2005                                      *
;*										*
;* This program is free software; you can redistribute it and/or modify		*
;* it under the terms of the GNU General Public License as published by		*
;* the Free Software Foundation; either version 2 of the License, or (at	*
;* your option) any later version.						*
;*										*
;* This program is distributed in the hope that it will be useful, but		*
;* WITHOUT ANY WARRANTY; without even the implied warranty of			*
;* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU		*
;* General Public License for more details.					*
;*										*
;* You should have received a copy of the GNU General Public License		*
;* along with this program; if not, write to the Free Software			*
;* Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.			*
;********************************************************************************

;--------------------------------------------------------------------------------
;	Short Interrupts
;--------------------------------------------------------------------------------

	org	P(0):VecHwReset

				BRA	>Start

	org	P(0):VecHostCommandDefault

HV_PARAMETER			JSR	>SelectParameter
HV_PARAMETER_DATA		JSR	>UpdateParameterData
HV_SET_SMOOTH_EDIT		JSR	>SetSmoothEdit
HV_SET_DIRECT_EDIT		JSR	>SetDirectEdit

HV_CTR1				MOVEP 	X:<<HRX,Y:globalCtr1
HV_CTR2				MOVEP 	X:<<HRX,Y:globalCtr2
HV_CTR3				MOVEP 	X:<<HRX,Y:globalCtr3
HV_MMOD				MOVEP 	X:<<HRX,Y:globalMmod
HV_PITCH_SHIFT			MOVEP 	X:<<HRX,Y:globalPitchShift
HV_MASTER_VOL			MOVEP 	X:<<HRX,Y:globalMasterVol
HV_MASTER_TUNE			MOVEP 	X:<<HRX,Y:globalMasterTune
HV_MIDI_PAN			MOVEP	X:<<HRX,Y:globalMidiPan

HV_READ_FRAG_MIPS		JSR	>GetMaxFragMips
HV_READ_TOTAL_MIPS		JSR	>GetMaxTotalMips

HV_READ_MAX_IN_LEVELS		JSR	>GetMaxInLevels
HV_READ_MAX_OUT_LEVELS		JSR	>GetMaxOutLevels

HV_VOICE_ASSIGN			MOVEP	X:<<HRX,Y:voiceAssign
HV_VOICE_ON			JSR	>VoiceOn
HV_VOICE_ON_SMOOTH		JSR	>VoiceOnSmooth
HV_VOICE_OFF			JSR	>VoiceOff
HV_VOICE_MUTE			JSR	>VoiceMute
HV_VOICE_SHIFT			JSR	>VoiceShift
HV_ACTIVATE_NEW_SOUND		JSR	>ActivateNewSound

HV_TRANSFER_ADDRESS		MOVEP	X:<<HRX,Y:transferAddress
HV_TRANSFER_SIZEM1		MOVEP	X:<<HRX,Y:transferSizem1

HV_TRANSFER_XDATA		JSR	>TransferXData
HV_TRANSFER_YDATA		JSR	>TransferYData
HV_TRANSFER_PDATA		JSR	>TransferPData

HV_MINIMISE_DSP			JSR	>MinimiseDSP
HV_MAXIMISE_DSP			JSR	>MaximiseDSP
HV_SOUND_PENDING		JSR	>SoundPending
HV_READY			JSR	>HostReady
HV_INHIBIT			JSR	>HostInhibit

HV_READ_DEBUG_DATA		JSR	>GetDebugData
HV_RESET_DEBUG_ADDRESS		JSR	>ResetDebugAddress

HV_MIDI_CLOCK			JSR	>MidiClock
HV_MIDI_SYNC_SUSTAIN_ON		JSR	>MidiSyncSustainOn
HV_MIDI_SYNC_SUSTAIN_OFF	JSR	>MidiSyncSustainOff
HV_MIDI_SYNC_PHASE		JSR	>MidiSyncPhase

;--------------------------------------------------------------------------------
;	Long Interrupts
;--------------------------------------------------------------------------------

	org	P(2):		; Use (cached) external memory

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - -  
; 	Reset
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - -  

Start	BSR	Initialise
	JMP	MainLoop

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - -  
; 	Receive event data from host
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - -  

VoiceOn	PUSHXYABR0

	MOVE			X:KFFFFFF,M0
	
	; The voice number, and VDIF and STDF values, are in Y:voiceAssign
	:
	; MS byte : mid byte : LS byte
	;   STDF  :   VDIF   :  voice    (all these values are 0..127)
	;	
	; The velocity and pitch (signed 16-bit cents), are in HRX
	;
	; MS byte :      LS word      
	; velocty :       pitch		 (velocity is 0..127)	    

	; First identify the voice number in the LS byte of voiceAssign

	CLR	B		Y:voiceAssign,A2
	MOVE			#vAssignState,B0	; keep LS byte
	MOVE			A2,X1
	ASL	B		#>voiceStateSize,X0

	; Get the base address for the assigned voice's state

	MAC	X0,X1,B
	ASR	B					; Compensate fractional multiply

	MOVEP 			X:<<HRX,A		; [+P]
	MOVE			B0,R0
	MOVE			Y:voiceAssign,B
	NOP
	NOP
	
	; Set voice note on state

	MOVE			X:K1,X0			; 1 = note on
	MOVE			X0,Y:(R0)+		; iOnOff
	MOVE			B,Y:(R0)+		; iVoice

	; Get pitch and velocity from note event data

	ASR	#16,A,B
	AND	#>$FF0000,A

	MOVE			B0,Y:(R0)+		; iPitch
	MOVE			A,Y:(R0)+		; iVelocity	

	PULLXYABR0
	
	RTI

VoiceOnSmooth

	; This is used for monophonic or poly-portamento playing style, where
	; there should be no zeroing of the envelope level when it is put 
	; into its attack phase.

	PUSHXYABR0

	MOVE			X:KFFFFFF,M0
	
	; The voice number, and VDIF and STDF values, are in Y:voiceAssign
	:
	; MS byte : mid byte : LS byte
	;   STDF  :   VDIF   :  voice    (all these values are 0..127)
	;	
	; The velocity and pitch (signed 16-bit cents), are in HRX
	;
	; MS byte :      LS word      
	; velocty :       pitch		 (velocity is 0..127)	    

	; First identify the voice number in the LS byte of voiceAssign

	CLR	B		Y:voiceAssign,A2
	MOVE			#vAssignState,B0	; keep LS byte
	MOVE			A2,X1
	ASL	B		#>voiceStateSize,X0

	; Get the base address for the assigned voice's state

	MAC	X0,X1,B
	ASR	B					; Compensate fractional multiply

	MOVEP 			X:<<HRX,A		; [+P]
	MOVE			B0,R0
	MOVE			Y:voiceAssign,B
	NOP
	
	; Set voice note on state

	MOVE			X:K2,X0			; 2 = note on smooth
	MOVE			X0,Y:(R0)+		; iOnOff
	MOVE			B,Y:(R0)+		; iVoice

	; Get pitch and velocity from note event data

	ASR	#16,A,B
	AND	#>$FF0000,A

	MOVE			B0,Y:(R0)+		; iPitch
	MOVE			A,Y:(R0)+		; iVelocity	

	PULLXYABR0
	
	RTI

VoiceShift

	; Unlike VoiceOn, all we are doing here is changing the pitch of an
	; already-sounding voice. 

	PUSHXYABR0

	MOVE			X:KFFFFFF,M0
	
	; The voice number, and VDIF and STDF values, are in Y:voiceAssign
	:
	; MS byte : mid byte : LS byte
	;   STDF  :   VDIF   :  voice    (all these values are 0..127)
	;	
	; However, STDF and VDIF will already have been correctly set 
	; for this voice during a preceding VoiceOn (the voice cannot
	; shift pitch, and be heard, unless it is already on), so
	; we do not use these items.
	;
	; The pitch (signed 16-bit cents) is in HRX
	;
	; MS byte :      LS word      
	; velocity:       pitch		 (velocity is ignored)

	; First identify the voice number in the LS byte of voiceAssign

	CLR	B		Y:voiceAssign,A2
	MOVE			#vAssignState+iPitch,B0	; keep LS byte
	MOVE			A2,X1
	ASL	B		#>voiceStateSize,X0

	; Get the base address for the assigned voice's state

	MAC	X0,X1,B
	ASR	B					; Compensate fractional multiply

	MOVEP 			X:<<HRX,A		; [+P]
	MOVE			B0,R0
	NOP
	
	; Get pitch from note event data

	ASR	#16,A,B
	NOP
	MOVE			B0,Y:(R0)+		; iPitch

	PULLXYABR0
	
	RTI

VoiceOff
	
	PUSHXYABR0

	MOVE			X:KFFFFFF,M0
	
	; The voice number (0..n) is in HRX. No other data.

	MOVEP 			X:<<HRX,X1
	CLR	B		#>voiceStateSize,X0
	MOVE			#vAssignState,B0	

	; Get the base address for the voice's state

	ASL	B
	MAC	X0,X1,B
	ASR	B					; Compensate fractional multiply
	NOP
	MOVE			B0,R0
	NOP
	NOP
	
	; Set voice note off state

	MOVE			X:KFFFFFF,X0		; -1 = note off
	MOVE			X0,Y:(R0)+		; iOnOff

	PULLXYABR0
	
	RTI

VoiceMute

	PUSHXYABR0

	MOVE			X:KFFFFFF,M0
	
	; The voice number (0..n) is in HRX. No other data.

	MOVEP 			X:<<HRX,X1
	CLR	B		#>voiceStateSize,X0
	MOVE			#vAssignState+vMute,B0	

	; Get the base address for the voice's state

	ASL	B
	MAC	X0,X1,B
	ASR	B			
	NOP
	MOVE			B0,R0
	NOP
	NOP
	
	; Set voice mute

	MOVE			X:KFFFFFF,X0
	MOVE			X0,Y:(R0)+		; vMute = on

	PULLXYABR0
	
	RTI

MinimiseDSP

	PUSHX

	MOVE			#>MinimalParameterGenerator,X0
	MOVE			X0,Y:sGenVector

	PULLX

	RTI

MaximiseDSP

	PUSHX

	BCLR	#HCR_HF3,X:HCR

	MOVE			#>ParameterGenerator,X0
	MOVE			X0,Y:sGenVector

	PULLX

	RTI

ActivateNewSound

	PUSHX

	MOVE			#>FadeForNewSound,X0
	MOVE			X0,Y:sGenVector

	PULLX

	RTI

SoundPending

	BSET	#soundPending,Y:flags
	BCLR	#editParaValid,Y:flags
	RTI

HostReady

	BSET	#readyFlag,Y:flags
	RTI

HostInhibit

	BCLR	#readyFlag,Y:flags
	RTI

ResetDebugAddress

	PUSHX

	MOVE			Y:debugStartAddress,X0
	MOVE			X0,Y:debugAddress

	PULLX

	RTI

GetDebugData	

	PUSHXAR0

	MOVE			Y:debugAddress,R0
	MOVE			Y:debugAddressIncrement,N0
	MOVE			X:KFFFFFF,M0
	NOP
	BTST	#0,Y:debugSpace			
	MOVE			X:(R0),X0	; data from X:
	MOVE			Y:(R0)+N0,A	; data from Y:
	TFR	X0,A	IFCC			; select memory space

	MOVE			R0,Y:debugAddress		
	MOVEP			A,X:<<HTX

	PULLXAR0

	RTI

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
; See midiLFOSync.txt for use of midi realtime signals
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

MidiClock

	BSET	#midiClockSignal,Y:iSignals
	RTI

MidiSyncSustainOn

	BSET	#midiSyncSustain,Y:iFlipFlops
	RTI

MidiSyncSustainOff

	BCLR	#midiSyncSustain,Y:iFlipFlops
	RTI

MidiSyncPhase

	PUSHAB

	MOVEP 			X:<<HRX,A

	ADD	#$40,A					; LFO waveform starts at minimum phase
	ASR	#7,A,B
	NOP
	MOVE			B0,Y:globalMidiSyncPhase

	BSET	#midiSyncPhaseUpdate,Y:iFlipFlops	; Cleared by midi clock processing

	PULLAB

	RTI

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - -  
; 	DMA block transfer from host
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - -  

INIT_DCR2 EQU $089AC0	; Block transfers use DMA channel 2

	; [23] 		DE 		0 	DMA disabled (until needed)
	; [22] 		DIE 		0 	DMA interrupts disabled
	; [21:19] 	DTM2:DTM0 	001 	by request, DE=0 when done, words
	; [18:17] 	DPR1:DPR0 	00 	DMA priority level 0 (lowest)
	; [16] 		DCON 		0 	continuous mode off
	; [15:11] 	DRS4:DRS0 	10011 	request on HRDF=1
	; [10] 		D3D 		0 	non-3D mode
	; [9:7] 	DAM5:DAM3 	101 	dest. is linear access, post +1
	; [6:4] 	DAM2:DAM0 	100 	source is linear access, no update
	; [3:2] 	DDS1:DDS0 	00 	X memory DMA destination (default)
	; [1:0] 	DSS1:DSS0 	00 	X memory DMA source

TransferXData

	MOVEP	#>HRX,X:<<DSR2 			; DMA source address
	MOVEP 	Y:transferAddress,X:<<DDR2 	; DMA destination (to be set by host command)
	MOVEP 	Y:transferSizem1,X:<<DCO2 	; Sound data size - 1 (to be set by host command)
	MOVEP 	#>INIT_DCR2,X:<<DCR2 		; Initial DMA control register value
	BSET	#DCR_DE,X:<<DCR2		; DMA2 ready for sound transfer from HOST
	RTI	

TransferYData

	MOVEP	#>HRX,X:<<DSR2 			; DMA source address
	MOVEP 	Y:transferAddress,X:<<DDR2 	; DMA destination (to be set by host command)
	MOVEP 	Y:transferSizem1,X:<<DCO2 	; Sound data size - 1 (to be set by host command)
	MOVEP 	#>INIT_DCR2,X:<<DCR2 		; Initial DMA control register value
	BSET	#DCR_DDS0,X:<<DCR2		; Destination Y
	BSET	#DCR_DE,X:<<DCR2		; DMA2 ready for sound transfer from HOST
	RTI	

TransferPData

	MOVEP	#>HRX,X:<<DSR2 			; DMA source address
	MOVEP 	Y:transferAddress,X:<<DDR2 	; DMA destination (to be set by host command)
	MOVEP 	Y:transferSizem1,X:<<DCO2 	; Sound data size - 1 (to be set by host command)
	MOVEP 	#>INIT_DCR2,X:<<DCR2 		; Initial DMA control register value
	BSET	#DCR_DDS1,X:<<DCR2		; Destination P
	BSET	#DCR_DE,X:<<DCR2		; DMA2 ready for sound transfer from HOST
	RTI	

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - -  
;	Parameter edit
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - -  

SelectParameter

	PUSHXYAB				; push main registers on stack

	; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	; The idea here is that the parameter is being defined as a parameter number 
	; (ie 0..127 in bits 0..7) and a matrix source (0 for base value, 0..n for matrix 
	; depths in bits 8..15). This information is currently in X:<<HRX. From this, an 
	; offset is calculated into Y:soundData or Y:soundTransferBuffer.

	MOVEP			X:<<HRX,A	; parameter index (0..127) in bits 0-7, source (0..n) in bits 8-15
	ASR	#32,A,B				; source (0..n)
	AND	#>$0000FF,A			; keep bits 0-7
	MOVE			A,Y0		; parameter number (0..127)

	MPYI	#>nMatrixSources+1,Y0,A		; Calculate offset from base address for parameters
	ASR	A				; compensate result, base address if writing live sound parameters
	ADD	B,A				; include source with offset
	NOP
	MOVE			A0,Y:parameter	; Offset from Y:soundData or Y:soundTransferBuffer

	MOVE			#>nDirect,Y1 		; Threshold to identify product-of-product matrix type
	TFR	Y0,A		#>dummyPnData,X0	; Parameter number (0..127)	In case Pn translation not required
	SUB	Y1,A		X0,Y:parameterPn	; 				Default is fake Pn destination
	JLT	_done

	; Product-of-products type parameter

	TST	B		A,Y0			; was the para a base value?	p-of-p parameter index (0..28)
	JEQ	_done					; Base value being edited, so no need to translate matrix depth

	MPYI	#>nMatrixSources,Y0,A			; address offset (soundPnData array doesn't include base values)
	ASR	A					; compensate fractional multiply
	ADD	A,B					; add source offset (0..n-1)
	MOVE			#>soundPnData-1,A	;				-1 because source values are 1..n here
	MOVE			B0,B
	ADD	A,B
	NOP	
	MOVE			B,Y:parameterPn

_done	

	BCLR	#editParaValid,Y:flags			; made valid when first new data arrives
	BSET	#newPara,Y:flags

	PULLXYAB

	RTI

UpdateParameterData

	MOVEP	X:<<HRX,Y:paraData
	BSET	#editParaValid,Y:flags

	RTI

SetSmoothEdit

	BSET	#smoothEdit,Y:flags
	RTI

SetDirectEdit

	BCLR	#smoothEdit,Y:flags
	RTI

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - -  
; 	Transfer real time data to host
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - -  

GetMaxOutLevels	

	PUSHXYAB

	; Put left in LS 12 bits, right in MS 12 bits

	MOVE 			Y:maxOutL,A
	MOVE			Y:maxOutR,B

	ABS	A		X:K7FFFFF,Y0
	TES	Y0,A				; in case of ABS(-1)
	LSR	#11,A
	ABS	B		#>$FFF000,X1
	TES	Y0,B
	LSL	B		A,X0
	AND	X1,B		#0,Y0
	OR	X0,B		Y0,Y:maxOutL
	MOVE			Y0,Y:maxOutR
	MOVEP			B1,X:<<HTX

	PULLXYAB

	RTI	

GetMaxInLevels

	PUSHXYAB
	
	; Put left in LS 12 bits, right in MS 12 bits
	
	MOVE 			Y:maxInL,A
	MOVE			Y:maxInR,B

	ABS	A		X:K7FFFFF,Y0
	TES	Y0,A				; in case of ABS(-1)
	LSR	#11,A
	ABS	B		#>$FFF000,X1
	TES	Y0,B
	LSL	B		A,X0
	AND	X1,B		#0,Y0
	OR	X0,B		Y0,Y:maxInL
	MOVE			Y0,Y:maxInR
	MOVEP			B1,X:<<HTX

	PULLXYAB

	RTI	

GetMaxTotalMips

	PUSHX

	MOVEP 	Y:maxTotalMips,X:<<HTX
	MOVE	#0,X0
	MOVE	X0,Y:maxTotalMips

	PULLX

	RTI	

GetMaxFragMips

	PUSHX

	MOVEP 	Y:maxFragMips,X:<<HTX
	MOVE	#0,X0
	MOVE	X0,Y:maxFragMips

	PULLX

	RTI	

