;********************************************************************************
;* E N T I T Y 									*
;********************************************************************************
;* File: dsp/memory.asm								*
;*                                                                         	*
;* Copyright (C) A.R.Hardie-Bick 2004-2005                                      *
;*										*
;* This program is free software; you can redistribute it and/or modify		*
;* it under the terms of the GNU General Public License as published by		*
;* the Free Software Foundation; either version 2 of the License, or (at	*
;* your option) any later version.						*
;*										*
;* This program is distributed in the hope that it will be useful, but		*
;* WITHOUT ANY WARRANTY; without even the implied warranty of			*
;* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU		*
;* General Public License for more details.					*
;*										*
;* You should have received a copy of the GNU General Public License		*
;* along with this program; if not, write to the Free Software			*
;* Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.			*
;********************************************************************************

; Universal Constants

N_CYCLES	equ	64	; Ratio of sample rate to parameter rate
N_VOICES_MEM	equ	6	; Number of voices to reserve memory for
N_VOICES_PROC	equ	6	; Number of voices to calculate

;------------------------------------------------------------------------------- 
; 				Memory map
;------------------------------------------------------------------------------- 
;
; DRAM usage (X: Y: and P: map to same physical locations in DRAM):
;
;		$800000
;		   .		Y: Large LUTs
;		$405000		Y: Medium-size LUTs
;		$404000		Y: Several 129-step LUTs (see lut.asm)
;		$400000		P: Program space (allow 16K)
;
; On-chip memory organisation:
;
;		 Addr.		    P:		    X:		    Y:
;
;				----------
;		  		 1K cache
;				----------
;		$000C00			    	----------	---------- 
;
;		$000800	    		
;					     	 3K words	 3K words
;		$000400	    	----------		
;				 1K words
;		$000000	    	----------	----------    	----------
;
; On-chip X memory usage:
;
;	        $000C00
;		$000B80		matrix destinations 			 (87)
;		$000A00		Voice-specific data			 (..)
;		   .
;		$000900	
;		$000800		Noise samples				(256)
;		$000700		soundPnData				(256)
;		$000600		Various small LUTs			(256)
;		$000400		Sine wave LUT (base values)		(512)
;		$000200		DAC/ADC double buffers			(512)
;		$000040		L: data
;		$000000		Short addressed data			 (64)
;
; On-chip Y memory usage:
;	
;		$000600		Sound data (rest of on-chip YRAM)      (1536)
;		$000400		Sine wave LUT (differences)		(512)
;		$000200		Y: voice state data			(512)
;		$000040		L: data
;		$000000		Short addressed data			 (64)
;

;------------------------------------------------------------------------------- 
;			ON-CHIP SRAM
;------------------------------------------------------------------------------- 

		org	L:$000000

gp			ds	1	; General purpose
gp1			ds	1
gp2			ds	1

vAccL			ds	1	; Voice accumulator left
vAccR			ds	1	; Voice accumulator right

DFBL			ds	1	; Output dither feedback left
DFBR			ds	1	;   "      "       "     right

filteredMCR48		ds	1	; High precision filter for midi clock rate

external__		ds	1	; External input left (X) and right (Y)
external		equ	@CVS(N,external__) ; X,Y or L addressing
externalRight		equ	@CVS(Y,external__) ; externalLeft has its own Y: location

_oscPhzCtrl		ds	1	; See oscillator code 
oscPhzCtrl		equ	@CVS(N,_oscPhzCtrl) ; X,Y or L addressing

shortSpace		equ	*	; Mark end of L: memory to avoid X & Y reuse

		org	X:shortSpace

rng			ds	1	; Random number generator (linear congruential) state

noisePtr		ds	1	; Pointer (for Rx) to next noise sample in fast gaussian noise array
noiseInc		ds	1	; Prime number increment (for Nx) for reading fast noise array 
					; (also use 250 (251 is prime), in modifier register Mx)
DACBufferPtr		ds	1
ADCBufferPtr		ds	1

preMixGain		ds	1	; Temp var holding overdrive-derived gain for pre-filter mix levels
voiceStateBaseAddress	ds	1	; Pointer to base address of current voice's assign, LFO and ENV states
voiceMatrixBaseAddress	ds	1	; Pointer to base address of current voice's matrix sources

voiceInterpolatedParasBaseAddress	; Pointer to base address of current voice's interpolated parameters
			ds	1
voiceNISampleParasBaseAddress		; Pointer to base address of current voices's non-interpolated parameters	
			ds	1	
voiceStatePtr		ds	1	; Helper pointer for address of next state in voice to look at
LFO56vLFOStatePtr	ds	1	; Pointer to LFO states of voice from which LFO5&6 paras shall derive

; Constants

KFFFFFF			dc	$FFFFFF
K800000			dc	$800000
K7FFFFF			dc	$7FFFFF
K3FFFF			dc	$3FFFF
KFFFF			dc	$FFFF
K8000			dc	$8000
KFF			dc	$FF
K7F			dc	$7F
K1			dc	$1
K2			dc	$2
K8768T256		dc	8768*256
K4800			dc	4800
K806			dc	806

KP999813		dc	0.999813
KP003333333		dc	0.003333333
KP000001		dc	0.000001
KP45			dc	0.45
KP55			dc	0.55

KP03505625		dc	0.03505625
KP04286875		dc	0.04286875
KP08193125		dc	0.08193125

KP2F64K			dc	P2F64K+$8000
KP2I64K			dc	P2I64K+$8000
Ksine512		dc	sine512
Ksine256K		dc	sine256K
KnoiseMask		dc	$1D872B			; See "Musical Applications of Microprocessors", P485
KModSources		dc	ModSources
KfltCoeffLUT12dB	dc	fltCoeffLUT12dB+$40000
KfltOvLUT64K		dc	fltOvLUT64K
Klim256LUT64K		dc	lim256LUT64K
Kexp1LUT1K		dc	exp1LUT1K
Kexp2LUT1K		dc	exp2LUT1K
KENVstaccatoPreAttackPhase1
			dc	ENVstaccatoPreAttackPhase1
KENVattackPhase		dc	ENVattackPhase
KENVreleasePhase	dc	ENVreleasePhase

KcyclesAvailable	dc	@CVI(@CVF(N_CYCLES)*100E6/48000.0) ; eg 133334 

KDACBuffer1		dc	DACBuffer1
KDACBuffer2		dc	DACBuffer2
KADCBuffer1		dc	ADCBuffer1
KADCBuffer2		dc	ADCBuffer2
KendDACBuffer		dc	endDACBuffer

	IF		@CVI(*)>$40
	FAIL		'X: memory definition overlapping L:'
	ENDIF
	
		org	Y:shortSpace

one			dc	1	; Easy access to this number

; Constants (placed here due to lack of short X: memory)

Knoise1M		dc	noise1M
KH1000			dc	$1000

; Flags

flags			ds	1

wasInRelease		equ	0	; Used to identify staccato 

skipCodeFragment	equ	1	; Measuring CPU usage for section of code
readyFlag		equ	2	; Host dialogue during initialisation
soundPending		equ	3	; Indicate a new sound is pending (causes edits to edit pending sound data)
newPara			equ	4	; Indicate parameter has changed, (so para filter needs initialising)
smoothEdit		equ	5	; Cause parameter edit changes to be filtered smooth
editParaValid		equ	6	; Indicates parameter address and data are valid
abruptParas		equ	7 	; Indicates to use abrupt parameter interpolation

LFOsHold		equ	9	; Local flag used in LFO generation 

; Signals (interrupt-triggered flags that reset after one parameter cycle)

iSignals		ds	1	; Flags set by interrupt
rSignals		ds	1	; Registered interrupt flags (stable for parameter cycle)

midiClockSignal		equ	0	; Midi clock signal

; Flip Flops (persistent flags that have to be explicitely cleared)

iFlipFlops		ds	1	; Flipflops modified by interrupt
rFlipFlops		ds	1	; Registered flipflops (stable for parameter cycle)

midiSyncSustain		equ	0	; Cause midi synced LFOs to hold frequency
midiSyncPhaseUpdate	equ	1	; Update phase for synced LFOs on next midi clock

; Audio

limMaxLR		ds	1
maxInL			ds	1	
maxInR			ds	1
maxOutL			ds	1
maxOutR			ds	1

externalLeft		ds	1

; Voice-related 

mostRecentVoiceIdx	ds	1	; Loop counter index of voice most recently assigned
oscPhSyn		ds	1 	; Temporary flag indicating audio oscillator reset
vOnOff			ds	1	; Registered copy of iOnOff valid through voice loop

; Voice assignment register

voiceAssign		ds	1	; An additional word of data for host interrupt routines

; Global RTC

globalCtr1		ds	1
globalCtr2		ds	1
globalCtr3		ds	1
globalMmod		ds	1

globalMasterVol		ds	1
globalMasterTune	ds	1
globalPitchShift	ds	1
globalMidiPan		ds	1

globalMidiSyncPhase	ds	1	; Set arbitrary phase for midi sync'd LFOs

; Filtered parameters

filteredMidiPan		ds	1
filteredMidiPitchShift	ds	1

; Parameter editing

paraData		ds	1	; Parameter data
parameter		ds	1	; Offset in Y:soundData or Y:soundTransferBuffer of parameter to edit
parameterPn		ds	1	; Address (possibly fake) to somewhere in X:soundPnData 

; Midi clock timing

nIterations		ds	1	; Number of parameter engine iterations per midi clock
midiClockRate		ds	1	; 1 / nIterations
filteredMCR		ds	1	; Filtered midi clock rate, value used during midi sync sustain

; Render Control

sGenVector		ds	1	; Sound generation vector

; MIPS benchmarks

maxTotalMips		ds	1
maxFragMips		ds	1
totalMips		ds	1

; Host block DMA

transferAddress		ds	1
transferSizem1		ds	1

; Debug

debugSpace		ds	1
debugAddress		ds	1
debugStartAddress	ds	1
debugAddressIncrement	ds	1

; Initialisation

F2IbShift		ds	1	; Block shift for F2Ib LUT
F2IcShift		ds	1	; Block shift for F2Ic LUT

; Scratch Pad

test1			ds	1
test2			ds	1
test3			ds	1

	IF		@CVI(*)>$40
	FAIL		'Y: memory definition overlapping L:'
	ENDIF
	
		org 	L:$000040

stack			ds	48	; (see macro.asm)

;------------------------------------------------------------------------------- 
; Per-voice non-interpolated parameters used in sample generation
;------------------------------------------------------------------------------- 

; Nomenclature: A 'z' is sometimes used to distinguish between parameters 
; used in sample generation, and an output from the parameter matrix from 
; which it is derived. For example, O1fmsrc is a matrix output, whereas
; O1zfmsrc is the translated parameter used in sample generation.

; Defined in approximate order of use in the code, to simplify address increments

; Non-interpolated paras are in X, 24-bit audio samples usually in Y, other stuff in L 

O1zfmsrc		equ	0	; X  *	address of FM source for oscillator
O1a			equ	0	; Y  	Oxa output
O1phaseAcc		equ	1	; L (*)	48-bit phase
O1freq			equ	2	; L  *	48-bit frequency
O1zfm			equ	3	; X  *	FM depth (external)
O1PM			equ	3	; Y  *	PM (brightness)
O1frqshift		equ	4	; X  *	0 or 1, selecting saw or square
O1b			equ	4	; Y	Oxb output
O1			equ	5	; Y	Ox output

O2zfmsrc		equ	6	; X  *	address of FM source for oscillator
O2a			equ	6	; Y	Oxa output
O2phaseAcc		equ	7	; L (*)	48-bit phase
O2freq			equ	8	; L  *	48-bit frequency
O2zfm			equ	9	; X  *	FM depth (external)
O2PM			equ	9	; Y  *	PM (brightness)
O2frqshift		equ	10	; X  *	0 or 1, selecting saw or square
O2b			equ	10	; Y	Oxb output
O2			equ	11	; Y	Ox output

O3zfmsrc		equ	12	; X  *	address of FM source for oscillator
O3a			equ	12	; Y	Oxa output
O3phaseAcc		equ	13	; L (*)	48-bit phase
O3freq			equ	14	; L  *	48-bit frequency
O3zfm			equ	15	; X  *	FM depth (external)
O3PM			equ	15	; Y  *	PM (brightness)
O3frqshift		equ	16	; X  *	0 or 1, selecting saw or square
O3b			equ	16	; Y	Oxb output
O3			equ	17	; Y	Ox output

RMzsrca			equ	18	; X  *	address of source a for RM
RMzsrcb			equ	19	; X  *	address of source b for RM
RM			equ	18	; Y	Ring Modulator output

NS			equ	19	; Y	Noise source (for modulation)

fltDaL			equ	20	; X	filter states (samples in X)
fltDbL			equ	21	; X
fltDaR			equ	22	; X	Right filter	"
fltDbR			equ	23	; X

fltD0L			equ	20	; Y	filter direct out (zero delay)
fltD1L			equ	21	; Y	filter delayed output
fltD0R			equ	22	; Y	
fltD1R			equ	23	; Y

; * indicates parameter updated by the parameter generator

NISampleParasPerVoice	equ	24	; Non-interpolated samples and parameters per voice

NISampleParas		ds	NISampleParasPerVoice*N_VOICES_MEM

;------------------------------------------------------------------------------- 
; General interpolated parameters
;------------------------------------------------------------------------------- 

interpolatedParas

MVOL			ds	1	; Front panel & midi volume

;------------------------------------------------------------------------------- 
; Per-voice interpolated parameters
;------------------------------------------------------------------------------- 

; These occupy X & Y memory, X contains the base value, Y contains the increment

; Defined in approximate order of use

O1zshape		equ	0	; Phase offset between Oxa and Oxb 
O2zshape		equ	1	; ("phase" parameter)		       
O3zshape		equ	2	;

O1L			equ	3	; O1 level (left)
O2L			equ	4	; O2
O3L			equ	5	; O3
RML			equ	6	; Ring mod

NGLR			equ	7	; Noise level (left and right)
EXLR			equ	8	; External input level

O1R			equ	9	; O1 level (right) 
O2R			equ	10	; O2
O3R			equ	11	; O3
RMR			equ	12	; Ring modulator

fltRs			equ	13	; Filter resonance
fltFa			equ	14	; 1st stage frequency coefficient 
fltFb			equ	15	; 2nd stage frequency coefficient
fltOn			equ	16	; Overdrive parameter n	(non-linearity)

VLL			equ	17	; Voice left, left mix level
VRL			equ	18	; Voice right, left mix level
VLR			equ	19	; Voice left, right mix level
VRR			equ	20	; Voice right, right mix level

interpolatedParasPerVoice	equ	21

vInterpolatedParas	ds	interpolatedParasPerVoice*N_VOICES_MEM

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -  

; Spare i. para in case odd no. of i. paras (interpolated paras are updated in pairs)

extraInterpolatedPara	ds	1	

; Number of parameters to interpolate after sample generation

nInterpolatedParas	equ	1+interpolatedParasPerVoice*N_VOICES_PROC+1

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -  

longSpace		equ	*	

		org	X:longSpace

	IF	@CVI(*)>$200
	FAIL	'X: Memory definition overlapping X:$000200'
	ENDIF

		org	Y:longSpace

;------------------------------------------------------------------------------- 
; Per-voice voice assignment state
;------------------------------------------------------------------------------- 

; Interrupt-generated event values

iOnOff			equ	0		; -ve->off +ve->on 0->no new event
iVoice			equ	1		; STDF:VDIF:voice (voice is redundant) 
iPitch			equ	2		; velocity(8-bits, 0..127):pitch(16-bits)
iVelocity		equ	3

; Registered event data (stable in parameter cycle)

vPitch			equ	4		; derived from iNote
vFilteredPitch		equ	5		; output of portamento filter
vVelocity		equ	6		; also derived from iNote
vVoiceDiff		equ	7		; derived from iVoice
vStackDiff		equ	8		;  	   "

; Others 

vMute			equ	9		; Voice muted if non-zero
vLegato			equ	10		; Derived from SNDportamnto

vAssignStateSize	equ	11 

;------------------------------------------------------------------------------- 
; Per-voice LFO state data
;------------------------------------------------------------------------------- 

vLFOstateOffset		equ	vAssignStateSize

; Voice LFOs (LFOs 1-4) use a subset of the state of sound LFOs (LFOs 5 and 6)

LFO_phase		equ	0		; LFO phase (24 bit waveform precursor, sawtooth)
LFO_delayLevel		equ	1		; State of amplitude delay 
LFO_data		equ	2		; Current data value (ie, LFO waveform before AM)
LFO_initial		equ	3		; Interpolation initial value (ie for pink noise)
LFO_target		equ	4		; Interpolation target value ( " )
LFO_stateA  		equ	5		; State A (eg pink noise filter 1)
LFO_stateB  		equ	6		; State B (eg pink noise filter 2)
LFO_stateC  		equ	7		; State C (eg pink noise filter 3)
LFO_signal		equ	8		; S&H trigger for other LFOs ($000001 = trigger)

vLFOstateSize		equ	9

; Additional sound LFO state offsets (memory not assigned as part of a voice LFO)

LFO_zIncr		equ	9		; (midi sync) phase change required per LFO iteration (freq)
LFO_zPrev		equ	10		; (midi sync) LFO phase at previous midi clock 
LFO_zSync		equ	11		; (midi sync) phase change required over next midi clock cycle
LFO_seq			equ	12		; (midi sync) sequence LUT pointer (see midiLFOsync.txt)
LFO_zStop		equ	13		; (midi sync) LFO phase on midi stop (for midi continue)
LFO_zErrOkay		equ	14		; (midi sync) can inhibit zErr after legitimate phase adjustment
LFO_shAddr		equ	15		; Address of S&H signal (S&H processing different for sLFOs) 
LFO_shType		equ	16		; Flag for indicating different types of S&H trigger (see code)

sLFOstateSize		equ	17

; Delay state for sound LFOs (part of voice state)

LFO5_delayLevel		equ	0
LFO6_delayLevel		equ	1

vLFO56stateSize		equ	2		; LFO5_delay and LFO6_delay states for a voice

; 4 LFOs plus LFO5 and LFO6 delay state in each voice

vLFOstateBlockSize	equ	vLFOstateSize*4+vLFO56stateSize	

;------------------------------------------------------------------------------- 
; Per-voice ENV state data
;------------------------------------------------------------------------------- 

; Envelopes 1 and 2

ENV_phase		equ	0		; Address of subroutine, ie ENVattack, ENVdecay, etc
ENV_level		equ	1

ENVstateSize		equ	2
ENVstateBlockSize	equ	ENVstateSize*2	; 2 ENVs per voice

; The order of usage is: Voice #1: assign, LFO1-6, ENV1,2, Voice #2: assign, LFO1-6, ENV1,2, Voice #3...

; Voice matrix, LFO and Envelope states in Y: memory 

voiceStateSize		equ	vAssignStateSize+vLFOstateBlockSize+ENVstateBlockSize

voiceState		ds	voiceStateSize*N_VOICES_MEM

vAssignState		equ	voiceState
vLFOstate		equ	vAssignState+vAssignStateSize
vLFO56state		equ	vLFOstate+vLFOstateSize*4
vENVstate		equ	vLFO56state+vLFO56stateSize

;------------------------------------------------------------------------------- 
; Additional non-voice state data
;------------------------------------------------------------------------------- 

; Sound LFO state

sLFOs			ds	sLFOstateSize*2	; 2 sound LFOs (common to all voices)

; Sound LFO sources

sLFOsources		ds	4		; direct & PoP for LFO5 and LFO6

;------------------------------------------------------------------------------- 

	IF	@CVI(*)>$400
	FAIL	'Y: Memory definition overlapping L:$000400'
	ENDIF

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -  

		org	X:$000200	; align these buffers to ^2

DACBuffer1		ds	2*N_CYCLES
DACBuffer2		ds	2*N_CYCLES
endDACBuffer		equ	*

ADCBuffer1		ds	2*N_CYCLES
ADCBuffer2		ds	2*N_CYCLES
endADCBuffer		equ	*

	IF	@CVI(*)>$400
	FAIL	'X: Memory definition overlapping L:$000400'
	ENDIF

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -  

	org	L:$000400		; This sine wave LUT is stored in long memory

sine512			ds	512	; Base values in X:, deltas/2 in Y:

	org	Y:$000600

initSpaceY 		; 	Also used as workspace during initialisation
soundData		ds	soundDataSize

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -  

	org	X:$000600

;------------------------------------------------------------------------------- 
; Various on-chip LUTs
;------------------------------------------------------------------------------- 

ModSources		dc	1,O1		; 1st arg = 1 if the address is relative
			dc	1,O2
			dc	1,O3
			dc	1,NS
			dc	1,RM
			dc	1,fltD0L
			dc	1,fltD0R
			dc	0,externalLeft	; 1st arg = 0 if address is absolute
			dc	0,externalRight

LFOwaves		dc	LFOtriangle
			dc	LFOsine
			dc	LFOsawtooth
			dc	LFOsawExp
			dc	LFOsquare
			dc	LFOpulse
			dc	LFOrandom
			dc	LFOgaussian
			dc	LFOpink
			dc	LFOrandom_S
			dc	LFOgauss_S
			dc	LFOpink_S

; Voice LFOs have a larger set of possible synchronisations than sound LFOs

vLFOsyncs		dc	vLFOnoSync,0
	
			; Restart LFO waveform on new note
			
			dc	vLFOsyncNote,0 	

			; Sync to any of LFO1-6 in antiphase or quadrature
			
			dc	vLFOsyncAntiInVoice,vLFOstateOffset+vLFOstateSize*0+LFO_phase
			dc	vLFOsyncQuadInVoice,vLFOstateOffset+vLFOstateSize*0+LFO_phase	
			dc	vLFOsyncAntiInVoice,vLFOstateOffset+vLFOstateSize*1+LFO_phase	
			dc	vLFOsyncQuadInVoice,vLFOstateOffset+vLFOstateSize*1+LFO_phase	
			dc	vLFOsyncAntiInVoice,vLFOstateOffset+vLFOstateSize*2+LFO_phase	
			dc	vLFOsyncQuadInVoice,vLFOstateOffset+vLFOstateSize*2+LFO_phase	
			dc	vLFOsyncAntiInVoice,vLFOstateOffset+vLFOstateSize*3+LFO_phase	
			dc	vLFOsyncQuadInVoice,vLFOstateOffset+vLFOstateSize*3+LFO_phase	
			dc	vLFOsyncAntiInSound,sLFOs+sLFOstateSize*0+LFO_phase	
			dc	vLFOsyncQuadInSound,sLFOs+sLFOstateSize*0+LFO_phase	
			dc	vLFOsyncAntiInSound,sLFOs+sLFOstateSize*1+LFO_phase	
			dc	vLFOsyncQuadInSound,sLFOs+sLFOstateSize*1+LFO_phase	

			; Sample and hold in response to the start of another LFO's cycle
			; or (NSH) reset LFO phase at start of new note, then sample and 
			; hold in response to the start of another LFO's cycle 

			dc	vLFOsyncLFO_S_H_InVoice,vLFOstateOffset+vLFOstateSize*0+LFO_signal	
			dc	vLFOsyncLFO_NSH_InVoice,vLFOstateOffset+vLFOstateSize*0+LFO_signal	
			dc	vLFOsyncLFO_S_H_InVoice,vLFOstateOffset+vLFOstateSize*1+LFO_signal	
			dc	vLFOsyncLFO_NSH_InVoice,vLFOstateOffset+vLFOstateSize*1+LFO_signal	
			dc	vLFOsyncLFO_S_H_InVoice,vLFOstateOffset+vLFOstateSize*2+LFO_signal	
			dc	vLFOsyncLFO_NSH_InVoice,vLFOstateOffset+vLFOstateSize*2+LFO_signal	
			dc	vLFOsyncLFO_S_H_InVoice,vLFOstateOffset+vLFOstateSize*3+LFO_signal	
			dc	vLFOsyncLFO_NSH_InVoice,vLFOstateOffset+vLFOstateSize*3+LFO_signal	
			dc	vLFOsyncLFO_S_H_InSound,sLFOs+sLFOstateSize*0+LFO_signal	
			dc	vLFOsyncLFO_NSH_InSound,sLFOs+sLFOstateSize*0+LFO_signal	
			dc	vLFOsyncLFO_S_H_InSound,sLFOs+sLFOstateSize*1+LFO_signal	
			dc	vLFOsyncLFO_NSH_InSound,sLFOs+sLFOstateSize*1+LFO_signal	

			; Sample and hold on new note

			dc	vLFOsyncNote_S_H,0

; Sound LFOs cannot sync to voice LFOs or new notes, so there is a reduced set of synchronisations
			
sLFOsyncs		dc	sLFOnoSync,0

			; Sync to LFO5 or LFO6 in antiphase or quadrature
	
			dc	sLFOsyncAnti,sLFOs+sLFOstateSize*0+LFO_phase	
			dc	sLFOsyncQuad,sLFOs+sLFOstateSize*0+LFO_phase	
			dc	sLFOsyncAnti,sLFOs+sLFOstateSize*1+LFO_phase	
			dc	sLFOsyncQuad,sLFOs+sLFOstateSize*1+LFO_phase	

			; Sample and hold in response to the start of another LFO's cycle

			dc	sLFOsyncLFO_S_H_InVoice,vLFOstateOffset+vLFOstateSize*0+LFO_signal	
			dc	sLFOsyncLFO_S_H_InVoice,vLFOstateOffset+vLFOstateSize*1+LFO_signal	
			dc	sLFOsyncLFO_S_H_InVoice,vLFOstateOffset+vLFOstateSize*2+LFO_signal	
			dc	sLFOsyncLFO_S_H_InVoice,vLFOstateOffset+vLFOstateSize*3+LFO_signal	
			dc	sLFOsyncLFO_S_H_InSound,sLFOs+sLFOstateSize*0+LFO_signal	
			dc	sLFOsyncLFO_S_H_InSound,sLFOs+sLFOstateSize*1+LFO_signal	
	
			; Sample and hold on new note

			dc	sLFOsyncNote_S_H,0

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -  
; LFO midi sync sequence LUTs

; These sequences are described in midiLFOSync.txt
; and generated by MidiLFOSyncSeq.java. They are
; used for controlling the phase of midi-synced LFOs

LFOS_01_32_1		dc	21,21,22
LFOS_02_24_1		dc	28,28,29,28,29,28,29,28,29
LFOS_03_16_1		dc	42,43,43
LFOS_04_12_1		dc	56,57,57,57,57,57,57,57,57	
LFOS_05_9_1		dc	75,76,76,76,76,76,75,76,76,76,76,76,76,75,76,76,76,76,76,76,75,76,76,76,76,76,76
LFOS_06_8_1		dc	85,85,86
LFOS_07_7_1		dc	97,98,97,98,97,98,97,98,97,98,97,98,97,98,97,98,97,98,97,98,98
LFOS_08_6_1		dc	113,114,114,114,113,114,114,114,114
LFOS_09_5_1		dc	136,137,136,137,136,137,136,137,136,137,136,137,136,137,137
LFOS_10_4_1		dc	170,171,171
LFOS_11_3_1		dc	227,228,227,228,227,228,227,228,228
LFOS_12_2_1		dc	341,341,342
LFOS_13_1_1		dc	682,683,683	
LFOS_14_1_2		dc	1365,1365,1366
LFOS_15_1_3		dc	1820,1820,1821,1820,1821,1820,1821,1820,1821
LFOS_16_1_4		dc	2730,2731,2731
LFOS_17_1_6		dc	3640,3641,3641,3641,3641,3641,3641,3641,3641
LFOS_18_1_8		dc	5461,5461,5462
LFOS_19_1_12		dc	8192
LFOS_20_1_16		dc	10922,10923,10923	
LFOS_21_1_24		dc	16384
LFOS_22_1_32		dc	21845,21845,21846
LFOS_23_1_48		dc	32768
LFOS_24_1_96		dc	65536
     
LFOSyncParas		; 	Sequence LUT address, sequence LUT length, sustainPhasePerClock

			dc	LFOS_01_32_1,3,0.5/3072.0
			dc	LFOS_02_24_1,9,0.5/2304.0
			dc	LFOS_03_16_1,3,0.5/1536.0
			dc	LFOS_04_12_1,9,0.5/1152.0
			dc	LFOS_05_9_1,27,0.5/864.0
			dc	LFOS_06_8_1,3,0.5/768.0
			dc	LFOS_07_7_1,21,0.5/672.0
			dc	LFOS_08_6_1,9,0.5/576.0
			dc	LFOS_09_5_1,15,0.5/480.0
			dc	LFOS_10_4_1,3,0.5/384.0
			dc	LFOS_11_3_1,9,0.5/288.0
			dc	LFOS_12_2_1,3,0.5/192.0
			dc	LFOS_13_1_1,3,0.5/96.0
			dc	LFOS_14_1_2,3,0.5/48.0
			dc	LFOS_15_1_3,9,0.5/36.0
			dc	LFOS_16_1_4,3,0.5/24.0
			dc	LFOS_17_1_6,9,0.5/18.0
			dc	LFOS_18_1_8,3,0.5/12.0
			dc	LFOS_19_1_12,1,0.5/8.0
			dc	LFOS_20_1_16,3,0.5/6.0
			dc	LFOS_21_1_24,1,0.5/4.0
			dc	LFOS_22_1_32,3,0.5/3.0
			dc	LFOS_23_1_48,1,0.5/2.0
			dc	LFOS_24_1_96,1,0.5/1.0
	
primes			; First 32 prime numbers (for accessing contents of random number array)

			dc	1,2,3,5,7,11,13,17,19,23,29,31,37,41,43,47,53,59
			dc	61,67,71,73,79,83,89,97,101,103,107,109,113,127

; List of parameters to define an audio filter and its characteristics

fltSelectLUT	
fltSelectLUT12dB	dc	fpc12dB64K+$8000 	; address of filter pitch correction LUT
			dc	res12dB64K+$8000	; address of resonance correction LUT

			dc	256*0000		; Pitch offsets of filter stages in cents
			dc     	256*7856		; (and shifted left 8 bits to form 24-bit fraction)

			dc	256*30000		; (Stages 3 & 4 bypassed in 12dB filter)
			dc      256*30000

			dc	0			; [unused]
			dc	0			; [unused]

fltSelectLUTsize	equ	*-fltSelectLUT

; Pink noise coefficients used in pink noise filter

pinkConstants		dc	0.986995		; Ka
			dc	0.877306		; Kb
			dc	0.0117927		; Ca
			dc	0.0259605		; Cb
			dc	0.0211587		; Cd
			dc	0.173071		; Kc
			dc	0.0499907		; Cc

	IF	@CVI(*)>$800
	FAIL	'Constants declared in memory not present in DSP56303 power-up state'
	ENDIF

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -  

	org	X:$000800				; noise256 has to start on boundary	

; Noise samples (copied from DRAM for quick access)

r250area						; Used during initialisation
noise256		ds	256	

;------------------------------------------------
; Parameter Filter Class
;------------------------------------------------

; Class data members

ParaFilt_nIter		equ	0
ParaFilt_incr		equ	1
ParaFilt_count		equ	2
ParaFilt_base		equ	3
ParaFilt_delta		equ	4
ParaFilt_output		equ	5
ParaFilt_target		equ	6

ParaFiltSize		equ	ParaFilt_target+1

; Class objects

Ctr1ParaFilt		ds	ParaFiltSize
Ctr2ParaFilt		ds	ParaFiltSize
Ctr3ParaFilt		ds	ParaFiltSize
MmodParaFilt		ds	ParaFiltSize
MvolParaFilt		ds	ParaFiltSize
MPanParaFilt		ds	ParaFiltSize
EditParaFilt		ds	ParaFiltSize
PsftParaFilt		ds	ParaFiltSize

;------------------------------------------------------------------------------- 
; Matrix data 
;------------------------------------------------------------------------------- 

; Results from pre-calculated matrix (same structure as sound.asm)

soundStructurePre
matrixDirectPre		ds	nDirect
matrixPoPsPre		ds	nPoPs

; Copy of some matrix outputs from voice with most recently assigned note
; (LFO5 & LFO6 use paras from most recently played note)

LFO56Direct		ds	8		; LFO5rate, LFO5wave...
LFO56PoPs		ds	2		; LFO5levl, LFO6levl

; Offsets used in the per-voice matrix calculation

nsLFOparasDirect	equ	LFO6del-LFO4del
nsLFOparasPoPs		equ	LFO6levl-LFO4levl

; Source locations with absolute addressing

matrixSources		ds	nMatrixSources
matrixPoPSources	ds	nMatrixSources

; These are the matrix source values that are unique to a voice and are 
; generated in a parameter cycle to be used as input to the next.

vKFLW			equ	0
vKVEL			equ	1
vVDIF			equ	2
vSTDF			equ	3
vLFO1			equ	4
vLFO2			equ	5
vLFO3			equ	6
vLFO4			equ	7
vLFO5			equ	8
vLFO6			equ	9
vENV1			equ	10
vENV2			equ	11

nvMatrixSources		equ	12

; These offsets are duplicated for PoP sources

vMatrixPoPoffset	equ	nvMatrixSources
vMatrixStateSize	equ	nvMatrixSources*2

vMatrixSources		ds	vMatrixStateSize*N_VOICES_MEM

;------------------------------------------------------------------------------- 

; Precalculated values used to simplify PoP matrices

soundPnData		ds	256		; Translated pn = 1 - gn values (maximum array size of 256 is
						; specified in ParaStructure::makeDspParaList() in sound.cpp)

dummyPnData		ds	1		; Location to write pn=1-gn during direct matrix para edits

;------------------------------------------------------------------------------- 
; 		   		DRAM
;------------------------------------------------------------------------------- 
; Note that DRAM is considered as Y: space, but in fact is X: or P: as well. 
; There is one place in the code where it is treated as X:, to speed up a loop. 
; This is at the start of TransferNewSound.
;------------------------------------------------------------------------------- 

	org	Y:$404000	; See contents of lut.asm for this area of memory

	org	Y:$405000

soundTransferBuffer	ds	$1000
sine4096		ds	$1000
sine1024		ds	$0400

F2Ia			ds	$0100	; Frequency-to-integrator-feedback for filters
F2Ib			ds	$0100
F2Ic			ds	$0100
			
P2Fc			ds	$0100	; Pitch-to-frequency for oscillators
P2Ff			ds	$0100

; Bipolar LUTs
			ds	$0200
exp1LUT1K		ds	$0200
			ds	$0200
exp2LUT1K		ds	$0200
			ds	$0200
sqrtLUT1K		ds	$0200
			ds	$0200
panLUT1K		ds	$0200
			ds	$0200
pdfLUT1K		ds	$0200

; Unipolar LUTs

sum4LUT1K		ds	$0400
res12dB257		ds	$0200
fpc12dB257		ds	$0200

soundCheck		ds	$0200	; For checking oscillator sine-wave look-up

	org	Y:$410000

P2F64K			ds	$10000	; P2F for oscillators
P2I64K			ds	$10000	; P2I for filters

fpc12dB64K		ds	$10000	; Pitch correction for 12dB filter
res12dB64K		ds	$10000	; Pitch-indexed resonance correction for 12dB filter
fltOvLUT64K		ds	$10000	; Overdrive to filter input level

lim256LUT64K		ds	$10001	; Dynamic output limiter LUT

	org	Y:$480000

; Filter Coefficient LUTS

; These combine data from other filter LUTs, including pitch correction, P2I and
; resonance correction. Index by pitch << 3 (8 items per row, 3 unused)

fltCoeffLUT12dB		ds	$80000	; Filter coefficient LUT (512K)

; Other Large LUTs

sine256K		ds	$40000	; Sine wave

	org	Y:$600000

noise1M			ds	$200000	; Large buffer of randomly accessed pre-calculated noise

