// File: Distortion.java

// Program to fine tune distortion transfer
// function for filters - see distortion.txt 

// This is for Entity Polyphonic, which is simpler 
// than the calculation in the monosynth.

import java.awt.*;
import java.awt.event.*;
import java.awt.geom.*;
import javax.swing.*;

public class Distortion extends JComponent {

	public Distortion() { }
		
	public void paint(Graphics g) {

		int x = 0; int y = 0;
		int oldx = 0; int oldy = 0;
		
		int cx = getSize().width / 2;
		int cy = getSize().height / 2;

		g.translate(cx, cy);
		
		cx -= 30; cy -= 30;

		if(cx < cy) cy = cx;
		if(cy < cx) cx = cy;

		// Draw axes

		g.drawLine(0, -cy, 0, +cy);
		g.drawLine(-cx, 0, +cx, 0);

		g.drawString("Y", -3, -cy-5);
		g.drawString("X", cx + 5, 5);
		g.drawString("+1.0", -40, -cy + 10);
		g.drawString("-1.0", -40, cy);
		g.drawString("+1.0", cx -30, 20);
		g.drawString("-1.0", -cx, 20);
				
		g.setColor(Color.blue);
		g.drawString("f(x) = x - 2(nx)", (-cx>>3)*7, 10-cy);
		g.drawString("n = 0.45v + 0.55", (-cx>>3)*7, 30-cy);
		g.drawString("where curve shape is defined by n", (-cx>>3)*7, 50-cy);
		g.drawString("and n varies between 0.55 and 1.0", (-cx>>3)*7, 70-cy);
		g.drawString("in response to overdrive, v (0..1)", (-cx>>3)*7, 90-cy);
		
		// Draw non-linearity curves - - - - - - - - - - - - - - - - - - - - - - - - - - - 
		
		g.setColor(Color.blue);
		
		for(double overdrive = 0; overdrive <= 1.0; overdrive += 0.1) {

			boolean start = true;

			double N = 0.0;
			double NS = 0.0;

			N = Math.pow(overdrive, 2.0) * 0.45 + 0.55;

			for(double X = -1.0; X <= 1.0; X += 0.01) {

				double Y;

				Y = X - 2.0 * Math.pow(N * X, 3.0);

				x = (int) (X * (double)cx + 0.5); 
				y = (int) (Y * (double)cy + 0.5);

				if(!start) g.drawLine(oldx, -oldy, x, -y);
				else g.setColor(Color.blue);
				
				oldx = x; oldy = y; start = false;
			}
		}
		
		// Draw distortion pre-gain function - - - - - - - - - - - - - - - - - - - - - - - 

		// This transfer function is used to modify the total level of signal supplied to
		// the filter. The filter's overdrive parameter therefore affects the curve
		// of the non-linearity used in the filter itself, and also this pre-gain 
		// characteristic (implemented as a LUT), which is required so that it's
		// possible to filter the signals in an (almost) entirely linear context
		// (provided there is no self-oscillation).

		boolean start = true;

		for(double X = 0.0; X <= 1.0; X += 0.01) {

			double Y = X;

			// f(x) = 1-(1-(x+1-(1-x)^5))^1.87

			Y = 1.0 - Math.pow(1.0 - 0.5 * (X + 1.0 - Math.pow(1.0 - X, 5) ), 1.87);
				
			x = (int) (X * (double)cx + 0.5); 
			y = (int) (Y * (double)cy + 0.5);

			if(!start) g.drawLine(oldx, -oldy, x, -y);
			else g.setColor(Color.red);
				
			oldx = x; oldy = y; start = false;
		}
		
		g.setColor(Color.red);
		g.drawString("f(v) = 1-(1-(v+1-(1-v)^5))^1.87",(cx>>3)*3, (-cy>>2)*3);
		g.drawString("filter pre-gain", (cx>>3)*3, 20+(-cy>>2)*3);
		
		// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
	}
	
	public static void main(String[] args) {

		JFrame frame = new JFrame("Filter distortion curves (shape is a function of overdrive)");
		Container c = frame.getContentPane();
		c.setLayout(new BorderLayout());
		c.add(new Distortion(), BorderLayout.CENTER);
		frame.setSize(700, 700);
		frame.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
		frame.setVisible(true);

	}
}
